unit BCConfig;

{************************************************************************
*                                                                       *
*                 Base Calculator Configuration Options                 *
*                Copyright (C) 1993-2000, John Zaitseff                 *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      1st April, 2000.
  Version:   2.1

  This file contains an object-oriented interface to the various
  configuration options (preferences) that the user can set for Base
  Calculator.  These preferences are read from the system registry on
  start-up and saved there on modification.  The interface to these
  options is encapsulated in TBCConfig.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


uses
  Registry, Graphics, Windows, Calc, BCDefs, BCDisp, BCFonts;


const
  { Prefix to the registry path used to store configuration options.
    This string should NOT be localised }
  RegPathPrefix = '\Software\Zaitseff\Base Calculator\';


type
  { Base Calculator configuration options object }
  TBCConfig = class
    private
      RegData           : TRegistry;

      FCalcMode         : TCalcMode;
      FCalcSize         : TCalcSize;
      FCalcSigned       : boolean;
      FMaxParenLevel    : integer;

      FFontResID        : TResID;
      FStatusFontResID  : TResID;

      FDispColor        : TColor;
      FDispFontOnColor  : TColor;
      FDispFontOffColor : TColor;

      FLeadingZeros     : boolean;
      FSeparator        : TSeparator;
      FUseFontOff       : boolean;
      FUsePalette       : boolean;

      FEnhancedPaste    : boolean;
      FOpPrecStyle      : TOpPrecStyle;

      FSampleValue      : string;

      FShowTipOfDay     : boolean;
      FTipOfDayIndex    : integer;
      FTipOfDayFile     : string;

      FSaveWindowPos    : boolean;
      FUseWindowPos     : boolean;
      FWindowPos        : TPoint;

      function RegReadFontName (const Key : string; const Default : string) : TResID;
      procedure RegWriteFontName (const Key : string; Value : TResID);

      procedure ReadCalcMode;
      procedure WriteCalcMode;
      procedure ReadCalcSize;
      procedure WriteCalcSize;
      procedure ReadCalcSigned;
      procedure WriteCalcSigned;

      procedure ReadMaxParenLevel;
      procedure ReadOpPrecStyle;
      procedure ReadSeparator;

      procedure SetMaxParenLevel (AlwaysWrite : integer; const Value : integer);

      procedure SetFontResID (AlwaysWrite : integer; const Value : TResID);
      procedure SetStatusFontResID (AlwaysWrite : integer; const Value : TResID);

      procedure SetDispColor (AlwaysWrite : integer; const Value : TColor);
      procedure SetDispFontOnColor (AlwaysWrite : integer; const Value : TColor);
      procedure SetDispFontOffColor (AlwaysWrite : integer; const Value : TColor);

      procedure SetLeadingZeros (AlwaysWrite : integer; const Value : boolean);
      procedure SetSeparator (AlwaysWrite : integer; const Value : TSeparator);
      procedure SetUseFontOff (AlwaysWrite : integer; const Value : boolean);
      procedure SetUsePalette (AlwaysWrite : integer; const Value : boolean);

      procedure SetEnhancedPaste (AlwaysWrite : integer; const Value : boolean);
      procedure SetOpPrecStyle (AlwaysWrite : integer; const Value : TOpPrecStyle);

      procedure SetSampleValue (AlwaysWrite : integer; const Value : string);

      procedure SetShowTipOfDay (AlwaysWrite : integer; const Value : boolean);
      procedure SetTipOfDayIndex (AlwaysWrite : integer; const Value : integer);
      procedure SetTipOfDayFile (AlwaysWrite : integer; const Value : string);

      procedure SetSaveWindowPos (AlwaysWrite : integer; const Value : boolean);
      procedure SetUseWindowPos (AlwaysWrite : integer; const Value : boolean);
      procedure SetWindowPos (AlwaysWrite : integer; const Value : TPoint);

    public
      constructor Create (const RegPath : string);
      destructor Destroy; override;

      property CalcMode         : TCalcMode    read FCalcMode   write FCalcMode;
      property CalcSize         : TCalcSize    read FCalcSize   write FCalcSize;
      property CalcSigned       : boolean      read FCalcSigned write FCalcSigned;

      property MaxParenLevel    : integer      index 0 read  FMaxParenLevel
                                                       write SetMaxParenLevel;

      property FontResID        : TResID       index 0 read  FFontResID
                                                       write SetFontResID;
      property StatusFontResID  : TResID       index 0 read  FStatusFontResID
                                                       write SetStatusFontResID;

      property DispColor        : TColor       index 0 read  FDispColor
                                                       write SetDispColor;
      property DispFontOnColor  : TColor       index 0 read  FDispFontOnColor
                                                       write SetDispFontOnColor;
      property DispFontOffColor : TColor       index 0 read  FDispFontOffColor
                                                       write SetDispFontOffColor;

      property LeadingZeros     : boolean      index 0 read  FLeadingZeros
                                                       write SetLeadingZeros;
      property Separator        : TSeparator   index 0 read  FSeparator
                                                       write SetSeparator;
      property UseFontOff       : boolean      index 0 read  FUseFontOff
                                                       write SetUseFontOff;
      property UsePalette       : boolean      index 0 read  FUsePalette
                                                       write SetUsePalette;

      property EnhancedPaste    : boolean      index 0 read  FEnhancedPaste
                                                       write SetEnhancedPaste;
      property OpPrecStyle      : TOpPrecStyle index 0 read  FOpPrecStyle
                                                       write SetOpPrecStyle;

      property SampleValue      : string       index 0 read  FSampleValue
                                                       write SetSampleValue;

      property ShowTipOfDay     : boolean      index 0 read  FShowTipOfDay
                                                       write SetShowTipOfDay;
      property TipOfDayIndex    : integer      index 0 read  FTipOfDayIndex
                                                       write SetTipOfDayIndex;
      property TipOfDayFile     : string       index 0 read  FTipOfDayFile
                                                       write SetTipOfDayFile;

      property SaveWindowPos    : boolean      index 0 read  FSaveWindowPos
                                                       write SetSaveWindowPos;
      property UseWindowPos     : boolean      index 0 read  FUseWindowPos
                                                       write SetUseWindowPos;
      property WindowPos        : TPoint       index 0 read  FWindowPos
                                                       write SetWindowPos;
    end;


implementation


uses
  SysUtils, Classes;


const
  { Default values for configuration options }
  DefCalcMode         = cmDecimal;
  DefCalcSize         = cs32Bit;
  DefCalcSigned       = True;
  DefOpPrecStyle      = psC;
  DefMaxParenLevel    = Calc.DefMaxParenLevel;
  DefFontResID        = 'BCF:LED';
  DefStatusFontResID  = 'BCF:Status';
  DefDispColor        = clBlack;
  DefDispFontOnColor  = $00FFFF;   { Full yellow   }
  DefDispFontOffColor = $000040;   { Very dark red }
  DefLeadingZeros     = False;
  DefSeparator        = spComma;
  DefUseFontOff       = True;
  DefUsePalette       = True;
  DefEnhancedPaste    = True;
  DefSampleValue      = '8C90F7';
  DefShowTipOfDay     = True;
  DefTipOfDayIndex    = 0;
  DefTipOfDayFile     = 'TIPS.DAT';
  DefSaveWindowPos    = True;
  DefUseWindowPos     = False;
  DefWindowPosX       = 0;
  DefWindowPosY       = 0;

  { Strings used for the system registry key entries.
    These should NOT be localised }
  RK_CalcMode         = 'CalcMode';
  RK_CalcSize         = 'CalcSize';
  RK_CalcSigned       = 'CalcSigned';
  RK_OpPrecStyle      = 'OpPrecStyle';
  RK_MaxParenLevel    = 'MaxParenLevel';
  RK_FontResID        = 'FontName';
  RK_StatusFontResID  = 'StatusFontName';
  RK_DispColor        = 'BackColor';
  RK_DispFontOnColor  = 'FontOnColor';
  RK_DispFontOffColor = 'FontOffColor';
  RK_LeadingZeros     = 'LeadingZeros';
  RK_Separator        = 'Separator';
  RK_UseFontOff       = 'UseFontOff';
  RK_UsePalette       = 'UsePalette';
  RK_EnhancedPaste    = 'EnhancedPaste';
  RK_SampleValue      = 'SampleValue';
  RK_ShowTipOfDay     = 'ShowTipOfDay';
  RK_TipOfDayIndex    = 'TipOfDayIndex';
  RK_TipOfDayFile     = 'TipOfDayFile';
  RK_SaveWindowPos    = 'SaveWindowPos';
  RK_UseWindowPos     = 'UseWindowPos';
  RK_WindowPosX       = 'WindowPosX';
  RK_WindowPosY       = 'WindowPosY';


{***********************************************************************}
{ Create the configuration object and read all of the configuration
  information from the system registry, using RegPath as the root of the
  registry path.  If a particular configuration option is not present,
  a default value is used and saved }
constructor TBCConfig.Create (const RegPath : string);

begin
  RegData := TRegistry.Create;

  RegData.OpenKey(RegPath, True);

  ReadCalcMode;
  ReadCalcSize;
  ReadCalcSigned;
  ReadOpPrecStyle;
  ReadMaxParenLevel;

  FFontResID := RegReadFontName(RK_FontResID, DefFontResID);
  FStatusFontResID := RegReadFontName(RK_StatusFontResID, DefStatusFontResID);

  ReadSeparator;

  try
    FMaxParenLevel := RegData.ReadInteger(RK_MaxParenLevel);
    FDispColor := RegData.ReadInteger(RK_DispColor);
    FDispFontOnColor := RegData.ReadInteger(RK_DispFontOnColor);
    FDispFontOffColor := RegData.ReadInteger(RK_DispFontOffColor);
    FLeadingZeros := RegData.ReadBool(RK_LeadingZeros);
    FUseFontOff := RegData.ReadBool(RK_UseFontOff);
    FUsePalette := RegData.ReadBool(RK_UsePalette);
    FEnhancedPaste := RegData.ReadBool(RK_EnhancedPaste);
    FSampleValue := RegData.ReadString(RK_SampleValue);
    FShowTipOfDay := RegData.ReadBool(RK_ShowTipOfDay);
    FTipOfDayIndex := RegData.ReadInteger(RK_TipOfDayIndex);
    FTipOfDayFile := RegData.ReadString(RK_TipOfDayFile);
    FSaveWindowPos := RegData.ReadBool(RK_SaveWindowPos);
    FUseWindowPos := RegData.ReadBool(RK_UseWindowPos);
    FWindowPos.X := RegData.ReadInteger(RK_WindowPosX);
    FWindowPos.Y := RegData.ReadInteger(RK_WindowPosY)
  except
    SetMaxParenLevel(Ord(True), DefMaxParenLevel);
    SetDispColor(Ord(True), DefDispColor);
    SetDispFontOnColor(Ord(True), DefDispFontOnColor);
    SetDispFontOffColor(Ord(True), DefDispFontOffColor);
    SetLeadingZeros(Ord(True), DefLeadingZeros);
    SetSeparator(Ord(True), DefSeparator);
    SetUseFontOff(Ord(True), DefUseFontOff);
    SetUsePalette(Ord(True), DefUsePalette);
    SetEnhancedPaste(Ord(True), DefEnhancedPaste);
    SetSampleValue(Ord(True), DefSampleValue);
    SetShowTipOfDay(Ord(True), DefShowTipOfDay);
    SetTipOfDayIndex(Ord(True), DefTipOfDayIndex);
    SetTipOfDayFile(Ord(True), DefTipOfDayFile);
    SetSaveWindowPos(Ord(True), DefSaveWindowPos);
    SetUseWindowPos(Ord(True), DefUseWindowPos);
    SetWindowPos(Ord(True), Point(DefWindowPosX, DefWindowPosY))
  end
end;


{***********************************************************************}
{ Destroy the configuration object.  Note that CalcMode, CalcSize and
  CalcSigned are only saved here, instead of at each modification, due
  to the frequency at which these values are changed }
destructor TBCConfig.Destroy;

begin
  if Assigned(RegData) then
    begin
      WriteCalcMode;
      WriteCalcSize;
      WriteCalcSigned;

      RegData.CloseKey;
      RegData.Free
    end;

  inherited Destroy
end;


{***********************************************************************}
{ Read the internal name of a font resource ID from the registry and
  convert it to the appropriate font resource ID.  If the registry key
  does not exist, or is not valid, Default is written out to the registry
  and is used for the resource ID }
function TBCConfig.RegReadFontName (const Key : string; const Default : string) : TResID;

begin
  Result := BCalcFonts.InternalNameToFontResID(RegData.ReadString(Key));
  if Result = 0 then
    begin
      Result := BCalcFonts.InternalNameToFontResID(Default);
      RegData.WriteString(Key, Default)
    end
end;


{***********************************************************************}
{ Write the internal name of the font resource ID to the registry }
procedure TBCConfig.RegWriteFontName (const Key : string; Value : TResID);

begin
  RegData.WriteString(Key, BCalcFonts.FontResIDToInternalName(Value))
end;


{***********************************************************************}
{ Read in the saved value of the calculator mode from the RK_CalcMode
  key.  This value is a decimal number representing the base }
procedure TBCConfig.ReadCalcMode;

var
  I : integer;

begin
  if RegData.ValueExists(RK_CalcMode) then
    begin
      I := RegData.ReadInteger(RK_CalcMode);
      case I of
        10 : FCalcMode := cmDecimal;
        16 : FCalcMode := cmHexadecimal;
        2  : FCalcMode := cmBinary;
        8  : FCalcMode := cmOctal
      else
        FCalcMode := DefCalcMode;
        WriteCalcMode
      end
    end
  else
    begin
      FCalcMode := DefCalcMode;
      WriteCalcMode
    end
end;


{***********************************************************************}
{ Write the current calculator mode as an integer number }
procedure TBCConfig.WriteCalcMode;

begin
  RegData.WriteInteger(RK_CalcMode, CalcModeToInt[FCalcMode])
end;


{***********************************************************************}
{ Read the saved value of the calculator word size }
procedure TBCConfig.ReadCalcSize;

var
  I : integer;

begin
  if RegData.ValueExists(RK_CalcSize) then
    begin
      I := RegData.ReadInteger(RK_CalcSize);
      case I of
        8   : FCalcSize := cs8Bit;
        16  : FCalcSize := cs16Bit;
        32  : FCalcSize := cs32Bit;
        64  : FCalcSize := cs64Bit;
        96  : FCalcSize := cs96Bit;
        128 : FCalcSize := cs128Bit
      else
        FCalcSize := DefCalcSize;
        WriteCalcSize
      end
    end
  else
    begin
      FCalcSize := DefCalcSize;
      WriteCalcSize
    end
end;


{***********************************************************************}
{ Write the current value of the calculator word size as an integer }
procedure TBCConfig.WriteCalcSize;

begin
  RegData.WriteInteger(RK_CalcSize, CalcSizeToBits[FCalcSize])
end;


{***********************************************************************}
{ Read the saved value of the calculator signed flag, writing it if not
  present }
procedure TBCConfig.ReadCalcSigned;

begin
  try
    FCalcSigned := RegData.ReadBool(RK_CalcSigned)
  except
    FCalcSigned := DefCalcSigned;
    WriteCalcSigned
  end
end;


{***********************************************************************}
{ Write the current value of the calculator signed flag to the registry }
procedure TBCConfig.WriteCalcSigned;

begin
  RegData.WriteBool(RK_CalcSigned, FCalcSigned)
end;


{***********************************************************************}
{ Read the MaxParenLevel value from the registry.  If the registry key
  does not exist, or is invalid, the default value is written to the
  registry and used as the maximum parenthesis nesting level }
procedure TBCConfig.ReadMaxParenLevel;

var
  I : integer;

begin
  if RegData.ValueExists(RK_MaxParenLevel) then
    begin
      I := RegData.ReadInteger(RK_MaxParenLevel);
      if (I >= 0) and (I <= Calc.AbsMaxParenLevel) then
        FMaxParenLevel := I
      else
        SetMaxParenLevel(Ord(True), DefMaxParenLevel)
    end
  else
    SetMaxParenLevel(Ord(True), DefMaxParenLevel)
end;


{***********************************************************************}
{ Read the OpPrecStyle value from the registry.  If the registry key does
  not exist, or is invalid, the default value is written to the registry
  and used as the operator precedence style }
procedure TBCConfig.ReadOpPrecStyle;

var
  I : integer;

begin
  if RegData.ValueExists(RK_OpPrecStyle) then
    begin
      I := RegData.ReadInteger(RK_OpPrecStyle);
      if (I >= Ord(Low(TOpPrecStyle))) and (I <= Ord(High(TOpPrecStyle))) then
        begin
          FOpPrecStyle := TOpPrecStyle(I);
          exit
        end
    end;

  { The key did not exist, or contained invalid data }
  SetOpPrecStyle(Ord(True), DefOpPrecStyle)
end;


{***********************************************************************}
{ Read the Separator value from the registry.  If the registry key does
  not exist, or is invalid, the default value is written to the registry
  and used as the separator }
procedure TBCConfig.ReadSeparator;

var
  I : integer;

begin
  if RegData.ValueExists(RK_Separator) then
    begin
      I := RegData.ReadInteger(RK_Separator);
      if (I >= Ord(Low(TSeparator))) and (I <= Ord(High(TSeparator))) then
        begin
          FSeparator := TSeparator(I);
          exit
        end
    end;

  { The key did not exist, or contained invalid data }
  SetSeparator(Ord(True), DefSeparator)
end;


{***********************************************************************}
{ Store the new value of MaxParenLevel and write it to the registry,
  if required }
procedure TBCConfig.SetMaxParenLevel (AlwaysWrite : integer; const Value : integer);

begin
  if (AlwaysWrite <> 0) or (Value <> FMaxParenLevel) then
    if (Value >= 0) and (Value <= Calc.AbsMaxParenLevel) then
      begin
        FMaxParenLevel := Value;
        RegData.WriteInteger(RK_MaxParenLevel, Value)
      end
end;


{***********************************************************************}
{ Store the new value of FontResID and write it to the registry,
  if required }
procedure TBCConfig.SetFontResID (AlwaysWrite : integer; const Value : TResID);

begin
  if (AlwaysWrite <> 0) or (Value <> FFontResID) then
    begin
      FFontResID := Value;
      RegWriteFontName(RK_FontResID, Value)
    end
end;


{***********************************************************************}
{ Store the new value of StatusFontResID and write it to the registry,
  if required }
procedure TBCConfig.SetStatusFontResID (AlwaysWrite : integer; const Value : TResID);

begin
  if (AlwaysWrite <> 0) or (Value <> FStatusFontResID) then
    begin
      FStatusFontResID := Value;
      RegWriteFontName(RK_StatusFontResID, Value)
    end
end;


{***********************************************************************}
{ Store the new value of DispColor and write it to the registry,
  if required }
procedure TBCConfig.SetDispColor (AlwaysWrite : integer; const Value : TColor);

begin
  if (AlwaysWrite <> 0) or (Value <> FDispColor) then
    begin
      FDispColor := Value;
      RegData.WriteInteger(RK_DispColor, Value)
    end
end;


{***********************************************************************}
{ Store the new value of DispFontOnColor and write it to the registry,
  if required }
procedure TBCConfig.SetDispFontOnColor (AlwaysWrite : integer; const Value : TColor);

begin
  if (AlwaysWrite <> 0) or (Value <> FDispFontOnColor) then
    begin
      FDispFontOnColor := Value;
      RegData.WriteInteger(RK_DispFontOnColor, Value)
    end
end;


{***********************************************************************}
{ Store the new value of DispFontOffColor and write it to the registry,
  if required }
procedure TBCConfig.SetDispFontOffColor (AlwaysWrite : integer; const Value : TColor);

begin
  if (AlwaysWrite <> 0) or (Value <> FDispFontOffColor) then
    begin
      FDispFontOffColor := Value;
      RegData.WriteInteger(RK_DispFontOffColor, Value)
    end
end;


{***********************************************************************}
{ Store the new value of LeadingZeros and write it to the registry,
  if required }
procedure TBCConfig.SetLeadingZeros (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FLeadingZeros) then
    begin
      FLeadingZeros := Value;
      RegData.WriteBool(RK_LeadingZeros, Value)
    end
end;


{***********************************************************************}
{ Store the new value of Separator and write it to the registry,
  if required }
procedure TBCConfig.SetSeparator (AlwaysWrite : integer; const Value : TSeparator);

begin
  if (AlwaysWrite <> 0) or (Value <> FSeparator) then
    begin
      FSeparator := Value;
      RegData.WriteInteger(RK_Separator, Ord(Value))
    end
end;


{***********************************************************************}
{ Store the new value of UseFontOff and write it to the registry,
  if required }
procedure TBCConfig.SetUseFontOff (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FUseFontOff) then
    begin
      FUseFontOff := Value;
      RegData.WriteBool(RK_UseFontOff, Value)
    end
end;


{***********************************************************************}
{ Store the new value of UsePalette and write it to the registry,
  if required }
procedure TBCConfig.SetUsePalette (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FUsePalette) then
    begin
      FUsePalette := Value;
      RegData.WriteBool(RK_UsePalette, Value)
    end
end;


{***********************************************************************}
{ Store the new value of EnhancedPaste and write it to the registry,
  if required }
procedure TBCConfig.SetEnhancedPaste (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FEnhancedPaste) then
    begin
      FEnhancedPaste := Value;
      RegData.WriteBool(RK_EnhancedPaste, Value)
    end
end;


{***********************************************************************}
{ Store the new value of OpPrecStyle and write it to the registry, if
  required }
procedure TBCConfig.SetOpPrecStyle (AlwaysWrite : integer; const Value : TOpPrecStyle);

begin
  if (AlwaysWrite <> 0) or (Value <> FOpPrecStyle) then
    begin
      FOpPrecStyle := Value;
      RegData.WriteInteger(RK_OpPrecStyle, Ord(Value))
    end
end;


{***********************************************************************}
{ Store the new value of SampleValue and write it to the registry,
  if required }
procedure TBCConfig.SetSampleValue (AlwaysWrite : integer; const Value : string);

begin
  if (AlwaysWrite <> 0) or (Value <> FSampleValue) then
    begin
      FSampleValue := Value;
      RegData.WriteString(RK_SampleValue, Value)
    end
end;


{***********************************************************************}
{ Store the new value of ShowTipOfDay and write it to the registry,
  if required }
procedure TBCConfig.SetShowTipOfDay (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FShowTipOfDay) then
    begin
      FShowTipOfDay := Value;
      RegData.WriteBool(RK_ShowTipOfDay, Value)
    end
end;


{***********************************************************************}
{ Store the new value of TipOfDayIndex and write it to the registry,
  if required }
procedure TBCConfig.SetTipOfDayIndex (AlwaysWrite : integer; const Value : integer);

begin
  if (AlwaysWrite <> 0) or (Value <> FTipOfDayIndex) then
    begin
      FTipOfDayIndex := Value;
      RegData.WriteInteger(RK_TipOfDayIndex, Value)
    end
end;


{***********************************************************************}
{ Store the new value of TipOfDayFile and write it to the registry,
  if required }
procedure TBCConfig.SetTipOfDayFile (AlwaysWrite : integer; const Value : string);

begin
  if (AlwaysWrite <> 0) or (Value <> FTipOfDayFile) then
    begin
      FTipOfDayFile := Value;
      RegData.WriteString(RK_TipOfDayFile, Value)
    end
end;


{***********************************************************************}
{ Store the new value of SaveWindowPos and write it to the registry,
  if required }
procedure TBCConfig.SetSaveWindowPos (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FSaveWindowPos) then
    begin
      FSaveWindowPos := Value;
      RegData.WriteBool(RK_SaveWindowPos, Value)
    end
end;


{***********************************************************************}
{ Store the new value of UseWindowPos and write it to the registry,
  if required }
procedure TBCConfig.SetUseWindowPos (AlwaysWrite : integer; const Value : boolean);

begin
  if (AlwaysWrite <> 0) or (Value <> FUseWindowPos) then
    begin
      FUseWindowPos := Value;
      RegData.WriteBool(RK_UseWindowPos, Value)
    end
end;


{***********************************************************************}
{ Store the new value of WindowPos and write it to the registry,
  if required }
procedure TBCConfig.SetWindowPos (AlwaysWrite : integer; const Value : TPoint);

begin
  if (AlwaysWrite <> 0) or
     (Value.X <> FWindowPos.X) or (Value.Y <> FWindowPos.Y)
  then
    begin
      FWindowPos := Value;
      RegData.WriteInteger(RK_WindowPosX, Value.X);
      RegData.WriteInteger(RK_WindowPosY, Value.Y)
    end
end;

end.

