unit Calc;

{************************************************************************
*                                                                       *
*                     Base Calculator State Machine                     *
*                Copyright (C) 1993-2000, John Zaitseff                 *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      1st April, 2000.
  Version:   2.1

  This file provides the implementation of a state machine for a base
  integer calculator, ie, one that operates in decimal, hexadecimal,
  binary and octal, with a word size of 8, 16, 32, 64, 96 or 128 bits.
  The TCalculator object implemented in this unit keeps track of all the
  necessary state information required by such a calculator.

  Note that Borland Delphi 4 or later is required to compile this unit.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


uses
  BCDefs, Math128;


const
  AbsMaxParenLevel = 4095;  { Absolute maximum amount of parenthesis nesting }
  DefMaxParenLevel = 32;    { Default max. amount of parenthesis nesting }


type
  { Calculator key press }
  TCalcKey     = (ck0, ck1, ck2, ck3, ck4, ck5, ck6, ck7,
                  ck8, ck9, ckA, ckB, ckC, ckD, ckE, ckF,
                  ckAdd, ckSubtract, ckMultiply, ckDivide, ckMod,
                  ckNot, ckNeg, ckAnd, ckNand, ckOr, ckNor, ckXor, ckEqv,
                  ckSHL, ckSHR, ckROL, ckROR, ckLParen, ckRParen,
                  ckEquals, ckMPlus, ckMMinus, ckMIn, ckMR,
                  ckClear, ckBackspace, ckSwapXY, ckSwapXM, ckNone);

  { Operator precedence styles: C or Pascal language }
  TOpPrecStyle = (psC, psPascal);


  { The following types are internal to TCalculator }

  { Calculator state machine entry state }
  TEntryState  = (esFirstKey, esNextKey, esError);

  { State of FString and FStack[FStackPtr].Value (TOS) }
  TValueState  = (vsBothValid,             { Both FString and TOS valid }
                  vsStringValid,           { FString valid, TOS invalid }
                  vsStringValidNotRounded, { FString valid, TOS not rounded }
                  vsNumberValid,           { FString invalid, TOS valid }
                  vsNotRounded);           { FString invalid, TOS not rounded }

  { Calculator operations stack }
  TCalcStack   = record
                   Value      : Int128;
                   Operation  : TCalcKey;
                   ParenLevel : integer
                 end;


type
  { Base Calculator state machine object }
  TCalculator = class
    private
      FMode          : TCalcMode;           { Current calculator mode }
      FSize          : TCalcSize;           { Current calculator word size }
      FSigned        : boolean;             { Current calculator signed flag }

      FEntryState    : TEntryState;         { Current state machine state }
      FValueState    : TValueState;         { State of FString and top of stack }
      FLastKey       : TCalcKey;            { Last key pressed (or simulated) }
      FConstantMode  : boolean;             { True if in CONST mode }

      FString        : string;              { String representation of value }

      FStack         : array of TCalcStack; { Calculator stack; dynamic array }
      FStackPtr      : integer;             { Pointer to CURRENT top of stack }
      FMaxParenLevel : integer;             { Max. level of parenthesis nesting }

      FMemory        : Int128;              { Value in calculator's memory }
      FMemInUse      : boolean;             { True if memory is used }

      FOpPrecStyle   : TOpPrecStyle;        { Styles of operator precedence }

      procedure SetMode (const AValue : TCalcMode);
      procedure SetSize (const AValue : TCalcSize);
      procedure SetSigned (const AValue : boolean);

      procedure UpdateCurrentValue (DoRound : boolean);
      procedure MakeStringInvalid (AssumeRounded : boolean);

      function GetCurrentValue : Int128;
      function GetCurrentString : string;

      function GetInError : boolean;

      function GetCurrentCalcOp : TCalcOp;
      function GetConstantMode : boolean;
      function GetParenLevel : integer;
      procedure SetMaxParenLevel (const Value : integer);

      procedure SetOpPrecStyle (const Value : TOpPrecStyle);

    public
      constructor Create; overload;
      constructor Create (const ParenLevel : integer); overload;
      destructor Destroy; override;

      property Mode          : TCalcMode    read FMode   write SetMode;
      property Size          : TCalcSize    read FSize   write SetSize;
      property Signed        : boolean      read FSigned write SetSigned;

      property CurrentValue  : Int128       read GetCurrentValue;
      property CurrentString : string       read GetCurrentString;

      property InError       : boolean      read GetInError;

      property MemoryValue   : Int128       read FMemory;
      property MemInUse      : boolean      read FMemInUse;

      property CurrentCalcOp : TCalcOp      read GetCurrentCalcOp;
      property ConstantMode  : boolean      read GetConstantMode;
      property ParenLevel    : integer      read GetParenLevel;
      property MaxParenLevel : integer      read FMaxParenLevel write SetMaxParenLevel;

      property OpPrecStyle   : TOpPrecStyle read FOpPrecStyle   write SetOpPrecStyle;

      procedure ClearAll;
      procedure ClearOperations;  { Clear key }
      procedure ClearMemory;

      function HandleKey (const Key : TCalcKey) : boolean;
      function AppendDigit (const Digit : integer) : boolean;
    end;


function ValToStr (Value : Int128; const Mode : TCalcMode;
                   const Size : TCalcSize; Signed : boolean) : string;
  { Convert a value to its string representation, as used by Base Calculator }


implementation


uses
  SysUtils, StringRes;


const
  { Convert a digit number into a character }
  DigitChar      : array [0..15] of char        = '0123456789ABCDEF';

  { Convert TCalcMode, TCalcSize into a maximum length of string }
  MaxStrLen      : array [TCalcMode, TCalcSize] of integer =
                   ((  3,   5,  10,  20,  29,  39),   { cmDecimal }
                    (  2,   4,   8,  16,  24,  32),   { cmHexadecimal }
                    (  8,  16,  32,  64,  96, 128),   { cmBinary }
                    (  3,   6,  11,  22,  32,  43));  { cmOctal }

  { Maximum values for the different word sizes in decimal mode }
  MaxDecValues   : array [TCalcSize] of string =
                   ('255',                                      { 8 bits   }
                    '65535',                                    { 16 bits  }
                    '4294967295',                               { 32 bits  }
                    '18446744073709551615',                     { 64 bits  }
                    '79228162514264337593543950335',            { 96 bits  }
                    '340282366920938463463374607431768211455'); { 128 bits }

  { Conversion table from TCalcKey to TCalcOp }
  CalcKeyToCalcOp : array [TCalcKey] of TCalcOp =
                    (coNone, coNone, coNone, coNone,  { ck0..ck3 }
                     coNone, coNone, coNone, coNone,  { ck4..ck7 }
                     coNone, coNone, coNone, coNone,  { ck8..ckB }
                     coNone, coNone, coNone, coNone,  { ckC..ckF }
                     coAdd, coSubtract, coMultiply, coDivide, coMod,
                     coNot, coNeg, coAnd, coNand, coOr, coNor, coXor, coEqv,
                     coSHL, coSHR, coROL, coROR,
                     coNone, coNone,         { ckLParen, ckRParen              }
                     coNone, coNone, coNone, { ckEquals, ckMPlus, ckMMinus     }
                     coNone, coNone, coNone, { ckMIn, ckMR, ckClear            }
                     coNone, coNone, coNone, { ckBackspace, ckSwapXY, ckSwapXM }
                     coNone);                { ckNone                          }


  MaxPrecedence    = 7;  { Number of operator precedence levels in tables below }
  UnusedPrecedence = -1; { "Precedence value" for unused entries }
  FinishAllPrec    = 0;  { Precedence value for "Finish all calculations" }
  FinishParenPrec  = 1;  { Precedence value for "Finish this parenthesis level" }
  StackSize        = MaxPrecedence;  { Number of stack entries to reserve per }
                                     { parenthesis level                      }

  { Tables of operator precedence for TCalcKey; unused keys are set to -1 }
  Precedence : array [TOpPrecStyle, TCalcKey] of integer =
               (                      { C language style operator precedence }
                (-1, -1, -1, -1, -1, -1, -1, -1,    { ck0..ck7 - not used    }
                 -1, -1, -1, -1, -1, -1, -1, -1,    { ck8..ckF - not used    }
                  6,  6,         { ckAdd, ckSubtract           - 2nd highest }
                  7,  7,  7,     { ckMultiply, ckDivide, ckMod - highest     }
                 -1, -1,         { ckNot, ckNeg                - not used    }
                  4,  4,         { ckAnd, ckNand               - 4th highest }
                  2,  2,         { ckOr, ckNor                 - 6th highest }
                  3,  3,         { ckXor, ckEqv                - 5th highest }
                  5,  5,  5,  5, { ckSHL, ckSHR, ckROL, ckROR  - 3rd highest }
                 -1,             { ckLParen                    - not used    }
                  1,             { ckRParen                    - 2nd lowest  }
                  0,  0,  0,     { ckMPlus, ckMMinus, ckEquals - lowest      }
                 -1, -1,         { ckMIn, ckMR                 - not used    }
                 -1, -1,         { ckClear, ckBackspace        - not used    }
                 -1, -1, -1),    { ckSwapXY, ckSwapXM, ckNone  - not used    }

                                 { Pascal language style operator precedence }
                (-1, -1, -1, -1, -1, -1, -1, -1,    { ck0..ck7 - not used    }
                 -1, -1, -1, -1, -1, -1, -1, -1,    { ck8..ckF - not used    }
                  2,  2,         { ckAdd, ckSubtract           - 2nd highest }
                  3,  3,  3,     { ckMultiply, ckDivide, ckMod - highest     }
                 -1, -1,         { ckNot, ckNeg                - not used    }
                  3,  3,         { ckAnd, ckNand               - highest     }
                  2,  2,         { ckOr, ckNor                 - 2nd highest }
                  2,  2,         { ckXor, ckEqv                - 2nd highest }
                  3,  3,  3,  3, { ckSHL, ckSHR, ckROL, ckROR  - highest     }
                 -1,             { ckLParen                    - not used    }
                  1,             { ckRParen                    - 2nd lowest  }
                  0,  0,  0,     { ckMPlus, ckMMinus, ckEquals - lowest      }
                 -1, -1,         { ckMIn, ckMR                 - not used    }
                 -1, -1,         { ckClear, ckBackspace        - not used    }
                 -1, -1, -1)     { ckSwapXY, ckSwapXM, ckNone  - not used    }
               );


{***********************************************************************}
{ Round a value to 8, 16, 32, 64, 96 or 128 bits and to the appropriate
  sign.  Note that if Size = cs128Bit, nothing needs to be done, as the
  number is already rounded }
procedure RoundVal (var Value : Int128; Size : TCalcSize; Signed : boolean);

begin
  if Size <> cs128Bit then
    Value := iExtend(Value, CalcSizeToBits[Size], Signed)
end;


{***********************************************************************}
{ Convert a value to its string representation, as used by Base Calculator.
  In particular, a negative sign is included ONLY if Mode = cmDecimal }
function ValToStr (Value : Int128; const Mode : TCalcMode;
                   const Size : TCalcSize; Signed : boolean) : string;

begin
  if Mode <> cmDecimal then
    Signed := False;

  RoundVal(Value, Size, Signed);

  Result := Int128ToStr(Value, 0, CalcModeToInt[Mode], Signed)
end;


{***********************************************************************}
{ Create the calculator state machine class with a default maximum number
  of parenthesis nesting }
constructor TCalculator.Create;

begin
  inherited Create;

  FMaxParenLevel := DefMaxParenLevel;
  SetLength(FStack, (FMaxParenLevel + 1) * StackSize);

  { Use C-language precedence table by default }
  FOpPrecStyle := psC;

  ClearAll
end;


{***********************************************************************}
{ Create the calculator state machine class.  ParenLevel is the maximum
  number of parenthesis nesting that this machine will have to handle.
  This must be between 0 and AbsMaxParenLevel (inclusive), or an
  ERangeError exception will be raised }
constructor TCalculator.Create (const ParenLevel : integer);

begin
  inherited Create;

  if (ParenLevel < 0) or (ParenLevel > AbsMaxParenLevel) then
    raise ERangeError.CreateFmt(SParenLevelRange, [ParenLevel, AbsMaxParenLevel]);

  FMaxParenLevel := ParenLevel;
  SetLength(FStack, (FMaxParenLevel + 1) * StackSize);

  { Use C-language precedence table by default }
  FOpPrecStyle := psC;

  ClearAll
end;


{***********************************************************************}
{ Destroy the calculator state machine class }
destructor TCalculator.Destroy;

begin
  FStack := nil;
  inherited Destroy
end;



{***********************************************************************}
{ Set the calculator mode to one of decimal, hexadecimal, binary or
  octal.  Setting the mode affects the internal state, even if the
  mode was the same as before.  If InError is True, setting the mode
  has no effect }
procedure TCalculator.SetMode (const AValue : TCalcMode);

begin
  if not InError then
    begin
      UpdateCurrentValue(False);
      FMode := AValue;
      MakeStringInvalid(False)
    end
end;


{***********************************************************************}
{ Set the calculator word size to one of 8, 16, 32, 64, 96 or 128 bits.
  Setting the size affects the internal state, even if it was the same
  as before.  If InError is True, setting the size has no effect }
procedure TCalculator.SetSize (const AValue : TCalcSize);

var
  I : integer;

begin
  if not InError then
    begin
      UpdateCurrentValue(True);

      { Truncate all numbers to the new size, if required }
      if AValue < FSize then
        begin
          for I := 0 to FStackPtr do
            with FStack[I] do
              RoundVal(Value, AValue, Signed);
          if MemInUse then
            begin
              RoundVal(FMemory, AValue, Signed);
              FMemInUse := (iSgn(FMemory) <> 0)
            end
        end;

      FSize := AValue;

      MakeStringInvalid(True)
    end
end;


{***********************************************************************}
{ Set the calculator state machine to either signed or unsigned integer
  operations.  Setting the signed flag affects the internal state, even
  if it was the same as before.  If InError is True, setting the signed
  flag has no effect }
procedure TCalculator.SetSigned (const AValue : boolean);

var
  I : integer;

begin
  if not InError then
    begin
      UpdateCurrentValue(False);

      FSigned := AValue;

      { Zero or sign extend the numbers appropriately }
      for I := 0 to FStackPtr do
        with FStack[I] do
          RoundVal(Value, Size, Signed);
      if MemInUse then
        RoundVal(FMemory, Size, Signed);
        { Note that changing the sign will not make FMemory zero.  Hence no
          need to test for this }

      MakeStringInvalid(True)
    end
end;


{***********************************************************************}
{ Ensure that the current value, ie, that at the top of the stack, is
  valid.  In other words, the string representation in FString is converted
  into an Int128 value, if necessary.  In addition, this function will
  round this value if requested }
procedure TCalculator.UpdateCurrentValue (DoRound : boolean);

var
  Value : Int128;

begin
  if InError or (FValueState = vsBothValid) then
    { Exit if calculator is in error, or both the string and value are valid }
    exit;

  if FValueState = vsStringValid then
    begin
      { The string is valid, but the number is not.  Convert the number, but
        do not round it yet }
      Value := StrToInt128(FString, CalcModeToInt[Mode]);
      FStack[FStackPtr].Value := Value;
      FValueState := vsStringValidNotRounded
    end;

  if DoRound and (FValueState <> vsNumberValid) then
    begin
      { If requested, round the value, but only if it has not already been
        rounded.  Note that rounding a value may make the string invalid
        (due, for example, to the number becoming negative) }
      RoundVal(FStack[FStackPtr].Value, Size, Signed);
      FEntryState := esFirstKey;
      if FValueState in [vsNotRounded, vsStringValidNotRounded] then
        FValueState := vsNumberValid
    end
end;


{***********************************************************************}
{ Make the current string representation invalid.  This also sets the
  entry state to esFirstKey.  The GetCurrentString function will make the
  string valid again (although it will not change the entry state) }
procedure TCalculator.MakeStringInvalid (AssumeRounded : boolean);

begin
  if not InError then
    begin
      FEntryState := esFirstKey;

      if FValueState = vsStringValid then
        { This state should not occur in this call! }
        UpdateCurrentValue(AssumeRounded);

      if AssumeRounded or (FValueState = vsBothValid) then
        FValueState := vsNumberValid
      else if FValueState = vsStringValidNotRounded then
        FValueState := vsNotRounded
    end
end;


{***********************************************************************}
{ Retrieve the current value in the calculator.  If InError is True,
  this function returns Zero.  Note that the result is rounded in such
  a way as NOT to affect the calculator entry state }
function TCalculator.GetCurrentValue : Int128;

begin
  if InError then
    Result := Zero
  else
    begin
      UpdateCurrentValue(False);
      Result := FStack[FStackPtr].Value;
      RoundVal(Result, Size, Signed)
    end
end;


{***********************************************************************}
{ Retrieve the current value in the calculator as a string representation.
  If InError is True, this function returns the empty string }
function TCalculator.GetCurrentString : string;

var
  TempSize : TCalcSize;

begin
  if InError then
    Result := ''
  else
    begin
      { There is no need to call UpdateCurrentValue }
      if FValueState = vsNumberValid then
        begin
          { Pass cs128Bit as the size if Signed is False, as in this case we
            do NOT need to round the value }
          if Signed then
            TempSize := Size
          else
            TempSize := cs128Bit;
          FString := ValToStr(FStack[FStackPtr].Value, Mode, TempSize, Signed);
          FValueState := vsBothValid
        end
      else if FValueState = vsNotRounded then
        begin
          FString := ValToStr(FStack[FStackPtr].Value, Mode, Size, Signed);
          FValueState := vsStringValidNotRounded
        end;

      Result := FString
    end
end;


{***********************************************************************}
{ Return True if the calculator state machine is in an error state and
  needs to be cleared by calling ClearOperations }
function TCalculator.GetInError : boolean;

begin
  Result := (FEntryState = esError)
end;


{***********************************************************************}
{ Return the current calculator operation.  If InError is True, the last
  such calculator operation is returned, even though it is essentially
  meaningless (since a ClearOperations call needs to be performed) }
function TCalculator.GetCurrentCalcOp : TCalcOp;

begin
  if (FStackPtr = 0) or
     (FStack[FStackPtr].ParenLevel <> FStack[FStackPtr - 1].ParenLevel)
  then
    Result := coNone
  else
    Result := CalcKeyToCalcOp[FStack[FStackPtr - 1].Operation]
end;


{***********************************************************************}
{ Return the CONST mode indicator.  If InError is True, the last value
  is returned, even though it is essentially meaningless (since a
  ClearOperations call needs to be performed) }
function TCalculator.GetConstantMode : boolean;

begin
  if FStack[FStackPtr].ParenLevel <> 0 then
    Result := False
  else
    Result := FConstantMode
end;


{***********************************************************************}
{ Return the current parenthesis nesting level in the calculator state
  machine.  If InError is True, the last nesting level is returned, even
  though it is essentially meaningless (since a ClearOperations call
  needs to be performed) }
function TCalculator.GetParenLevel : integer;

begin
  Result := FStack[FStackPtr].ParenLevel
end;


{***********************************************************************}
{ Set the maximum level of parenthesis nesting that can occur in the
  calculator state machine.  This number must be between 0 and
  AbsMaxParenLevel inclusive.  Each level of nesting requires StackSize
  stack locations, with each stack location occupying approx. 22 bytes.
  If Value is illegal, an ERangeError exception is raised.  If Value is
  valid, this procedure will reallocate the stack to handle the
  appropriate parenthesis nesting level and clear all operations }
procedure TCalculator.SetMaxParenLevel (const Value : integer);

begin
  if (Value < 0) or (Value > AbsMaxParenLevel) then
    raise ERangeError.CreateFmt(SParenLevelRange, [Value, AbsMaxParenLevel]);

  FMaxParenLevel := Value;
  SetLength(FStack, (FMaxParenLevel + 1) * StackSize);
  ClearOperations
end;


{***********************************************************************}
{ Set the precedence table to reflect the requested style of operator
  precedence (either psC or psPascal).  Note that setting this property
  to a new (ie, different) value clears operations }
procedure TCalculator.SetOpPrecStyle (const Value : TOpPrecStyle);

begin
  if Value <> FOpPrecStyle then
    begin
      FOpPrecStyle := Value;
      ClearOperations
    end
end;


{***********************************************************************}
{ Clear the calculator state machine to its initial values }
procedure TCalculator.ClearAll;

begin
  { Set the default values }
  FMode := cmDecimal;
  FSize := cs32Bit;
  FSigned := True;

  ClearOperations;
  ClearMemory
end;


{***********************************************************************}
{ Clear all of the calculator operations.  This is the same as
  HandleKey(ckClear), ie, the <ESC> key or [Clear] button in the Base
  Calculator application }
procedure TCalculator.ClearOperations;

begin
  { Push a zero onto the stack }
  FStackPtr := 0;
  with FStack[FStackPtr] do
    begin
      Value := Zero;
      Operation := ckNone;
      ParenLevel := 0
    end;
  FConstantMode := False;

  FString := '0';

  FEntryState := esFirstKey;
  FValueState := vsBothValid;
  FLastKey := ckNone
end;


{***********************************************************************}
{ Clear the calculator's memory }
procedure TCalculator.ClearMemory;

begin
  FMemInUse := False;
  FMemory := Zero
end;


{***********************************************************************}
{ Handle a calculator key (digits, operations and functions).  This
  function returns True if the key was handled successfully.  Note that
  the display may still need to be updated if False is returned.
  Additionally, InError may also return True after this function }
function TCalculator.HandleKey (const Key : TCalcKey) : boolean;

  { Internal function:  Handle the Backspace key.  This function only works
    (and returns True) if a digit key has already been pressed (ie,
    FEntryState is esNextKey).  Note that the string representation is
    always valid when FEntryState = esNextKey }
  function Backspace : boolean;

  begin
    if FEntryState <> esNextKey then
      begin
        Result := False;
        exit
      end;

    { Delete the last digit appended to the string representation }
    Delete(FString, Length(FString), 1);

    if (FString = '') or (FString = '-') then
      begin
        FEntryState := esFirstKey;
        FString := '0'
      end;

    FValueState := vsStringValid;
    Result := True
  end;

  { Internal function:  Store the current value in memory, replacing any
    value already present there }
  function StoreInMem : boolean;

  begin
    UpdateCurrentValue(True);

    with FStack[FStackPtr] do
      begin
        FMemory := Value;
        FMemInUse := (iSgn(Value) <> 0)
      end;

    Result := True
  end;

  { Internal function:  Retrieve the value currently in memory, replacing
    the current value }
  function RetrieveMem : boolean;

  begin
    FStack[FStackPtr].Value := FMemory;
    FValueState := vsNumberValid;
    FEntryState := esFirstKey;
    Result := True
  end;

  { Internal function:  Swap the current value (X) with the previous value
    (Y) if that previous value is at the same parenthesis level }
  function SwapXY : boolean;

  var
    Temp : Int128;

  begin
    UpdateCurrentValue(True);

    with FStack[FStackPtr] do
      if (FStackPtr > 0) and (ParenLevel = FStack[FStackPtr - 1].ParenLevel) then
        begin
          { A previous value exists at the same parenthesis level }
          Temp := Value;
          Value := FStack[FStackPtr - 1].Value;
          FStack[FStackPtr - 1].Value := Temp;
          FValueState := vsNumberValid;
          Result := True
        end
      else
        Result := False
  end;

  { Internal function:  Swap the current value (X) with the value stored
    in memory (M) }
  function SwapXM : boolean;

  var
    Temp : Int128;

  begin
    UpdateCurrentValue(True);

    with FStack[FStackPtr] do
      begin
        Temp := Value;
        Value := FMemory;
        FMemory := Temp;
        FMemInUse := (iSgn(FMemory) <> 0);
        FValueState := vsNumberValid
      end;

    Result := True
  end;

  { Internal procedure:  Add the current value to the value in memory.  Note
    that the current value has already been rounded }
  procedure AddToMemory;

  begin
    FMemory := iAdd(FMemory, FStack[FStackPtr].Value);
    FMemInUse := (iSgn(FMemory) <> 0)
  end;

  { Internal procedure:  Subtract the current value from the value in memory.
    Note that the current value has already been rounded }
  procedure SubtractFromMemory;

  begin
    FMemory := iSub(FMemory, FStack[FStackPtr].Value);
    FMemInUse := (iSgn(FMemory) <> 0)
  end;

  { Internal function:  Perform a unary operation, either NOT or NEG, on the
    current value }
  function DoUnaryOp (Key : TCalcKey) : boolean;

  begin
    UpdateCurrentValue(True);

    with FStack[FStackPtr] do
      begin
        case Key of
          ckNot : Value := iNot(Value);
          ckNeg : Value := iNeg(Value)
        end;
        RoundVal(Value, Size, Signed)
      end;

    FValueState := vsNumberValid;
    Result := True
  end;

  { Internal function:  Perform a left parenthesis operation }
  function DoLParen : boolean;

  begin
    if ParenLevel = MaxParenLevel then
      { There is (potentially) no more space in the stack }
      Result := False
    else
      with FStack[FStackPtr] do
        begin
          Value := Zero;
          Inc(ParenLevel);
          FValueState := vsNumberValid;
          Result := True
        end
  end;

  { Internal procedure:  Push the current value onto the stack again, storing
    the current operation and maintaining the parenthesis level }
  procedure PushStack (Key : TCalcKey);

  begin
    if FStackPtr = High(FStack) then
      raise EStackOverflow.Create(SCalcStackOverflow);

    FStack[FStackPtr].Operation := Key;
    Inc(FStackPtr);
    with FStack[FStackPtr] do
      begin
        Value := FStack[FStackPtr - 1].Value;
        Operation := ckNone;
        ParenLevel := FStack[FStackPtr - 1].ParenLevel
      end
  end;

  { Internal function:  Perform all operations on the stack which are higher
    in precedence than the current operation in CurOp.  False is returned if
    an error occurs }
  function DoPreviousOps (Key : TCalcKey) : boolean;

  var
    StartPrec,
    StartParenLevel,
    CurPrec,
    PrevParenLevel,
    SP              : integer;
    CurOp           : TCalcKey;
    ConstModeDone   : boolean;

  begin
    UpdateCurrentValue(True);
    StartPrec := Precedence[OpPrecStyle, Key];
    StartParenLevel := FStack[FStackPtr].ParenLevel;

    Result := True;

    { Process all stack entries that qualify }
    while True do
      begin
        if FStackPtr > 0 then
          begin
            CurPrec := Precedence[OpPrecStyle, FStack[FStackPtr - 1].Operation];
            PrevParenLevel := FStack[FStackPtr - 1].ParenLevel
          end
        else
          begin
            CurPrec := UnusedPrecedence;
            PrevParenLevel := 0
          end;

        if (StartPrec > FinishParenPrec) and (StartParenLevel > PrevParenLevel)
        then
          { Don't go below the current parenthesis level for ordinary
            operations (ie, not for ckRParen, ckEquals, ckMPlus and ckMMinus) }
          break;

        if (StartPrec = FinishParenPrec) and (StartParenLevel > PrevParenLevel)
        then
          begin
            { Finish off a parenthesis level for ckRParen }
            Dec(FStack[FStackPtr].ParenLevel);
            break
          end;

        if FConstantMode and (StartPrec > FinishParenPrec) and
           (StartParenLevel = 0)
        then
          { Any ordinary operator at parenthesis level 0 turns off CONST mode }
          FConstantMode := False;

        if FStackPtr = 0 then
          begin
            { We have reached the bottom of the stack }
            if StartPrec = FinishAllPrec then
              { Finish off all calculations on ckEquals, ckMPlus, ckMMinus }
              FStack[FStackPtr].ParenLevel := 0;
            break
          end;

        if StartPrec > CurPrec then
          { If current operator's precedence is higher than that on stack,
            we have finished processing }
          break;

        ConstModeDone := (FConstantMode and (FStackPtr = 1));
        if ConstModeDone then
          begin
            { Constant mode: do NOT remove the last stack element }
            SP := 0;
            CurOp := FStack[SP].Operation;
            if StartPrec = FinishAllPrec then
              { Finish off all parenthesis levels on ckEquals, ckMPlus, ckMMinus }
              FStack[FStackPtr].ParenLevel := 0
          end
        else
          begin
            { Perform the operation, decrementing the stack pointer }
            SP := FStackPtr;
            Dec(FStackPtr);
            CurOp := FStack[FStackPtr].Operation
          end;

        { Perform the actual operation }
        with FStack[FStackPtr] do
          begin
            case CurOp of
              ckAdd      : Value := iAdd(Value, FStack[SP].Value);
              ckSubtract : Value := iSub(Value, FStack[SP].Value);
              ckMultiply : Value := iMul(Value, FStack[SP].Value);
              ckDivide   : if iSgn(FStack[SP].Value) <> 0 then
                             Value := iDiv(Value, FStack[SP].Value, Signed)
                           else
                             begin
                               FEntryState := esError;
                               Result := False;
                               break;
                             end;
              ckMod      : if iSgn(FStack[SP].Value) <> 0 then
                             Value := iMod(Value, FStack[SP].Value, Signed)
                           else
                             begin
                               FEntryState := esError;
                               Result := False;
                               break;
                             end;
              ckAnd      : Value := iAnd(Value, FStack[SP].Value);
              ckNand     : Value := iNand(Value, FStack[SP].Value);
              ckOr       : Value := iOr(Value, FStack[SP].Value);
              ckNor      : Value := iNor(Value, FStack[SP].Value);
              ckXor      : Value := iXor(Value, FStack[SP].Value);
              ckEqv      : Value := iEqv(Value, FStack[SP].Value);
              ckSHL      : Value := iShl(Value, FStack[SP].Value, Signed);
              ckSHR      : Value := iShr(Value, FStack[SP].Value, Signed);
              ckROL      : Value := iRolBit(Value, FStack[SP].Value,
                                            CalcSizeToBits[Size], Signed);
              ckROR      : Value := iRorBit(Value, FStack[SP].Value,
                                            CalcSizeToBits[Size], Signed)
            end;
            RoundVal(Value, Size, Signed);
            Operation := ckNone
          end;
        FValueState := vsNumberValid;

        if ConstModeDone then
          { Constant mode: the operation has been performed }
          break
      end
  end;


var
  CurPrec, OldPrec : integer;

begin { HandleKey }
  { The Clear key can be pressed at any time, even if the calculator is in
    an error state }
  if Key = ckClear then
    begin
      ClearOperations;
      Result := True;
      exit
    end;

  { No other operation can take place when the calculator is in an error mode }
  if InError then
    begin
      Result := False;
      exit
    end;

  { Handle one of the operation keys }
  case Key of
    ck0 .. ckF  : Result := AppendDigit(Ord(Key));
    ckBackspace : Result := Backspace;

    ckMIn       : Result := StoreInMem;
    ckMR        : Result := RetrieveMem;
    ckSwapXY    : Result := SwapXY;
    ckSwapXM    : Result := SwapXM;

    ckNot,
    ckNeg       : Result := DoUnaryOp(Key);

    ckLParen    : Result := DoLParen;

    ckNone      : Result := True
  else
    { Test if calculator should enter CONST mode }
    Result := True;
    CurPrec := Precedence[OpPrecStyle, Key];
    OldPrec := Precedence[OpPrecStyle, FLastKey];

    if (CurPrec > FinishParenPrec) and (OldPrec > FinishParenPrec) then
      begin
        if Key = FLastKey then
          begin
            if (FStackPtr = 1) and (FStack[FStackPtr].ParenLevel = 0) then
              { Enter or leave CONST mode }
              FConstantMode := not FConstantMode
            else
              { CONST mode cannot be entered }
              Result := False
          end
        else if FLastKey <> ckLParen then
          begin
            { Last key (not a left parenthesis) was different from current key }
            if CurPrec >= OldPrec then
              begin
                { Since the new operation has higher precedence, simply replace
                  old operation with the new one }
                FStack[FStackPtr - 1].Operation := Key;
                FConstantMode := False
              end
            else
              begin
                { The new operation has LOWER precedence; hence, need to
                  perform previous operations.  Undo the stack from the
                  last time }
                Dec(FStackPtr);
                FStack[FStackPtr].Operation := ckNone;
                FConstantMode := False;
                Result := DoPreviousOps(Key);
                if Result then
                  PushStack(Key)
              end
          end
      end
    else
      begin
        { The last key was NOT an CONST-able operation key }
        Result := DoPreviousOps(Key);
        if Result then
          begin
            if Precedence[OpPrecStyle, Key] > FinishParenPrec then
              PushStack(Key)
            else if Key = ckMPlus then
              AddToMemory
            else if Key = ckMMinus then
              SubtractFromMemory
          end
      end
  end;

  FLastKey := Key
end;


{***********************************************************************}
{ Handle a digit key '0' to 'F' (passed as an integer 0 to 15).  True is
  returned if the digit was successfully appended to the current value;
  False is returned if it was not, or if Digit was invalid }
function TCalculator.AppendDigit (const Digit : integer) : boolean;

var
  MaxL       : integer;
  FirstDigit : char;

begin
  if InError or (Digit < 0) or (Digit >= CalcModeToInt[Mode]) then
    begin
      Result := False;
      exit
    end;

  if FEntryState = esFirstKey then
    begin
      { First digit in the string.  Note that leading zeros are allowed }
      FEntryState := esNextKey;
      FString := DigitChar[Digit]
    end
  else { FEntryState = esNextKey }
    begin
      { NB: String representation will always be OK when FEntryState =
        esNextKey.  This is because any function that affects the value
        automatically sets FEntryState to esFirstKey }
      MaxL := MaxStrLen[Mode, Size];
      if Length(FString) = MaxL then
        begin
          { The string representation is at the maximum length }
          if FString[1] = '0' then
            { Delete a leading zero if possible }
            Delete(FString, 1, 1)
          else
            begin
              Result := False;
              exit
            end
        end;

      if Length(FString) = MaxL - 1 then
        begin
          { The string is one less than the maximum length.  Check to see if
            digit can be appended without causing an integer overflow.  This
            only needs to be done for decimal and octal modes }
          if Mode = cmDecimal then
            begin
              if FString + DigitChar[Digit] > MaxDecValues[Size] then
                begin
                  Result := False;
                  exit
                end
            end
          else if Mode = cmOctal then
            begin
              FirstDigit := FString[1];
              if ((Size in [cs8Bit, cs32Bit, cs128Bit]) and (FirstDigit > '3')) or
                 ((Size in [cs16Bit, cs64Bit]) and (FirstDigit > '1')) then
                begin
                  Result := False;
                  exit
                end
            end
        end;

      FString := FString + DigitChar[Digit]
    end;

  FValueState := vsStringValid;
  FLastKey := TCalcKey(Digit);
  Result := True
end;

end.

