unit BCFonts;

{************************************************************************
*                                                                       *
*                 Base Calculator Font File Definitions                 *
*                Copyright (C) 1993-2000, John Zaitseff                 *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      1st April, 2000.
  Version:   2.1

  This file provides the font file definitions for the Base Calculator
  application.  By default, Base Calculator comes supplied with six fonts
  for its display:  LED, OCR-A, Sans Serif, Courier, Times Roman and
  Handwritten characters.  These are stored as resources in the executable.

  Note that Borland Delphi 4 or later is required to compile this unit.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.


  FONT FILES
  ==========

  The following is a description of the font files.  Each font set
  consists of four monochrome (black and white) bitmaps, labelled Small,
  Medium, Large and Error (the Error bitmap is a special case, and will be
  discussed separately).  Each of the three font bitmaps, Small, Medium
  and Large, is divided into two rows. The top row of each bitmap contains
  character outlines to be displayed in bright colour.  The bottom row
  contains optional character outlines to be displayed in dim colour. The
  idea behind this is especially relevant to 7-segment LED character
  outlines:  the top row represents segments that are on, while the bottom
  row represents segments that are off (dim).

  Each row is divided into 18 full-width characters and 4 half-width
  characters.  The full-width characters are (from left to right):

      0 1 2 3 4 5 6 7 8 9 A B C D E F - (space)

  The half-width characters are:

      (comma) (full-stop) (space) (space)

  The half-width characters appear directly next to the full-width ones.

  The widths and heights of the characters are directly taken from each of
  the font bitmaps.  If a font bitmap is 200x32 pixels, for example, then
  each full-width character has a size of 10x16 and each half-width
  character has a size of 5x16 pixels.

  For best results, the following are the recommended bitmap sizes:

      Small  - 200x32    (full-width character 10x16, half-width 5x16)
      Medium - 400x64    (full-width character 20x32, half-width 10x32)
      Large  - 800x128   (full-width character 40x64, half-width 20x64)

  Each of the font bitmaps is treated as a mask:  black pixels will be
  treated as transparent, white pixels as the required colour.  When
  displaying a character, the dim colour (bottom row character) will be
  drawn first (if this option is selected by the user of the program),
  followed by the bright colour (top row character).  The only exceptions
  to this are the half-width characters:  if one is to be displayed as on,
  only the top row is drawn in bright colour; conversely, if one is to be
  displayed as off, only the bottom row is drawn in dim colour.

  The Error bitmap is a special case:  it contains the word "Error", and
  is displayed whenever an error occurs in calculation.  As with the other
  three bitmaps, the Error bitmap has both a top row and a bottom row,
  drawn in the same way described above.  The height of the Error bitmap
  is to be the same as the height of the Large bitmap (recommended: 128
  pixels).  The width is to be the equivalent of 9.5 full-width Large
  characters (suggested: 380 pixels).

  Each font has an associated resource script (RC file) which must be
  compiled into a RES file.  Each font has two resource strings
  associated with it.  The first is the (localised) name of the font,
  used to display the font name to the user.  The second is the internal
  name of the font, used in storing the font name to the registry.  Each
  font bitmap appears as a BITMAP resource, in the order Small, Medium,
  Large and Error.  The starting resource number of the font bitmaps is
  the same as the first string resource number; the second string resource
  immediately follows the first.

  Whenever a new font is added, the table of fonts contained in this unit
  (in BMNumFonts, the number of fonts present, and BMFontResID, the font
  number to resource ID mapping) must be modified and its resource added
  in the implementation section.  These sections are flagged with MODIFY:

  In addition to the four bitmaps for each font set, Base Calculator makes
  use of another font bitmap, called Status.  This special-case bitmap is
  used to draw the status indicators above the numerical display.  This
  bitmap is divided into 21 rows and one column (except for the first
  row). The first row is again a special case, and consists of the digits
  0 to 9, in that order.  Each row in this bitmap has the same height and
  width (suggested: a height of 17 pixels, and a width of 80 pixels).  The
  second to 21st rows consist of one word each, in order:  MEM, PAREN,
  CONST, ADD, SUB, MUL, DIV, MOD, NOT, NEG, AND, NAND, OR, NOR, XOR, EQV,
  SHL, SHR, ROL, ROR.  MEM indicates that the calculator's memory is being
  used. PAREN indicates the parenthesis nesting level; the digits in the
  first row are used for this purpose.  CONST indicates a constant
  calculator operation, one of ADD to ROR.

  The Status bitmap is essentially font-set-independent; it does not need
  to change when changing the Small, Medium, Large and Error font bitmaps.
  For this reason, this font bitmap has its own resource ID.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


type
  TResID = type integer;


const
  { Font offsets from BCFonts.FontResID[] }
  SmallFontOfs    = 0;
  MediumFontOfs   = 1;
  LargeFontOfs    = 2;
  ErrorFontOfs    = 3;


type
  { Base Calculator fonts object }
  TBCFonts = class
    private
      FNumFonts               : integer;

      FFontDisplayName        : array of string;
      FFontInternalName       : array of string;

      FStatusFontResID        : TResID;
      FStatusFontDisplayName  : string;
      FStatusFontInternalName : string;

      function GetFontResID (N : integer) : TResID;
      function GetFontDisplayName (N : integer) : string;
      function GetFontInternalName (N : integer) : string;

    public
      constructor Create;

      property NumFonts                        : integer read FNumFonts;

      property FontResID [N : integer]         : TResID  read GetFontResID;
      property FontDisplayName [N : integer]   : string  read GetFontDisplayName;
      property FontInternalName [N : integer]  : string  read GetFontInternalName;

      property StatusFontResID                 : TResID  read FStatusFontResID;
      property StatusFontDisplayName           : string  read FStatusFontDisplayName;
      property StatusFontInternalName          : string  read FStatusFontInternalName;

      function FontResIDToFontNumber (Value : TResID) : integer;
      function FontResIDToInternalName (Value : TResID) : string;
      function InternalNameToFontResID (const Value : string) : TResID;
    end;


var
  BCalcFonts : TBCFonts;


implementation


uses
  SysUtils, StringRes;


{***********************************************************************}
{ MODIFY:  Number of bitmapped fonts.  This needs to be modified if
           additional fonts are added }
const
  BMNumFonts  = 6;


{***********************************************************************}
{ MODIFY:  Mapping of font numbers (1..BMNumFonts) to font resource IDs.
           The order of resource IDs in this table determines the order
           that fonts are listed in the Options dialog box in Base
           Calculator.  This table needs to be modified if additional
           fonts are added }
const
  BMFontResID : array [1..BMNumFonts] of TResID =
                (1000, 1004, 1008, 1012, 1016, 1020);


{***********************************************************************}
{ Miscellaneous constants }

const
  { Status font bitmap resource ID }
  BMStatusFontResID   = 1100;

  { String resource offsets from BCFonts.FontResID }
  BMDisplayNameOfs    = 0;
  BMInternalNameOfs   = 1;


{***********************************************************************}
{ MODIFY: Font resource definitions.  If any fonts are added, the
          associated RES file needs to be included in this list }

{$R LED.res}
{$R OCR.res}
{$R SSf.res}
{$R Cou.res}
{$R Rom.res}
{$R Hnd.res}

{$R Status.res}


{***********************************************************************}
{ Construct the Base Calculator fonts object, reading in the display and
  internal names of the fonts }
constructor TBCFonts.Create;

var
  I : integer;

begin
  inherited Create;

  FNumFonts := BMNumFonts;

  SetLength(FFontDisplayName, BMNumFonts);
  SetLength(FFontInternalName, BMNumFonts);

  for I := 1 to BMNumFonts do
    begin
      FFontDisplayName[I - 1] := LoadStr(BMFontResID[I] + BMDisplayNameOfs);
      FFontInternalName[I - 1] := LoadStr(BMFontResID[I] + BMInternalNameOfs)
    end;

  FStatusFontResID := BMStatusFontResID;
  FStatusFontDisplayName := LoadStr(BMStatusFontResID + BMDisplayNameOfs);
  FStatusFontInternalName := LoadStr(BMStatusFontResID + BMInternalNameOfs)
end;


{***********************************************************************}
{ Return the resource ID of a font number (1..NumFonts).  If the parameter
  is out of range, an ERangeError exception is raised }
function TBCFonts.GetFontResID (N : integer) : TResID;

begin
  if (N < 1) or (N > NumFonts) then
    raise ERangeError.CreateFmt(SFontNumRange, [N, NumFonts]);

  Result := BMFontResID[N]
end;


{***********************************************************************}
{ Return the display name of a font number (1..NumFonts).  If the parameter
  is out of range, an ERangeError exception is raised }
function TBCFonts.GetFontDisplayName (N : integer) : string;

begin
  if (N < 1) or (N > NumFonts) then
    raise ERangeError.CreateFmt(SFontNumRange, [N, NumFonts]);

  Result := FFontDisplayName[N - 1]
end;


{***********************************************************************}
{ Return the internal name of a font number (1..NumFonts).  If the
  parameter is out of range, an ERangeError exception is raised }
function TBCFonts.GetFontInternalName (N : integer) : string;

begin
  if (N < 1) or (N > NumFonts) then
    raise ERangeError.CreateFmt(SFontNumRange, [N, NumFonts]);

  Result := FFontInternalName[N - 1]
end;


{***********************************************************************}
{ Convert a font resource ID to the position it occupies in TBCFonts.
  If an illegal font resource ID is supplied, zero is returned }
function TBCFonts.FontResIDToFontNumber (Value : TResID) : integer;

var
  I : integer;

begin
  for I := Low(BMFontResID) to High(BMFontResID) do
    if Value = BMFontResID[I] then
      begin
        Result := I;
        exit
      end;

  Result := 0
end;


{***********************************************************************}
{ Convert a font resource ID to its equivalent internal name.  If an
  illegal font resource ID is supplied, a blank string is returned }
function TBCFonts.FontResIDToInternalName (Value : TResID) : string;

var
  I : integer;

begin
  for I := Low(BMFontResID) to High(BMFontResID) do
    if Value = BMFontResID[I] then
      begin
        Result := FontInternalName[I];
        exit
      end;

  if Value = StatusFontResID then
    Result := StatusFontInternalName
  else
    Result := ''
end;


{***********************************************************************}
{ Convert an internal font name to its equivalent font resource ID.  If
  an illegal font name is supplied, 0 is returned }
function TBCFonts.InternalNameToFontResID (const Value : string) : TResID;

var
  I : integer;

begin
  for I := Low(BMFontResID) to High(BMFontResID) do
    if CompareText(Value, FontInternalName[I]) = 0 then
      begin
        Result := BMFontResID[I];
        exit
      end;

  if CompareText(Value, StatusFontInternalName) = 0 then
    Result := StatusFontResID
  else
    Result := 0
end;


initialization
  BCalcFonts := TBCFonts.Create
finalization
  BCalcFonts.Free
end.

