unit OptionsFrm;

{************************************************************************
*                                                                       *
*            Base Calculator Configuration Options Dialog Box           *
*                Copyright (C) 1993-2000, John Zaitseff                 *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      1st April, 2000.
  Version:   2.1

  This file provides the user interface code to create and manipulate the
  Configuration Options dialog box.

  It would have been great if Borland Delphi and Windows automatically
  handled requests for help for the sample display (using the Help button
  in the dialog box title bar).  Unfortunately, this is not the case.
  This unit contains code to simulate such handling, capturing a number
  of Windows messages and processing LButtonDown events for the sample
  display.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, ComCtrls, StdCtrls, ExtCtrls, Spin,
  BCDefs, BCDisp, BCConfig;


type
  { Base Calculator Options dialog box }
  TOptionsWin = class(TForm)
      OptionsPageCtl        : TPageControl;
      BtnOK                 : TButton;
      BtnCancel             : TButton;
      ColorDialog           : TColorDialog;
      DisplayTS             : TTabSheet;
      FontDiv               : TLabel;
      FontBev               : TBevel;
      FontLbl               : TLabel;
      FontCmB               : TComboBox;
      ColoursDiv            : TLabel;
      ColoursBev            : TBevel;
      ColorBtn              : TButton;
      FontOnColorBtn        : TButton;
      FontOffColorBtn       : TButton;
      UseFontOffCB          : TCheckBox;
      UsePaletteCB          : TCheckBox;
      MiscDiv               : TLabel;
      MiscBev               : TBevel;
      LeadingZerosCB        : TCheckBox;
      UseSeparatorCB        : TCheckBox;
      SeparatorLbl          : TLabel;
      SeparatorCmB          : TComboBox;
      SampleDiv             : TLabel;
      SampleBev             : TBevel;
      SamplePanel           : TPanel;
      SampleDisplay         : TBCalcSampleDisplay;
      OtherTS               : TTabSheet;
      ParenBev              : TBevel;
      ParenDiv              : TLabel;
      UseDefMaxParenLevelCB : TCheckBox;
      MaxParenLevelLbl      : TLabel;
      MaxParenLevelLbl2     : TLabel;
      MaxParenLevelPanel    : TPanel;
      MaxParenLevelEB       : TEdit;
      MaxParenLevelUD       : TUpDown;
      PrecDiv               : TLabel;
      PrecBev               : TBevel;
      OpPrecLbl             : TLabel;
      OpPrecCmB             : TComboBox;
      Misc2Div              : TLabel;
      Misc2Bev              : TBevel;
      EnhancedPasteCB       : TCheckBox;
      SaveWindowPosCB       : TCheckBox;

      procedure FormCreate (Sender : TObject);
      procedure FormShow (Sender : TObject);

      procedure BtnOKClick (Sender : TObject);
      procedure SampleDisplayClick (Sender : TObject);

      procedure FontCmBChange (Sender : TObject);
      procedure ColorBtnClick (Sender : TObject);
      procedure FontOnColorBtnClick (Sender : TObject);
      procedure FontOffColorBtnClick (Sender : TObject);
      procedure UseFontOffCBClick (Sender : TObject);
      procedure UsePaletteCBClick (Sender : TObject);
      procedure LeadingZerosCBClick (Sender : TObject);
      procedure UseSeparatorCBClick (Sender : TObject);
      procedure SeparatorCmBChange (Sender : TObject);
      procedure UseDefMaxParenLevelCBClick (Sender : TObject);

    private
      FConfig         : TBCConfig;
      FSeparator      : TSeparator;
      FColor          : TColor;
      FFontOnColor    : TColor;
      FFontOffColor   : TColor;

      FSysCommandHelp : boolean;  { True if a request for help has been started }

      procedure WMDisplayChange (var Message : TWMDisplayChange); message WM_DISPLAYCHANGE;

      procedure WMSysCommand (var Message : TWMSysCommand); message WM_SYSCOMMAND;
      procedure WMHelp (var Message : TWMHelp); message WM_HELP;

    public
      property Config : TBCConfig read FConfig write FConfig;
    end;


var
  OptionsWin : TOptionsWin;


implementation


uses
  BCFonts, Calc;


{$R *.DFM}


{***********************************************************************}
{ Create the Options dialog box, initialising the font name strings }
procedure TOptionsWin.FormCreate (Sender : TObject);

var
  I : integer;

begin
  with FontCmB do
    begin
      Items.BeginUpdate;
      Items.Clear;
      for I := 1 to BCalcFonts.NumFonts do
        Items.Add(BCalcFonts.FontDisplayName[I]);
      Items.EndUpdate;
      ItemIndex := 0
    end;

  MaxParenLevelUD.Max := AbsMaxParenLevel
end;


{***********************************************************************}
{ Show the Options dialog box, making it centred above its parent form.
  This procedure also initialises all of the controls to the required
  values }
procedure TOptionsWin.FormShow (Sender : TObject);

var
  ParentForm  : TForm;
  FontNum     : integer;
  BM          : TPicture;
  AllowSepCmB : boolean;
  AllowMaxPL  : boolean;

begin
  { To whom does this form belong? }
  if Owner is TApplication then
    ParentForm := TApplication(Owner).MainForm
  else if Owner is TForm then
    ParentForm := TForm(Owner)
  else
    ParentForm := nil;

  { Centre the form above its parent }
  if Assigned(ParentForm) then
    begin
      Left := (ParentForm.Left + ParentForm.Width div 2) - Width div 2;
      Top := (ParentForm.Top + ParentForm.Height div 2) - Height div 2
    end
  else
    Position := poScreenCenter;

  { Set the control values to those stored in FConfig }
  if Assigned(Config) then
    with Config do
      begin
        FontNum := BCalcFonts.FontResIDToFontNumber(FontResID);
        if FontNum <> 0 then
          FontCmB.ItemIndex := FontNum - 1;

        UseFontOffCB.Checked := UseFontOff;
        SampleDisplay.UseFontOff := UseFontOff;

        UsePaletteCB.Enabled := SampleDisplay.CanUsePalette;
        UsePaletteCB.Checked := UsePalette;
        SampleDisplay.UsePalette := UsePalette;

        LeadingZerosCB.Checked := LeadingZeros;
        SampleDisplay.LeadingZeros := LeadingZeros;

        { FSeparator contains the separator, except when UseSeparatorCB.Checked
          is False, in which case it contains the last non-spNone selected
          separator }
        AllowSepCmB := (Separator <> spNone);
        if AllowSepCmB then
          begin
            FSeparator := Separator;
            SeparatorCmB.Color := clWindow;
            SampleDisplay.Separator := Separator
          end
        else
          begin
            FSeparator := spComma;
            SeparatorCmB.Color := clBtnFace;
            SampleDisplay.Separator := spNone
          end;
        UseSeparatorCB.Checked := AllowSepCmB;
        SeparatorCmB.Enabled := AllowSepCmB;
        SeparatorCmB.ItemIndex := Ord(FSeparator) - 1;

        FColor := DispColor;
        FFontOnColor := DispFontOnColor;
        FFontOffColor := DispFontOffColor;

        with SampleDisplay do
          begin
            Color := FColor;
            FontOnColor := FFontOnColor;
            FontOffColor := FFontOffColor;
            Value := SampleValue
          end;

        BM := TPicture.Create;
        try
          BM.Bitmap.LoadFromResourceID(HInstance, FontResID + MediumFontOfs);
          SampleDisplay.MediumFont := BM
        finally
          BM.Free
        end;

        MaxParenLevelEB.Text := '0';
        MaxParenLevelUD.Position := -1;
        MaxParenLevelUD.Position := MaxParenLevel;
        AllowMaxPL := not (MaxParenLevel = DefMaxParenLevel);
        UseDefMaxParenLevelCB.Checked := not AllowMaxPL;
        MaxParenLevelEB.Enabled := AllowMaxPL;
        MaxParenLevelUD.Enabled := AllowMaxPL;
        if AllowMaxPL then
          MaxParenLevelPanel.Color := clWindow
        else
          MaxParenLevelPanel.Color := clBtnFace;

        OpPrecCmB.ItemIndex := Ord(OpPrecStyle);
        EnhancedPasteCB.Checked := EnhancedPaste;
        SaveWindowPosCB.Checked := SaveWindowPos
      end
end;


{***********************************************************************}
{ Save all of the configuration options to the Config object, which
  will save any changed settings to the system registry }
procedure TOptionsWin.BtnOKClick (Sender : TObject);

var
  FontNum : integer;

begin
  if Assigned(Config) then
    with Config do
      begin
        FontNum := FontCmB.ItemIndex + 1;
        if FontNum <> 0 then
          FontResID := BCalcFonts.FontResID[FontNum];

        DispColor := FColor;
        DispFontOnColor := FFontOnColor;
        DispFontOffColor := FFontOffColor;

        LeadingZeros := LeadingZerosCB.Checked;
        if UseSeparatorCB.Checked then
          Separator := FSeparator
        else
          Separator := spNone;
        UseFontOff := UseFontOffCB.Checked;
        UsePalette := UsePaletteCB.Checked;

        MaxParenLevel := MaxParenLevelUD.Position;
        OpPrecStyle := TOpPrecStyle(OpPrecCmB.ItemIndex);
        EnhancedPaste := EnhancedPasteCB.Checked;
        SaveWindowPos := SaveWindowPosCB.Checked
      end
end;


{***********************************************************************}
{ The user selected an option from the font names combobox }
procedure TOptionsWin.FontCmBChange (Sender : TObject);

var
  FontNum,
  FontResID : integer;
  BM        : TPicture;

begin
  FontNum := FontCmB.ItemIndex + 1;
  if FontNum = 0 then
    exit;

  FontResID := BCalcFonts.FontResID[FontNum];
  SampleDisplay.UseFontOff := UseFontOffCB.Checked;

  BM := TPicture.Create;
  try
    BM.Bitmap.LoadFromResourceID(HInstance, FontResID + MediumFontOfs);
    SampleDisplay.MediumFont := BM
  finally
    BM.Free
  end;
end;


{***********************************************************************}
{ Select a colour and propagate it to the sample display, if the user
  pressed OK }
procedure TOptionsWin.ColorBtnClick (Sender : TObject);

begin
  with ColorDialog do
    begin
      Color := FColor;
      if Execute then
        begin
          FColor := Color;
          SampleDisplay.Color := Color
        end
    end
end;


{***********************************************************************}
{ Select a colour and propagate it to the sample display as the Font On
  colour, if the user pressed OK }
procedure TOptionsWin.FontOnColorBtnClick (Sender : TObject);

begin
  with ColorDialog do
    begin
      Color := FFontOnColor;
      if Execute then
        begin
          FFontOnColor := Color;
          SampleDisplay.FontOnColor := Color
        end
    end
end;


{***********************************************************************}
{ Select a colour and propagate it to the sample display as the Font Off
  colour, if the user pressed OK }
procedure TOptionsWin.FontOffColorBtnClick (Sender : TObject);

begin
  with ColorDialog do
    begin
      Color := FFontOffColor;
      if Execute then
        begin
          FFontOffColor := Color;
          SampleDisplay.FontOffColor := Color
        end
    end
end;


{***********************************************************************}
{ Propagate the UseFontOffColorCB setting to the sample display }
procedure TOptionsWin.UseFontOffCBClick (Sender : TObject);

begin
  SampleDisplay.UseFontOff := UseFontOffCB.Checked
end;


{***********************************************************************}
{ Propagate the UsePalette setting to the sample display }
procedure TOptionsWin.UsePaletteCBClick (Sender : TObject);

begin
  SampleDisplay.UsePalette := UsePaletteCB.Checked
end;


{***********************************************************************}
{ Propagate the LeadingZeros setting to the sample display }
procedure TOptionsWin.LeadingZerosCBClick (Sender : TObject);

begin
  SampleDisplay.LeadingZeros := LeadingZerosCB.Checked
end;


{***********************************************************************}
{ Propagate the Separator setting to the sample display }
procedure TOptionsWin.UseSeparatorCBClick (Sender : TObject);

begin
  if UseSeparatorCB.Checked then
    begin
      SampleDisplay.Separator := FSeparator;
      with SeparatorCmB do
        begin
          Enabled := True;
          ItemIndex := Ord(FSeparator) - 1;
          Color := clWindow
        end
    end
  else
    begin
      { UseSeparator is False }
      SampleDisplay.Separator := spNone;
      with SeparatorCmB do
        begin
          Enabled := False;
          ItemIndex := Ord(FSeparator) - 1;
          Color := clBtnFace
        end
    end
end;


{***********************************************************************}
{ Propagate the Separator setting to the sample display }
procedure TOptionsWin.SeparatorCmBChange (Sender : TObject);

begin
  if UseSeparatorCB.Checked then
    begin
      FSeparator := TSeparator(SeparatorCmB.ItemIndex + 1);
      SampleDisplay.Separator := FSeparator
    end
end;


{***********************************************************************}
{ Propagate the value of UseDefMaxParenLevelCB.Checked to other controls }
procedure TOptionsWin.UseDefMaxParenLevelCBClick (Sender : TObject);

var
  AllowMaxPL : boolean;
  OldMaxPL   : integer;

begin
  AllowMaxPL := not UseDefMaxParenLevelCB.Checked;

  OldMaxPL := MaxParenLevelUD.Position;
  MaxParenLevelEB.Text := '';
  MaxParenLevelUD.Position := -1;

  MaxParenLevelEB.Enabled := AllowMaxPL;
  MaxParenLevelUD.Enabled := AllowMaxPL;

  if AllowMaxPL then
    begin
      MaxParenLevelUD.Position := OldMaxPL;
      MaxParenLevelPanel.Color := clWindow
    end
  else
    begin
      MaxParenLevelUD.Position := DefMaxParenLevel;
      MaxParenLevelPanel.Color := clBtnFace
    end
end;


{***********************************************************************}
{ The screen display has changed; this may affect whether UsePalette
  should be enabled or not }
procedure TOptionsWin.WMDisplayChange (var Message : TWMDisplayChange);

begin
  with SampleDisplay do
    begin
      DisplayChanged;
      UsePaletteCB.Enabled := CanUsePalette;
      Repaint
    end
end;


{***********************************************************************}
{ Display context-sensitive help for the sample display.  Unfortunately,
  this display does NOT have a HelpContext property, nor does it (or its
  parent, SamplePanel) respond to the help cursor click.  Hence, this unit
  captures the state of the help cursor (via the SC_CONTEXTHELP in the
  WM_SYSCOMMAND message) and simulates the WM_HELP that would have
  resulted }
procedure TOptionsWin.SampleDisplayClick (Sender : TObject);

var
  HelpID : integer;

begin
  if FSysCommandHelp then
    begin
      HelpID := SamplePanel.HelpContext;
      if HelpID <> 0 then
        Application.HelpCommand(HELP_CONTEXTPOPUP, HelpID);
      FSysCommandHelp := False
    end
end;


{***********************************************************************}
{ If the WM_SYSCOMMAND command type is SC_CONTEXTHELP, the user has
  initiated a request for help using the Help button in the dialog
  box's title bar.  Save this for a possible left mouse button click
  in the sample display }
procedure TOptionsWin.WMSysCommand (var Message : TWMSysCommand);

begin
  if (Message.CmdType and $FFF0) = SC_CONTEXTHELP then
    FSysCommandHelp := True;
  inherited
end;


{***********************************************************************}
{ A previous WM_SYSCOMMAND / SC_CONTEXTHELP generated a WM_HELP message.
  Hence any left mouse button clicks after this will NOT be a request
  for help }
procedure TOptionsWin.WMHelp (var Message : TWMHelp);

begin
  FSysCommandHelp := False;
  inherited
end;

end.

