unit BCDisp;

{************************************************************************
*                                                                       *
*                      Base Calculator LED Display                      *
*                Copyright (C) 1993-2000, John Zaitseff                 *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      1st April, 2000.
  Version:   2.1

  This file provides the Base Calculator LED Display custom component,
  TBCalcDisplay, as well as a stripped-down version of this component,
  TBCalcSampleDisplay, for the Options dialog box.  These custom
  components are specific to Base Calculator, and were written in this
  way to simplify the main program.  TBCalcSampleDisplay differs from
  TBCalcDisplay in that only one line is displayed in the Large font,
  the Status font is not used, and the mode is always Hexadecimal (ie,
  all digits 0-F are allowed and separator spacing is set at 4 digits).

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, BCDefs;


type
  TDisplayMode  = (dmNormal, dmBlank, dmError);
  TSeparator    = (spNone, spSpace, spComma, spFullStop);


type
  { "Abstract" Base Calculator LED Display, from which the real and sample
    displays are derived }
  TCustomBCalcDisplay = class(TGraphicControl)
    private
      FBuffer         : TBitmap;      { Off-screen display buffer }
      FBufferValid    : boolean;      { True if FBuffer contents are valid }
      FDrawing        : boolean;      { True if currently redrawing }

      FCanUsePalette  : boolean;      { True if using a palettable device }
      FPaletteUsed    : boolean;      { True if palette is being used }
      FPalette        : HPalette;     { The actual palette }

      FRColor         : TColor;       { Real background colour }
      FRFontOnColor   : TColor;       { Real font on colour }
      FRFontOffColor  : TColor;       { Real font off colour }

      FValue          : string;       { Original string value }

      FCalcMode       : TCalcMode;
      FCalcSize       : TCalcSize;
      FDisplayMode    : TDisplayMode;
      FSeparator      : TSeparator;
      FLeadingZeros   : boolean;
      FColor          : TColor;
      FFontOnColor    : TColor;
      FFontOffColor   : TColor;
      FUseFontOff     : boolean;
      FUsePalette     : boolean;

      FSmallFont      : TPicture;
      FMediumFont     : TPicture;
      FLargeFont      : TPicture;
      FErrorFont      : TPicture;
      FStatusFont     : TPicture;

      FMemInUse       : boolean;
      FParenLevel     : cardinal;
      FCurrentCalcOp  : TCalcOp;
      FConstantMode   : boolean;

      FSampleDisplay  : boolean;      { True if sample display, False if real }

      procedure SetValue (AValue : string);
      procedure SetCalcMode (Value : TCalcMode);
      procedure SetCalcSize (Value : TCalcSize);
      procedure SetDisplayMode (Value : TDisplayMode);
      procedure SetSeparator (Value : TSeparator);
      procedure SetLeadingZeros (Value : boolean);
      procedure SetColor (Value : TColor);
      procedure SetFontOnColor (Value : TColor);
      procedure SetFontOffColor (Value : TColor);
      procedure SetUseFontOff (Value : boolean);
      procedure SetUsePalette (Value : boolean);
      procedure SetSmallFont (Value : TPicture);
      procedure SetMediumFont (Value : TPicture);
      procedure SetLargeFont (Value : TPicture);
      procedure SetErrorFont (Value : TPicture);
      procedure SetStatusFont (Value : TPicture);
      procedure SetMemInUse (Value : boolean);
      procedure SetParenLevel (Value : cardinal);
      procedure SetCurrentCalcOp (Value : TCalcOp);
      procedure SetConstantMode (Value : boolean);

      procedure SetSampleDisplay (Value : boolean);
      procedure CheckBufferSize;
      function GetCanvas : TCanvas;

    protected
      procedure Paint; override;
      procedure BufferChanged (Sender : TObject);
      function GetPalette : HPalette; override;

      procedure CreatePalette;
      procedure DestroyPalette;
      procedure ChangePalette (Value : TColor; Index : cardinal; var RealColor : TColor);

      property SampleDisplay  : boolean      read FSampleDisplay write SetSampleDisplay stored False;
      property Canvas         : TCanvas      read GetCanvas      stored False;
      property Palette        : HPalette     read FPalette       stored False;

  { public }
      property CanUsePalette  : boolean      read FCanUsePalette stored False;
      property PaletteUsed    : boolean      read FPaletteUsed   stored False;

  { published }
      property Value          : string       read FValue         write SetValue;

      property CalcMode       : TCalcMode    read FCalcMode      write SetCalcMode      default cmDecimal;
      property CalcSize       : TCalcSize    read FCalcSize      write SetCalcSize      default cs32Bit;
      property DisplayMode    : TDisplayMode read FDisplayMode   write SetDisplayMode   default dmNormal;
      property Separator      : TSeparator   read FSeparator     write SetSeparator     default spComma;
      property LeadingZeros   : boolean      read FLeadingZeros  write SetLeadingZeros  default False;

      property Color          : TColor       read FColor         write SetColor         default clBlack;
      property FontOnColor    : TColor       read FFontOnColor   write SetFontOnColor   default $0000FF;
      property FontOffColor   : TColor       read FFontOffColor  write SetFontOffColor  default $000040;
      property UseFontOff     : boolean      read FUseFontOff    write SetUseFontOff    default True;
      property UsePalette     : boolean      read FUsePalette    write SetUsePalette    default True;

      property SmallFont      : TPicture     read FSmallFont     write SetSmallFont;
      property MediumFont     : TPicture     read FmediumFont    write SetMediumFont;
      property LargeFont      : TPicture     read FLargeFont     write SetLargeFont;
      property ErrorFont      : TPicture     read FErrorFont     write SetErrorFont;
      property StatusFont     : TPicture     read FStatusFont    write SetStatusFont;

      property MemInUse       : boolean      read FMemInUse      write SetMemInUse      default False;
      property ParenLevel     : cardinal     read FParenLevel    write SetParenLevel    default 0;
      property CurrentCalcOp  : TCalcOp      read FCurrentCalcOp write SetCurrentCalcOp default coNone;
      property ConstantMode   : boolean      read FConstantMode  write SetConstantMode  default False;

    public
      constructor Create (AOwner : TComponent); override;
      destructor Destroy; override;

      procedure DisplayChanged;
    end;


type
  { Base Calculator main (real) LED display }
  TBCalcDisplay = class(TCustomBCalcDisplay)
    public
      constructor Create (AOwner : TComponent); override;

      property CanUsePalette;
      property PaletteUsed;

    published
      property Value;
      property CalcMode;
      property CalcSize;
      property DisplayMode;
      property Separator;
      property LeadingZeros;

      property Color;
      property FontOnColor;
      property FontOffColor;
      property UseFontOff;
      property UsePalette;

      property SmallFont;
      property MediumFont;
      property LargeFont;
      property ErrorFont;
      property StatusFont;

      property MemInUse;
      property ParenLevel;
      property CurrentCalcOp;
      property ConstantMode;

      { Publish properties inherited from TGraphicControl }
      property Height default 81;
      property Width  default 381;

      property Align;
      property Anchors;
      property Constraints;
      property DragCursor;
      property DragKind;
      property DragMode;
      property Enabled;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDock;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDock;
      property OnStartDrag;
    end;


type
  { Base Calculator sample LED display }
  TBCalcSampleDisplay = class(TCustomBCalcDisplay)
    protected
      property CalcMode nodefault;
      property CalcSize nodefault;

    public
      constructor Create (AOwner : TComponent); override;

      property CanUsePalette;
      property PaletteUsed;

    published
      property Value;
      property Separator;
      property LeadingZeros;

      property Color;
      property FontOnColor;
      property FontOffColor;
      property UseFontOff;
      property UsePalette;

      property MediumFont;

      { Publish properties inherited from TGraphicControl }
      property Height nodefault;
      property Width  nodefault;

      property Align;
      property Anchors;
      property Constraints;
      property DragCursor;
      property DragKind;
      property DragMode;
      property Enabled;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDock;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDock;
      property OnStartDrag;
    end;


procedure Register;
  { Register TBCalcDisplay and TBCalcSampleDisplay with the Borland Delphi IDE }


implementation


const
  { Colour palette indices }
  ppBlack        = 0;                  { Black colour palette index }
  ppWhite        = 1;                  { White colour palette index }
  ppColor        = 2;                  { Background colour palette index }
  ppFontOnColor  = 3;                  { Font "on" colour palette index  }
  ppFontOffColor = 4;                  { Font "off" colour palette index }
  ppMaxColors    = 5;                  { Maximum number of entries in palette }


{***********************************************************************}
{ Create the Base Calculator LED Display custom component, setting the
  default values as needed }
constructor TCustomBCalcDisplay.Create (AOwner : TComponent);

begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable, csOpaque] - [csSetCaption];
  Caption := '';  { Not exposed to the user }

  FBuffer := TBitmap.Create;
  FBuffer.OnChange := BufferChanged;
  FBufferValid := False;

  FSmallFont := TPicture.Create;
  FMediumFont := TPicture.Create;
  FLargeFont := TPicture.Create;
  FErrorFont := TPicture.Create;
  FStatusFont := TPicture.Create;

  DisplayChanged;

  { Set the custom component's default values }
  Height := 81;
  Width := 381;

{ Value := ''; }                        { Already set by inherited Create }
{ CalcMode := cmDecimal; }              { Already set by inherited Create }
  CalcSize := cs32Bit;
{ DisplayMode := dmNormal; }            { Already set by inherited Create }
  Separator := spComma;
{ LeadingZeros := False; }              { Already set by inherited Create }

  UsePalette := True;
  Color := clBlack;
  FontOnColor := $0000FF;               { Full red }
  FontOffColor := $000040;              { Very dark red }
  UseFontOff := True;

{ MemInUse := False; }                  { Already set by inherited Create }
{ ParenLevel := 0; }                    { Already set by inherited Create }
{ CurrentCalcOp := coNone; }            { Already set by inherited Create }
{ InConstOpMode := False; }             { Already set by inherited Create }

  SampleDisplay := False
end;


{***********************************************************************}
{ Destroy the created custom component }
destructor TCustomBCalcDisplay.Destroy;

begin
  FErrorFont.Free;
  FLargeFont.Free;
  FMediumFont.Free;
  FSmallFont.Free;
  FStatusFont.Free;
  DestroyPalette;                         { Destroy palette BEFORE FBuffer }
  FBuffer.Free;

  inherited Destroy
end;


{***********************************************************************}
{ Check the dimensions of the off-screen buffer (FBuffer) and alter it
  to match the dimensions of the actual on-screen component, if necessary }
procedure TCustomBCalcDisplay.CheckBufferSize;

begin
  if FBuffer.Width <> Width then
    begin
      FBufferValid := False;
      FBuffer.Width := Width
    end;
  if FBuffer.Height <> Height then
    begin
      FBufferValid := False;
      FBuffer.Height := Height
    end;
  if not FBufferValid and not FDrawing then
    Invalidate
end;


{***********************************************************************}
{ Return the Canvas of the off-screen buffer bitmap, not the Canvas of
  the real display.  This canvas should be treated as read-only; writing
  to the canvas is not encouraged, although possible }
function TCustomBCalcDisplay.GetCanvas : TCanvas;

begin
  CheckBufferSize;
  Result := FBuffer.Canvas
end;


{***********************************************************************}
{ Draw the current value and status indicators on the custom control,
  using all of the current settings.  This code is VERY dependent on the
  position of digits, words, etc in the various font bitmaps.  For more
  information on these bitmaps, please see Fonts\BCFonts.pas }
procedure TCustomBCalcDisplay.Paint;

type
  TFontType    = (ftLarge, ftMedium, ftSmall);
  TDispDetails = record
                   Font     : TFontType; { Font type to use }
                   MaxLen   : word;      { Maximum number of digits }
                   Lines    : word;      { Number of lines in display }
                   L1Len    : word;      { Line length of first line }
                   LxLen    : word;      { Line length of all other lines }
                 end;

const
  ROP_DSPDxax  = $E20746; { ROP code for bitmap display }

  StatusX      = 0;       { Horizontal offset of the status indicators }
  StatusY      = 0;       { Vertical offset of the status indicators }
  MainX        = 0;       { Horizontal offset of the main LED digits display }
  MainYnoStat  = 0;       { Vertical offset of the main part of the LED display,
                            if FSampleDisplay = True (added to StatusY) }

  { Status indicator relative vertical offsets within the StatusFont bitmap }
  siyNumLines  = 21;      { Number of lines in the StatusFont bitmap }
  siyDigits    = 0;       { Position of digits in StatusFont bitmap }
  siyMem       = 1;       { Position of MEM word in StatusFont bitmap }
  siyParen     = 2;       { Position of PAREN word }
  siyConst     = 3;       { Position of CONST word }
  siyOperators = 3;       { Position of the start of the operators - 1 }

  ParenDigits  = 4;       { Minimum number of digits to reserve for ParenLevel }

  SepCountDecimal = 3;    { Separator count for cmDecimal mode }
  SepCountOther   = 4;    { Separator count for other modes }

  { Display details for Value, depending on CalcMode and CalcSize }
  DispDetails  : array [TCalcMode, TCalcSize] of TDispDetails =
     { Decimal mode }
   (((Font: ftLarge;  MaxLen: 3;   Lines: 1; L1Len: 3;  LxLen: 0),   { 8-bit   }
     (Font: ftLarge;  MaxLen: 5;   Lines: 1; L1Len: 5;  LxLen: 0),   { 16-bit  }
     (Font: ftMedium; MaxLen: 10;  Lines: 1; L1Len: 10; LxLen: 0),   { 32-bit  }
     (Font: ftMedium; MaxLen: 20;  Lines: 2; L1Len: 5;  LxLen: 15),  { 64-bit  }
     (Font: ftMedium; MaxLen: 29;  Lines: 2; L1Len: 14; LxLen: 15),  { 96-bit  }
     (Font: ftSmall;  MaxLen: 39;  Lines: 2; L1Len: 9;  LxLen: 30)), { 128-bit }
     { Hexadecimal mode }
    ((Font: ftLarge;  MaxLen: 2;   Lines: 1; L1Len: 2;  LxLen: 0),   { 8-bit   }
     (Font: ftLarge;  MaxLen: 4;   Lines: 1; L1Len: 4;  LxLen: 0),   { 16-bit  }
     (Font: ftMedium; MaxLen: 8;   Lines: 1; L1Len: 8;  LxLen: 0),   { 32-bit  }
     (Font: ftMedium; MaxLen: 16;  Lines: 1; L1Len: 16; LxLen: 0),   { 64-bit  }
     (Font: ftMedium; MaxLen: 24;  Lines: 2; L1Len: 12; LxLen: 12),  { 96-bit  }
     (Font: ftMedium; MaxLen: 32;  Lines: 2; L1Len: 16; LxLen: 16)), { 128-bit }
     { Binary mode }
    ((Font: ftLarge;  MaxLen: 8;   Lines: 1; L1Len: 8;  LxLen: 0),   { 8-bit   }
     (Font: ftMedium; MaxLen: 16;  Lines: 1; L1Len: 16; LxLen: 0),   { 16-bit  }
     (Font: ftMedium; MaxLen: 32;  Lines: 2; L1Len: 16; LxLen: 16),  { 32-bit  }
     (Font: ftSmall;  MaxLen: 64;  Lines: 2; L1Len: 32; LxLen: 32),  { 64-bit  }
     (Font: ftSmall;  MaxLen: 96;  Lines: 3; L1Len: 32; LxLen: 32),  { 96-bit  }
     (Font: ftSmall;  MaxLen: 128; Lines: 4; L1Len: 32; LxLen: 32)), { 128-bit }
     { Octal mode }
    ((Font: ftLarge;  MaxLen: 3;   Lines: 1; L1Len: 3;  LxLen: 0),   { 8-bit   }
     (Font: ftLarge;  MaxLen: 6;   Lines: 1; L1Len: 6;  LxLen: 0),   { 16-bit  }
     (Font: ftMedium; MaxLen: 11;  Lines: 1; L1Len: 11; LxLen: 0),   { 32-bit  }
     (Font: ftMedium; MaxLen: 22;  Lines: 2; L1Len: 6;  LxLen: 16),  { 64-bit  }
     (Font: ftMedium; MaxLen: 32;  Lines: 2; L1Len: 16; LxLen: 16),  { 96-bit  }
     (Font: ftSmall;  MaxLen: 43;  Lines: 2; L1Len: 11; LxLen: 32)));{ 128-bit }

  { Absolute maximum number of lines and columns that a given font can occupy }
  AbsMaxLines       : array [TFontType] of byte = (1, 2, 4);
  AbsMaxCols        : array [TFontType] of byte = (9, 19, 36);
  AbsMaxColsNoSep   : array [TFontType] of byte = (8, 16, 32);

  { Absolute maximum number of lines and columns that a given font can occupy,
    for the sample display }
  AbsMaxLinesSample : array [TFontType] of byte = (1, 1, 1);
  AbsMaxColsSample  : array [TFontType] of byte = (4, 9, 19);


  { Internal procedure: draw a single digit at (X,Y).  This procedure depends
    on the layout of digits in the font bitmap }
  procedure DrawSingleDigit (C : char; BFont : TBitmap; X, Y, W, H : integer);

  var
    COff : integer;   { Character offset within the font }
    R    : TRect;

  begin
    C := UpCase(C);
    case C of
      '0'..'9' : COff := Ord(C) - Ord('0');
      'A'..'F' : COff := Ord(C) - Ord('A') + 10;
      '-'      : COff := 16;
    else
      { Space and other unknown characters }
      COff := 17
    end;

    with FBuffer.Canvas do
      begin
        R := Rect(X,Y, X+W,Y+H);
        if UseFontOff then
          begin
            Brush.Color := FRFontOffColor;
            CopyRect(R, BFont.Canvas, Rect(COff*W,H, (COff+1)*W,H*2))
          end;
        Brush.Color := FRFontOnColor;
        CopyRect(R, BFont.Canvas, Rect(COff*W,0, (COff+1)*W,H))
      end
  end;


  { Internal procedure: draw a half-width separator at (X,Y).  This procedure
    depends on the layout of digits in the font bitmap }
  procedure DrawSepChar (Sep : TSeparator; UseFontOn : boolean;
                         BFont : TBitmap; X, Y, W, H : integer);

  var
    COff : integer;   { Half-character offset within font }
    R    : TRect;

  begin
    if (Sep = spSpace) or ((not UseFontOn) and (not UseFontOff)) then
      exit;

    with FBuffer.Canvas do
      begin
        R := Rect(X,Y, X+W,Y+H);
        COff := 0;
        case Sep of
          spSpace    : COff := 38; { Note that these offsets are in half-widths }
          spComma    : COff := 36;
          spFullStop : COff := 37;
        end;

        if UseFontOn then
          begin
            Brush.Color := FRFontOnColor;
            CopyRect(R, BFont.Canvas, Rect(COff*W,0, (COff+1)*W,H))
          end
        else
          begin
            Brush.Color := FRFontOffColor;
            CopyRect(R, BFont.Canvas, Rect(COff*W,H, (COff+1)*W,H*2))
          end
      end
  end;


  { Internal procedure: draw a single row of digits }
  procedure DrawDigitsRow (S : string; const Details : TDispDetails;
                           LastLine : boolean; BFont : TBitmap;
                           X, Y, W, H : integer);

  var
    SepCount,
    CurSepCount, I : integer;
    UseFontOn      : boolean;

  begin
    { Determine where to place separators }
    if CalcMode = cmDecimal then
      SepCount := SepCountDecimal
    else
      SepCount := SepCountOther;

    with FBuffer.Canvas do
      begin
        CurSepCount := SepCount;
        UseFontOn := True;

        { Process the string of digits backwards }
        for I := Length(S) downto 1 do
          begin
            { Check for and draw the separator }
            if (S[I] = ' ') or (S[I] = '-') then
              UseFontOn := False;
            if (Separator <> spNone) and (CurSepCount = SepCount) then
              begin
                Dec(X, W div 2);
                if LastLine then
                  begin
                    { The last line never ends in a visible separator }
                    DrawSepChar(spSpace, False, BFont, X, Y, W div 2, H);
                    LastLine := False
                  end
                else
                  DrawSepChar(Separator, UseFontOn, BFont, X, Y, W div 2, H)
              end;
            if CurSepCount = 1 then
              CurSepCount := SepCount
            else
              Dec(CurSepCount);

            { Draw the digit }
            Dec(X, W);
            DrawSingleDigit(S[I], BFont, X, Y, W, H)
          end
      end
  end;


  { Internal procedure: draw all of the digits in Value }
  procedure DrawAllDigits;

  var
    Details     : TDispDetails;
    FontToUse   : TPicture;
    MLen, Len,
    SPos, SCnt,
    MainY,
    I, J,
    X, Y, W, H  : integer;
    V, S        : string;
    C           : char;
    Neg         : boolean;

  begin
    { Determine which font to use }
    Details := DispDetails[CalcMode, CalcSize];
    FontToUse := nil;

    case Details.Font of
      ftLarge  : FontToUse := LargeFont;
      ftMedium : FontToUse := MediumFont;
      ftSmall  : FontToUse := SmallFont;
    end;

    { Check if font bitmap is actually present }
    if not (FontToUse.Graphic is TBitmap) then
      exit;

    { Check if Value is negative (only for cmDecimal) }
    Neg := (Value <> '') and (Value[1] = '-');
    if not Neg then
      V := Value
    else
      begin
        { Strip out the negative sign from the string }
        V := Copy(Value, 2, Length(Value) - 1);
        Neg := (CalcMode = cmDecimal)
      end;

    { Make the value in V the exact length required }
    Len := Length(V);
    MLen := Details.MaxLen;
    if Len <= MLen then
      begin
        { Pad the value with either zeros or spaces }
        if LeadingZeros then
          C := '0'
        else
          C := ' ';
        V := StringOfChar(C, MLen - Len) + V
      end
    else { Length(Value) > ML }
      { Use the last portion of the string }
      V := Copy(V, Len - MLen + 1, MLen);

    W := FontToUse.Bitmap.Width div 20;
    H := FontToUse.Bitmap.Height div 2;

    if not SampleDisplay then
      begin
        if StatusFont.Graphic is TBitmap then
          MainY := StatusY + StatusFont.Bitmap.Height div siyNumLines
        else
          MainY := StatusY;
        if Separator <> spNone then
          X := MainX + (Width + AbsMaxCols[Details.Font] * W) div 2
        else
          X := MainX + (Width + AbsMaxColsNoSep[Details.Font] * W) div 2;
        Y := MainY + ((AbsMaxLines[Details.Font] - Details.Lines) * H) div 2
      end
    else
      begin
        X := MainX + (Width + AbsMaxColsSample[Details.Font] * W) div 2;
        Y := StatusY + MainYnoStat + (Height - (AbsMaxLinesSample[Details.Font]
             - Details.Lines + 1) * H) div 2
      end;

    { Draw each line of the display }
    SPos := 1;
    SCnt := Details.L1Len;
    for I := 1 to Details.Lines do
      begin
        S := Copy(V, SPos, SCnt);

        { Insert the negative sign if necessary }
        if Neg and (S[Length(S)] <> ' ') then
          begin
            for J := Length(S) downto 1 do
              if S[J] = ' ' then
                begin
                  { String contains a space (not the last one) that can be
                    replaced with a negative sign }
                  S[J] := '-';
                  Neg := False;
                  break
                end;
            if Neg then
              begin
                { The string does not contain a space that can be replaced;
                  simply add it in front }
                S := '-' + S;
                Neg := False
              end
          end;

        { Draw the row of digits }
        DrawDigitsRow(S, Details, (I = Details.Lines), FontToUse.Bitmap, X, Y, W, H);
        Inc(Y, H);

        Inc(SPos, SCnt);
        SCnt := Details.LxLen;
      end
  end;


  { Internal procedure: draw the status indicators MEM, PAREN and CONST }
  procedure DrawStatusIndicators;

  var
    DoDraw  : boolean;
    W, WW,
    H, HH,
    I, L, N : integer;
    S       : string;
    R       : TRect;

  begin
    if StatusFont.Graphic is TBitmap then
      with FBuffer.Canvas do
        begin
          W := StatusFont.Width;
          H := StatusFont.Height div siyNumLines;
          R := Rect(StatusX,StatusY, StatusX+W,StatusY+H);

          { Draw the MEM (memory in use) status indicator }
          DoDraw := True;
          if MemInUse then
            Brush.Color := FRFontOnColor
          else if UseFontOff then
            Brush.Color := FRFontOffColor
          else
            DoDraw := False;

          if DoDraw then
            begin
              CopyRect(R, StatusFont.Bitmap.Canvas,
                       Rect(0,siyMem*H, W,(siyMem+1)*H))
            end;
          Inc(R.Left, W);
          Inc(R.Right, W);

          { Draw the PAREN (parenthesis level) status indicator }
          DoDraw := True;
          if ParenLevel > 0 then
            Brush.Color := FRFontOnColor
          else if UseFontOff then
            Brush.Color := FRFontOffColor
          else
            DoDraw := False;

          WW := W div 10;
          if DoDraw then
            begin
              CopyRect(R, StatusFont.Bitmap.Canvas,
                       Rect(0,siyParen*H, W,(siyParen+1)*H));
              Inc(R.Left, W);
              Inc(R.Right, W);

              S := IntToStr(ParenLevel);
              L := Length(S);
              R.Right := R.Left + WW;
              for I := 1 to L do
                begin
                  N := Ord(S[I]) - Ord('0');
                  CopyRect(R, StatusFont.Bitmap.Canvas,
                           Rect(N*WW,siyDigits*H, (N+1)*WW,(siyDigits+1)*H));
                  Inc(R.Left, WW);
                  Inc(R.Right, WW)
                end;
              if L < ParenDigits then
                Inc(R.Left, (ParenDigits - L) * WW);
              R.Right := R.Left + W
            end
          else { not DoDraw }
            begin
              Inc(R.Left, W + ParenDigits * WW);
              Inc(R.Right, W + ParenDigits * WW)
            end;

          { Draw the CONST (constant operation) status indicator }
          DoDraw := True;
          if ConstantMode then
            Brush.Color := FRFontOnColor
          else if UseFontOff then
            Brush.Color := FRFontOffColor
          else
            DoDraw := False;

          if DoDraw then
            CopyRect(R, StatusFont.Bitmap.Canvas, Rect(0,siyConst*H, W,(siyConst+1)*H));

          { Draw the current operation status indicator, if not coNone }
          if CurrentCalcOp <> coNone then
            begin
              Inc(R.Left, W);
              Inc(R.Right, W);
              Brush.Color := FRFontOnColor;
              HH := (siyOperators + Ord(CurrentCalcOp)) * H;
              CopyRect(R, StatusFont.Bitmap.Canvas, Rect(0,HH, W,HH+H))
            end
        end
  end;


  { Internal procedure: draw the error message bitmap }
  procedure DrawErrorMessage;

  var
    MainY,
    X, Y,
    W, H     : integer;
    R        : TRect;

  begin
    if ErrorFont.Graphic is TBitmap then
      with FBuffer.Canvas do
        begin
          if StatusFont.Graphic is TBitmap then
            MainY := StatusY + StatusFont.Bitmap.Height div siyNumLines
          else
            MainY := StatusY;

          W := ErrorFont.Width;
          H := ErrorFont.Height div 2;
          X := MainX + (Width - W) div 2;
          Y := (Height - MainY - H) div 2 + MainY;

          R := Rect(X,Y, X+W,Y+H);

          if UseFontOff then
            begin
              Brush.Color := FRFontOffColor;
              CopyRect(R, ErrorFont.Bitmap.Canvas, Rect(0,H, W,H*2))
            end;
          Brush.Color := FRFontOnColor;
          CopyRect(R, ErrorFont.Bitmap.Canvas, Rect(0,0, W,H))
        end
  end;


var
  OldDrawing   : boolean;
  RealCanvas   : TCanvas;

begin  { TCustomBCalcDisplay.Paint }
  OldDrawing := FDrawing;
  FDrawing := True;

  try
    CheckBufferSize;
    RealCanvas := inherited Canvas;

    { Check if the buffer is valid; if not, redraw the entire display }
    if not FBufferValid then
      with FBuffer.Canvas do
        begin
          CopyMode := cmSrcCopy;
          Brush.Color := FRColor;
          FillRect(ClientRect);
          CopyMode := ROP_DSPDxax;

          if DisplayMode <> dmBlank then
            begin
              { Draw the status indicators, if this is not a sample display }
              if not SampleDisplay then
                DrawStatusIndicators;

              { Draw the actual digits, or the error message }
              if DisplayMode = dmError then
                DrawErrorMessage
              else { FDisplayMode = dmNormal }
                DrawAllDigits
            end;

          FBufferValid := True
        end;

    { Draw the display using the off-screen buffer bitmap }
    RealCanvas.Draw(0,0, FBuffer)
  finally
    FDrawing := OldDrawing
  end
end;


{***********************************************************************}
{ This event procedure is called when the off-screen buffer, FBuffer, is
  modified.  If we are not drawing (ie, when FDrawing is False), such a
  modification is due to the user directly accessing the buffer via the
  Canvas property; in this case, the display needs to be refreshed.  Note
  that such direct accesses are not encouraged }
procedure TCustomBCalcDisplay.BufferChanged (Sender : TObject);

begin
  if not FDrawing then
    Invalidate
end;


{***********************************************************************}
{ Return the current palette, if used, for Windows to use }
function TCustomBCalcDisplay.GetPalette : HPalette;

begin
  Result := FPalette
end;


{***********************************************************************}
{ Create a colour palette, if needed, and inform Windows of it }
procedure TCustomBCalcDisplay.CreatePalette;

var
  Pal : TMaxLogPalette;

begin
  if (not CanUsePalette) or PaletteUsed then
    exit;

  CheckBufferSize;

  Pal.palVersion := $300;
  Pal.palNumEntries := ppMaxColors;
  TColor(Pal.palPalEntry[ppBlack]) := clBlack;
  TColor(Pal.palPalEntry[ppWhite]) := clWhite;
  TColor(Pal.palPalEntry[ppColor]) := ColorToRGB(FColor);
  TColor(Pal.palPalEntry[ppFontOnColor]) := ColorToRGB(FFontOnColor);
  TColor(Pal.palPalEntry[ppFontOffColor]) := ColorToRGB(FFontOffColor);

  FPalette := Windows.CreatePalette(PLogPalette(@Pal)^);
  if FPalette <> 0 then
    begin
      FPaletteUsed := True;
      FBuffer.Palette := FPalette;

      FRColor := PaletteIndex(ppColor);
      FRFontOnColor := PaletteIndex(ppFontOnColor);
      FRFontOffColor := PaletteIndex(ppFontOffColor)
    end
end;


{***********************************************************************}
{ Destroy the current palette, if one exists }
procedure TCustomBCalcDisplay.DestroyPalette;

begin
  if FPalette <> 0 then
    begin
      FRColor := ColorToRGB(FColor);
      FRFontOnColor := ColorToRGB(FFontOnColor);
      FRFontOffColor := ColorToRGB(FFontOffColor);

      FBuffer.Palette := 0;
      DeleteObject(FPalette);
      FPalette := 0;
      FPaletteUsed := False
    end
end;


{***********************************************************************}
{ Modify the custom control's palette to include the new colour }
procedure TCustomBCalcDisplay.ChangePalette (Value : TColor; Index : cardinal;
                                             var RealColor : TColor);

begin
  RealColor := ColorToRGB(Value);
  if FPaletteUsed then
    begin
      SetPaletteEntries(FPalette, Index, 1, RealColor);
      RealColor := PaletteIndex(Index);

      { This is somewhat extreme, but it does ensure that the palette is
        correctly modified and reflected into all windows.  Simply sending
        a WM_QUERYNEWPALETTE does not seem to be enough }
      DestroyPalette;
      CreatePalette
    end
end;


{***********************************************************************}
{ The display screen has changed; check to see if a graphics palette can
  now be used.  This function should be called by the parent form on
  receiving the WM_DISPLAYCHANGED message.  This function does NOT seem
  to work when switching from 15/16/24/32-bit mode to 8-bit mode }
procedure TCustomBCalcDisplay.DisplayChanged;

var
  DC : HDC;

begin
  { Determine if a palette can be used }
  DC := GetDC(0);
  try
    FCanUsePalette := (GetDeviceCaps(DC, RasterCaps) and RC_PALETTE) <> 0;
    if FCanUsePalette then
      FCanUsePalette := (GetDeviceCaps(DC, SizePalette) >=
                        (GetDeviceCaps(DC, NumReserved) + ppMaxColors))
  finally
    ReleaseDC(0, DC)
  end;

  if FCanUsePalette and FUsePalette then
    { A palette can be used, and, in fact, is requested.  Create one }
    CreatePalette
  else if (not FCanUsePalette) and FPaletteUsed then
    { A palette can no longer be used.  Destroy the current one }
    DestroyPalette
end;


{***********************************************************************}
{ Set the display's value to the string passed in AValue.  This string,
  a sequence of digits 0-9 and A-F, is interpreted according to CalcMode,
  CalcSize and DisplayMode }
procedure TCustomBCalcDisplay.SetValue (AValue : string);

begin
  if AValue <> FValue then
    begin
      FValue := AValue;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the display's value mode: decimal, hexadecimal, binary or octal.
  This setting is used to determine where separators are placed, if
  necessary, and where to place line breaks, if required }
procedure TCustomBCalcDisplay.SetCalcMode (Value : TCalcMode);

begin
  if Value <> FCalcMode then
    begin
      FCalcMode := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the number of bits used in the value.  This setting is used to
  determine which font (small, medium or large) to use to draw the value
  digits }
procedure TCustomBCalcDisplay.SetCalcSize (Value : TCalcSize);

begin
  if Value <> FCalcSize then
    begin
      FCalcSize := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the display's mode: normal, blank or error.  Normal mode draws the
  value and status indicators as normal, blank mode does not draw anything,
  and error mode displays the "Error" font bitmap and the status indicators }
procedure TCustomBCalcDisplay.SetDisplayMode (Value : TDisplayMode);

begin
  if Value <> FDisplayMode then
    begin
      FDisplayMode := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Select which type of separator (none, space, comma or full stop) to
  use between digits in the display }
procedure TCustomBCalcDisplay.SetSeparator (Value : TSeparator);

begin
  if Value <> FSeparator then
    begin
      FSeparator := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set whether leading zeros are displayed or not in the value }
procedure TCustomBCalcDisplay.SetLeadingZeros (Value : boolean);

begin
  if Value <> FLeadingZeros then
    begin
      FLeadingZeros := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the background colour and redraw the display.  If a palette is being
  used, change and realise the palette as well }
procedure TCustomBCalcDisplay.SetColor (Value : TColor);

begin
  if Value <> FColor then
    begin
      FColor := Value;
      ChangePalette(Value, ppColor, FRColor);
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the font "on" colour and redraw the display.  This colour is used
  to draw portions of the display that are "lit up".  If a palette is
  being used, change and realise the palette as well }
procedure TCustomBCalcDisplay.SetFontOnColor (Value : TColor);

begin
  if Value <> FFontOnColor then
    begin
      FFontOnColor := Value;
      ChangePalette(Value, ppFontOnColor, FRFontOnColor);
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the font "off" colour and redraw the display.  This colour is used
  to draw portions of the display (in particular, the LED segments) that
  are "unlit".  This colour is usually close to the background colour in
  terms of brightness/contrast.  If a palette is being used, change and
  realise the palette as well }
procedure TCustomBCalcDisplay.SetFontOffColor (Value : TColor);

begin
  if Value <> FFontOffColor then
    begin
      FFontOffColor := Value;
      ChangePalette(Value, ppFontOffColor, FRFontOffColor);
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Select whether to use the font "off" colour or not }
procedure TCustomBCalcDisplay.SetUseFontOff (Value : boolean);

begin
  if Value <> FUseFontOff then
    begin
      FUseFontOff := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Select whether to use a colour palette on a palettable device (typically,
  a 256-colour display) for the component or not }
procedure TCustomBCalcDisplay.SetUsePalette (Value : boolean);

begin
  if Value <> FUsePalette then
    begin
      FUsePalette := Value;
      if Value then
        CreatePalette
      else
        DestroyPalette;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the small font bitmap to the new value }
procedure TCustomBCalcDisplay.SetSmallFont (Value : TPicture);

begin
  FSmallFont.Assign(Value);
  FBufferValid := False;
  Invalidate
end;


{***********************************************************************}
{ Set the medium font bitmap to the new value }
procedure TCustomBCalcDisplay.SetMediumFont (Value : TPicture);

begin
  FMediumFont.Assign(Value);
  FBufferValid := False;
  Invalidate
end;


{***********************************************************************}
{ Set the large font bitmap to the new value }
procedure TCustomBCalcDisplay.SetLargeFont (Value : TPicture);

begin
  FLargeFont.Assign(Value);
  FBufferValid := False;
  Invalidate
end;


{***********************************************************************}
{ Set the error font bitmap to the new value }
procedure TCustomBCalcDisplay.SetErrorFont (Value : TPicture);

begin
  FErrorFont.Assign(Value);
  FBufferValid := False;
  Invalidate
end;


{***********************************************************************}
{ Set the status indicators Status font bitmap to the new value }
procedure TCustomBCalcDisplay.SetStatusFont (Value : TPicture);

begin
  FStatusFont.Assign(Value);
  FBufferValid := False;
  Invalidate
end;


{***********************************************************************}
{ Set the MEM status indicator to either on or off }
procedure TCustomBCalcDisplay.SetMemInUse (Value : boolean);

begin
  if Value <> FMemInUse then
    begin
      FMemInUse := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the PAREN status indicator to show the number of open parentheses.
  Setting this value to 0 turns this indicator off }
procedure TCustomBCalcDisplay.SetParenLevel (Value : cardinal);

begin
  if Value <> FParenLevel then
    begin
      FParenLevel := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the status indicator to show the current calculator operation.
  Setting this value to coNone turns this indicator off }
procedure TCustomBCalcDisplay.SetCurrentCalcOp (Value : TCalcOp);

begin
  if Value <> FCurrentCalcOp then
    begin
      FCurrentCalcOp := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Set the CONST status indicator to show whether the current operation
  is being performed in CONST mode.  Setting this value to False turns
  this indicator off }
procedure TCustomBCalcDisplay.SetConstantMode (Value : boolean);

begin
  if Value <> FConstantMode then
    begin
      FConstantMode := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Select whether to use the display as a sample or not.  This should be
  set by components derived from this one, but should NOT be made public
  or published }
procedure TCustomBCalcDisplay.SetSampleDisplay (Value : boolean);

begin
  if Value <> FSampleDisplay then
    begin
      FSampleDisplay := Value;
      FBufferValid := False;
      Invalidate
    end
end;


{***********************************************************************}
{ Create the Base Calculator LED Display custom component, setting the
  default values as needed }
constructor TBCalcDisplay.Create (AOwner : TComponent);

begin
  inherited Create(AOwner);

  { Set the custom component's default values }
  Height := 81;
  Width := 381;

  SampleDisplay := False
end;


{***********************************************************************}
{ Create the Base Calculator LED Display custom component, setting the
  default values as needed }
constructor TBCalcSampleDisplay.Create (AOwner : TComponent);

begin
  inherited Create(AOwner);

  { Set the custom component's default values }
  Height := 49;
  Width := 289;

  SampleDisplay := True;

  CalcMode := cmHexadecimal;
  CalcSize := cs32Bit;
  DisplayMode := dmNormal
end;


{***********************************************************************}
{ Register TBCalcDisplay and TBCalcSampleDisplay with the Borland Delphi IDE }
procedure Register;

begin
  RegisterComponents('Base Calculator', [TBCalcDisplay, TBCalcSampleDisplay])
end;

end.

