unit Math128;

{************************************************************************
*                                                                       *
*                 128-bit Integer Mathematical Routines                 *
*                   Copyright (C) 1999, John Zaitseff                   *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      7th April, 1999.
  Version:   1.2

  This unit contains type definitions, constants and routines for 128-bit
  integer arithmetic.  For information on each function, please see the
  comments in the interface.

  Note that Borland Delphi 4 or later is required to compile this unit.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


type
  { Type definition for 128-bit integers }
  PInt128 = ^Int128;
  Int128  = record
              case integer of
                0 : (L        : array [0..3] of LongWord);
                1 : (B        : array [0..15] of byte);
                2 : (Byte     : byte);       { Least significant 8 bits  }
                3 : (Word     : word);       { Least significant 16 bits }
                4 : (LongWord : LongWord);   { Least significant 32 bits }
              end;


const
  { Zero as an Int128 }
  Zero : Int128 = (L: (0,0,0,0));


{***********************************************************************}

function Int128ToStr (const Value : Int128; Digits, Base : integer;
                      Signed : boolean) : string;
  { Convert the 128-bit number in Value to a string representation in base
    Base, and return this string.  Base must be between 2 and 36 (inclusive),
    or zero to automatically select base 10 (decimal).  If Signed is True,
    Value is treated as a signed integer; in particular, if Value is
    negative, a '-' sign will be prepended to the string and the absolute
    value converted, irrespective of the base (ie, -10 in base 10 is -A in
    base 16).  Digits specifies the minimum number of digits to return, not
    including any '-' sign.  For bases greater than 10, the uppercase letters
    'A' to 'Z' will be used, using as many as required to fully represent
    the base.  For example, if Value = 918203, 'JOHN' will be returned if
    Base = 36, 'E02BB' if Base = 16 and '918203' if Base = 10.  If Base is
    invalid, an ERangeError exception will be raised }


procedure Val128 (const S : AnsiString; Base : integer;
                  var Result : Int128; var Code : integer);
  { Convert a string representation of a number in base Base, into a
    128-bit number, and return this number in Result.  Base must be
    between 2 and 36 (inclusive), or zero to automatically select
    between base 10 (decimal) and 16 (hexadecimal).  If Base is outside
    this legal range, Base is treated is if it was zero.  S is the
    string representation of the number; leading blanks are automatically
    skipped.  Code is set to 0 if no error occurred during conversion, or
    the position of the error in the string if an error did occur; in this
    case, Result contains as much of the string as could be converted.  If
    S is empty (other than blanks), Result is set to 0 and Code is set to
    the length of S + 1.  The number may start with a '+' or '-' sign,
    which will be interpreted correctly.  If Base is zero, the string is
    treated as if Base = 10 (decimal), unless the number starts with '$',
    'x', 'X', '0x' or '0X', in which case it will be treated as if Base = 16
    (hexadecimal).  Note that these prefixes are only valid if Base = 0.
    For bases > 10, the alphabetic characters 'A' to 'Z' are treated as
    digits (case is ignored), as many as are needed to fully represent all
    the digits in that base.  For example, 'John' in base 36 is the same
    as '918203' in base 10 or 'E02BB' in base 16; 'JOHN' in base 25 is the
    same as '  +312323' or '  0x4C403' with Base = 0 }

function StrToInt128 (const S : string; Base : integer) : Int128;
  { Convert a string representation of a number in base Base, into a
    128-bit number.  This function is similar to Val128, except that
    if Base is not 0 or between 2 and 36, an ERangeError exception is
    raised.  In addition, if S is not a valid representation of the
    number in base Base, an EConvertError exception is raised.  Note
    that overflow is ignored and does not raise an exception }

function StrToInt128Def (const S : string; Base : integer;
                         const Default : Int128) : Int128;
  { Convert a string representation of a number in base Base, into a
    128-bit number.  If S is invalid, no exception is raised; instead,
    Default is returned.  If Base is invalid, the function call is
    treated as if Base was 0 }


function IntToInt128 (const Value : integer) : Int128; overload;
function IntToInt128 (const Value : Int64) : Int128; overload;
  { Convert Value, either a 32-bit or a 64-bit signed number, into a
    128-bit integer by sign-extending it }


{***********************************************************************}

function iAdd (const A, B : Int128) : Int128;
  { Add two numbers, A and B, and return the sum (A + B), ignoring overflow }

function iSub (const A, B : Int128) : Int128;
  { Subtract B from A and return the difference (A - B), ignoring underflow }

function iMul (const A, B : Int128) : Int128;
  { Multiply two numbers, A and B, together and return the product, ignoring
    any overflow.  Note that the multiplication of two 128-bit numbers
    normally requires a result of 256 bits.  This function, however, ignores
    the top 128 bits, returning only the lower 128 bits.  Note that A is the
    multiplicand, and B is the multiplier }

procedure iDivMod (const A, B : Int128; Signed : boolean;
                   var Quot, Rem : Int128);
  { Divide two numbers, A and B, and return the quotient and remainder.
    A is the dividend and B is the divisor, ie, A := Quot * B + Rem.  If
    Signed is True, a signed division is performed.  Any overflow is
    ignored; however, if B is zero, a division by zero will occur.  If Signed
    is True, A and B are treated as signed numbers.  The signs of Quot and
    Rem will be as follows:  +A, +B => +Quot, +Rem; +A, -B => -Quot, +Rem;
    -A, +B => -Quot, -Rem; -A, -B => +Quot, -Rem }

procedure iDivModU (const A : Int128; const B : LongWord;
                    var Quot : Int128; var Rem : LongWord);
  { Divide two unsigned numbers, A and B, and return the quotient and
    remainder.  A is the 128-bit unsigned dividend and B is the 32-bit
    unsigned divisor, ie, A := Quot * B + Rem.  If B is zero, a division
    by zero will occur }

function iDiv (const A, B : Int128; Signed : boolean) : Int128;
  { Divide two numbers (A div B) and return the result.  If Signed is True,
    a signed division will be performed; otherwise, an unsigned division
    is done.  Any overflows are ignored; if B is zero, a divide by zero
    exception will occur }

function iMod (const A, B : Int128; Signed : boolean) : Int128;
  { Divide two numbers (A mod B) and return the remainder.  If Signed is
    True, a signed modulus will be performed; otherwise, an unsigned
    modulus is done.  Any overflows are ignored; if B is zero, a divide by
    zero exception will occur }


{***********************************************************************}

function iNeg (const A : Int128) : Int128;
  { Return the negative of A, treating A as a signed integer }

function iAbs (const A : Int128) : Int128;
  { Return the absolute value of A, treating A as a signed integer }


{***********************************************************************}

function iNot (const A : Int128) : Int128;
  { Return the bitwise (one's) complement of A, ie, NOT A }

function iAnd (const A, B : Int128) : Int128;
  { Return the bitwise conjunction of A and B, ie, A AND B }

function iNand (const A, B : Int128) : Int128;
  { Return the one's complement of the bitwise conjunction of A and B,
    ie, NOT (A AND B) }

function iOr (const A, B : Int128) : Int128;
  { Return the bitwise disjunction of A and B, ie, A OR B }

function iNor (const A, B : Int128) : Int128;
  { Return the one's complement of the bitwise disjunction of A and B,
    ie, NOT (A OR B) }

function iXor (const A, B : Int128) : Int128;
  { Return the exclusive OR of A and B, ie, A XOR B }

function iEqv (const A, B : Int128) : Int128;
  { Return the equivalence (exclusive NOR) of A and B, ie, NOT (A XOR B) }


{***********************************************************************}

function iSgn (const A : Int128) : integer;
  { Return the sign of A, treating A as a signed integer.  In other words,
    return 0 if A is zero, 1 if A is positive, and -1 if A is negative }

function iCmp (const A, B : Int128; Signed : boolean) : integer;
  { Compare A and B, and return 0 if they are equal, -1 if A < B, or 1
    if A > B.  If Signed is True, a signed comparison is performed, else
    an unsigned comparison is performed }


{***********************************************************************}

function iExtend (const A : Int128; BitSize : integer; Signed : boolean) : Int128;
  { Extend the number of bits (specified in BitSize) from A into a full
    128-bit number and return this number.  If Signed is True, the number
    is sign-extended, otherwise it is zero-extended.  If BitSize <= 0,
    zero is returned.  If BitSize >= 128, A is returned unchanged }

function iTruncate (const A : Int128; BitSize : integer) : Int128;
  { Truncate the 128-bit number A to the number of bits specified in
    BitSize, setting the remaining bits to zero, and return the new
    number.  If BitSize <= 0, 0 is returned.  If BitSize >= 128, A is
    returned unchanged }


{***********************************************************************}

function iShl (const A, Count : Int128; Signed : boolean) : Int128;
  { Shift the number A to the left by Count bits.  Count is treated as
    a signed number if Signed is True; negative left shifts are treated
    as a call to iShr(A, iNeg(Count), True).  If Count >= 128, the result
    is zero, by definition (all of the bits have been "shifted off" from
    A).  If Count = 0, A is returned unchanged }

function iShr (const A, Count : Int128; Signed : boolean) : Int128;
  { Shift the number A to the right by Count bits.  If Signed is True, an
    arithmetic right shift is done, where the right-most Count bits become
    the same as the sign bit in A (the right-most bit).  If Signed is
    False, a logical right shift is done, where the right-most Count bits
    become zero.  Count is treated as a signed number if Signed is True;
    negative right shifts are treated as a call to iShl(A, iNeg(Count), True).
    If Count = 0, A is returned unchanged }

function iRol (const A, Count : Int128; Signed : boolean) : Int128;
  { Rotate the number A to the left by Count bits.  Count is treated as a
    signed number if Signed is True; negative left rotations are treated as
    a positive right rotation.  If Count = 0, A is returned unchanged.  Note
    that (Count mod 128) is used instead of just Count }

function iRor (const A, Count : Int128; Signed : boolean) : Int128;
  { Rotate the number A to the right by Count bits.  Count is treated as a
    signed number if Signed is True.  Note that a rotation to the right by
    Count bits is the same as a rotation to the left by (128 - Count mod 128) }

function iRolBit (const A, Count : Int128; BitSize : integer;
                  Signed : boolean) : Int128;
  { Rotate bits 0..BitSize-1 of the number A to the left by Count bits.
    Count is treated as a signed number if Signed is True; negative left
    rotations are treated as a positive right rotation.  Note that only
    the lower BitSize bits are rotated; the upper bits (if any) are extended
    via a call to iExtend (using Signed as a parameter).  If Count = 0 or
    BitSize <= 1, A is returned appropriately extended.  Also note that
    (Count mod BitSize) is used instead of just Count }

function iRorBit (const A, Count : Int128; BitSize : integer;
                  Signed : boolean) : Int128;
  { Rotate bits 0..BitSize-1 of the number A to the right by Count bits.
    Count is treated as a signed number if Signed is True; negative right
    rotations are treated as a positive left rotation.  Note that only
    the lower BitSize bits are rotated; the upper bits (if any) are extended
    via a call to iExtend (using Signed as a parameter).  If Count = 0 or
    BitSize <= 1, A is returned appropriately extended.  Also note that
    (Count mod BitSize) is used instead of just Count }


implementation


uses
  SysUtils;


resourcestring
  SBaseOutOfRange = '%d is not a legal base (needs to be 0 or 2..36)';
  SInvalidInteger = '''%s'' is not a valid Int128 value in base %d';


{***********************************************************************}
{ Convert the 128-bit number in Value to a string representation in base
  Base, and return this string.  Base must be between 2 and 36 (inclusive),
  or zero to automatically select base 10 (decimal).  If Signed is True,
  Value is treated as a signed integer; in particular, if Value is
  negative, a '-' sign will be prepended to the string and the absolute
  value converted, irrespective of the base (ie, -10 in base 10 is -A in
  base 16).  Digits specifies the minimum number of digits to return, not
  including any '-' sign.  For bases greater than 10, the uppercase letters
  'A' to 'Z' will be used, using as many as required to fully represent
  the base.  For example, if Value = 918203, 'JOHN' will be returned if
  Base = 36, 'E02BB' if Base = 16 and '918203' if Base = 10.  If Base is
  invalid, an ERangeError exception will be raised }

function Int128ToStr (const Value : Int128; Digits, Base : integer;
                      Signed : boolean) : string;

const
  DigitChar : array [0..35] of char = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';

var
  Temp : Int128;
  Neg  : boolean;
  D    : cardinal;
  L    : integer;
  S    : string;

begin
  if (Base < 0) or (Base > 36) or (Base = 1) then
    raise ERangeError.CreateFmt(SBaseOutOfRange, [Base]);

  if Base = 0 then
    Base := 10;

  Result := '';
  Neg := (Signed = True) and (iSgn(Value) < 0);
  if Neg then
    Temp := iNeg(Value)
  else
    Temp := Value;

  { Build the string from the right-hand side (least significant digit) }

  while (iSgn(Temp) <> 0) do            { while Temp <> 0 do      }
    begin                               {   D := Temp mod Base    }
      iDivModU(Temp, Base, Temp, D);    {   Temp := Temp div Base }
      Result := DigitChar[D] + Result
    end;

  { Check for zero result }

  if Result = '' then
    Result := '0';

  { Pad result string with extra 0 digits, if necessary }

  L := Digits - Length(Result);
  if L > 0 then
    begin
      SetLength(S, L);
      FillChar(S[1], L, '0');
      Result := S + Result
    end;

  if Neg then
    Result := '-' + Result
end;


{***********************************************************************}
{ Convert a string representation of a number in base Base, into a
  128-bit number, and return this number in Result.  Base must be
  between 2 and 36 (inclusive), or zero to automatically select
  between base 10 (decimal) and 16 (hexadecimal).  If Base is outside
  this legal range, Base is treated is if it was zero.  S is the
  string representation of the number; leading blanks are automatically
  skipped.  Code is set to 0 if no error occurred during conversion, or
  the position of the error in the string if an error did occur; in this
  case, Result contains as much of the string as could be converted.  If
  S is empty (other than blanks), Result is set to 0 and Code is set to
  the length of S + 1.  The number may start with a '+' or '-' sign,
  which will be interpreted correctly.  If Base is zero, the string is
  treated as if Base = 10 (decimal), unless the number starts with '$',
  'x', 'X', '0x' or '0X', in which case it will be treated as if Base = 16
  (hexadecimal).  Note that these prefixes are only valid if Base = 0.
  For bases > 10, the alphabetic characters 'A' to 'Z' are treated as
  digits (case is ignored), as many as are needed to fully represent all
  the digits in that base.  For example, 'John' in base 36 is the same
  as '918203' in base 10 or 'E02BB' in base 16; 'John' in base 25 is the
  same as '  +312323' or '0x4C403' with Base = 0 }

procedure Val128 (const S : AnsiString; Base : integer;
                  var Result : Int128; var Code : integer);

var
  iBase, iDigit, iTemp : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of S, EDX = value of Base, ECX = address of Result,
    [EBP+8] = address of Code }

        push    esi
        push    edi
        push    ebx

        mov     esi,S           { ESI = address of string S }
        mov     edi,Result      { EDI = address of Result }
        mov     ebx,[Code]      { EBX = address of Code }

        push    esi             { Save a copy of the string pointer, for errors }

        mov     eax,Base        { Check that Base is either 0 or in the }
        cmp     eax,0           {   range 2..36.  If not, set Base to 0 }
        je      @bsok           { Note that EDX = Base }
        cmp     eax,2
        jl      @bserr
        cmp     eax,36
        jle     @bsok
@bserr: xor     eax,eax
        mov     Base,eax

@bsok:  mov     dword ptr [iBase],eax  { Store Base in a local 128-bit variable }
        xor     eax,eax
        mov     dword ptr [iBase + 4],eax
        mov     dword ptr [iBase + 8],eax
        mov     dword ptr [iBase + 12],eax

        mov     ecx,4
        push    edi
        push    ecx

        rep                     { Clear Result to zero (EAX = 0) }
        stosd

        pop     ecx

        lea     edi,[iDigit]
        rep                     { Clear iDigit to zero }
        stosd

        pop     edi

        test    esi,esi         { Check if S is an empty string }
        jz      @empt           { If so, return with Code = 1 (an error) }

@blank: lodsb                   { Skip spaces in string }
        cmp     al,' '
        je      @blank

        mov     cl,False        { CL = True to negate number }

        cmp     al,'-'          { Is the first character a sign? }
        je      @minus          { If so, remember the fact and set Base to 10 }
        cmp     al,'+'          {   if it is zero                             }
        je      @plus

        test    edx,edx         { Check if Base = 0.  If so, a leading $, x,   }
        jnz     @fsdig          {   X, 0x or 0X means base 16 (hex), else base }
                                {   10 (decimal) is assumed                    }

        cmp     al,'$'          { Is the first character '$', 'x' or 'X'? }
        je      @hexb           { If so, set Base to 16 (hexadecimal) }
        cmp     al,'x'
        je      @hexb
        cmp     al,'X'
        je      @hexb
        cmp     al,'0'          { Is the first character '0'? }
        jne     @decb           { If not, we need to check for base 10 }

        lodsb
        cmp     al,'x'          { Is the second character 'x' or 'X'? }
        je      @hexb           { If so, set Base to 16 (hexadecimal) }
        cmp     al,'X'
        je      @hexb

        test    al,al           { Is second character the end of string? }
        jz      @okex           { If so, string is '0', so return zero }

@decb:  test    al,al           { AL = 1st or 2nd character }
        jz      @empt           { If AL = 0 here, string is blank (an error) }

        mov     edx,10          { Set Base to 10 (decimal) }
        jmp     @stbs

@hexb:  mov     edx,16          { Set Base to 16 (hexadecimal) }
        lodsb
@stbs:  mov     dword ptr [iBase],edx

@fsdig: test    al,al           { First digit of the actual number }
        jz      @empt           { If it is end of string, return an error }

@diglp: cmp     al,'0'          { If char < '0', return an error }
        jb      @err
        cmp     al,'9'          { If char in ['0'..'9'], it is a decimal digit }
        jbe     @ddig
        cmp     al,'a'          { If char >= 'a', make it uppercase }
        jb      @upcs
        sub     al,'a' - 'A'
@upcs:  cmp     al,'A'          { If char > '9' and char < 'A', return an error }
        jb      @err
        sub     al,'A' - '0' - 10  { Start converting an alphabetic character }
                                   {   into a digit                           }
@ddig:  sub     al,'0'          { Convert char into a digit, 0..36 }
        cmp     al,dl           { DL = Base; if digit >= Base, return an error }
        jae     @err

        mov     dword ptr [iDigit],eax  { Store digit as a 128-bit number }

        push    eax             { Calculate Result := Result * Base + Digit }
        push    edx             { Note that we must use a temporary variable }
        push    ecx             {   for iMul, since it is overwritten during }
                                {   calculations                             }

        mov     eax,edi         { iTemp := iMul(Result, iBase) }
        lea     edx,[iBase]
        lea     ecx,[iTemp]
        call    iMul

        lea     eax,[iTemp]     { Result := iAdd(iTemp, iDigit) }
        lea     edx,[iDigit]
        mov     ecx,edi
        call    iAdd

        pop     ecx
        pop     edx
        pop     eax

        lodsb                   { Load the next character into AL }
        test    al,al           { If not the end of string, continue loop }
        jnz     @diglp

@okex:  test    cl,cl           { Do we negate the final number? }
        jz      @okex2          { If CL = True, yes, we do }

        mov     eax,edi         { EAX = address of Result, parameter to iNeg }
        mov     edx,eax         { EDX = address of Result, result of iNeg }
        call    iNeg            { Negate Result }

@okex2: pop     eax             { Discard previously saved string pointer }
        xor     eax,eax         { Exit with Code = 0 }
        jmp     @exit

@minus: inc     cl              { CL = True to negate the number at the end }
@plus:  lodsb                   { Discard '+' or '-' and process the rest }
        test    edx,edx         { If Base = 0, set the Base to 10 first }
        jz      @decb
        jmp     @fsdig

@empt:  inc     esi             { Return with Code = 1, due to an empty string }

@err:   pop     eax             { Retrieve the string's starting address }
        xchg    eax,esi         { Return with Code = (ESI - string start address) }
        sub     eax,esi

@exit:  mov     [ebx],eax       { Set the error code, Code, to EAX }

        pop     ebx
        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Convert a string representation of a number in base Base, into a
  128-bit number.  This function is similar to Val128, except that
  if Base is not 0 or between 2 and 36, an ERangeError exception is
  raised.  In addition, if S is not a valid representation of the
  number in base Base, an EConvertError exception is raised.  Note
  that overflow is ignored and does not raise an exception }
function StrToInt128 (const S : string; Base : integer) : Int128;

var
  E : integer;

begin
  if (Base < 0) or (Base > 36) or (Base = 1) then
    raise ERangeError.CreateFmt(SBaseOutOfRange, [Base]);
  Val128(S, Base, Result, E);
  if E <> 0 then
    raise EConvertError.CreateFmt(SInvalidInteger, [S, Base])
end;


{***********************************************************************}
{ Convert a string representation of a number in base Base, into a
  128-bit number.  If S is invalid, no exception is raised; instead,
  Default is returned.  If Base is invalid, the function call is
  treated as if Base was 0 }
function StrToInt128Def (const S : string; Base : integer;
                         const Default : Int128) : Int128;

var
  E : integer;

begin
  Val128(S, Base, Result, E);
  if E <> 0 then
    Result := Default
end;


{***********************************************************************}
{ Convert Value, a 32-bit signed number, into a 128-bit integer by sign-
  extending it }
function IntToInt128 (const Value : integer) : Int128; overload;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = Value, EDX = address of Result }

        mov     ecx,edx
        mov     [ecx],eax       { Result.L[0] := Value }

        cdq                     { Sign-extend EAX into EDX:EAX }

        mov     [ecx + 4],edx   { Result.L[1..3] := SignExtend(Value) }
        mov     [ecx + 8],edx
        mov     [ecx + 12],edx
end;


{***********************************************************************}
{ Convert Value, a 64-bit signed number, into a 128-bit integer by sign-
  extending it }
function IntToInt128 (const Value : Int64) : Int128; overload;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of Result, [EBP+8] = Value }

        mov     ecx,eax

        mov     eax, dword ptr [Value]   { Result.L[0..1] := Value }
        mov     [ecx],eax
        mov     eax, dword ptr [Value + 4]
        mov     [ecx + 4],eax

        cdq                              { Sign-extend EAX into EDX:EAX }

        mov     [ecx + 8],edx            { Result.L[2..3] := SignExtend(Value) }
        mov     [ecx + 12],edx
end;


{***********************************************************************}
{ Add two 128-bit numbers together and return the sum as an 128-bit
  number.  This function ignores integer overflow during the addition.
  This function could have been implemented as a loop (with a count of
  4), but it is just as easy (and much faster) to unroll such a loop }
function iAdd (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := A.L[0] + B.L[0] }
        add     eax,[B]         { Note: not ADC, since carry is not defined }
        stosd

        lodsd                   { Result.L[1] := A.L[1] + B.L[1] + Carry }
        adc     eax,[B + 4]
        stosd

        lodsd                   { Result.L[2] := A.L[2] + B.L[2] + Carry }
        adc     eax,[B + 8]
        stosd

        lodsd                   { Result.L[3] := A.L[3] + B.L[3] + Carry }
        adc     eax,[B + 12]
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Subtract two 128-bit numbers from each other (A - B) and return the
  difference as an 128-bit number.  This function ignores integer overflow
  during the subtraction.  This function could have been implemented as
  a loop (with a count of 4), but it is just as easy (and much faster)
  to unroll such a loop }
function iSub (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := A.L[0] - B.L[0] }
        sub     eax,[B]         { Note: not SBB, since carry is not defined }
        stosd

        lodsd                   { Result.L[1] := A.L[1] - B.L[1] - Carry }
        sbb     eax,[B + 4]
        stosd

        lodsd                   { Result.L[2] := A.L[2] - B.L[2] - Carry }
        sbb     eax,[B + 8]
        stosd

        lodsd                   { Result.L[3] := A.L[3] - B.L[3] - Carry }
        sbb     eax,[B + 12]
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Multiply two numbers, A and B, together and return the product, ignoring
  any overflow.  Note that the multiplication of two 128-bit numbers
  normally requires a result of 256 bits.  This function, however, ignores
  the top 128 bits, returning only the lower 128 bits.  Note that A is the
  multiplicand, and B is the multiplier.

  This function is implemented using an algorithm that essentially treats
  the single DWORDs in A and B as single digits, multiplying them one at
  a time.  The routines have been optimised in view of the fact that
  overflow is ignored.  For best results, B should be the smaller number.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointers to A or B.  This is because Result is used to store
  intermediate results of the calculation }
function iMul (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi
        push    ebx

        mov     esi,A           { ESI = address of A }
        mov     ebx,B           { EBX = address of B }
        mov     edi,Result      { EDI = address of Result }

        xor     eax,eax         { Clear Result to zero }
        mov     ecx,4
        push    edi
        rep
        stosd
        pop     edi

        mov     ecx,[ebx]       { ECX = B.L[0] }
        jecxz   @skz1           { If B.L[0] is zero, don't try to multiply }

        push    esi
        push    edi

        lodsd
        mul     ecx             { EDX:EAX = A.L[0] * B.L[0] }
        stosd                   { Result.L[1..0] = EDX:EAX }
        mov     [edi],edx

        lodsd
        mul     ecx             { EDX:EAX = A.L[1] * B.L[0] }
        add     eax,[edi]       { Result.L[3..1] = EDX:EAX + Result.L[1] }
        adc     edx,0
        stosd
        mov     [edi],edx
        adc     dword ptr [edi+4],0  { Result.L[3] = Carry of EDX:EAX + Result.L[1] }

        lodsd
        mul     ecx             { EDX:EAX = A.L[2] * B.L[0] }
        add     eax,[edi]       { Result.L[3..2] = EDX:EAX + Result.L[2] }
        adc     edx,0
        stosd
        mov     [edi],edx

        lodsd
        mul     ecx             { EDX:EAX = A.L[3] * B.L[0] }
        add     [edi],eax       { Result.L[3] = EAX + Result.L[3] }

        pop     edi
        pop     esi

@skz1:  lea     edi,[edi+4]     { EDI = address of Result.L[1] }
        mov     ecx,[ebx+4]     { ECX = B.L[1] }
        jecxz   @skz2

        push    esi
        push    edi

        lodsd
        mul     ecx             { EDX:EAX = A.L[0] * B.L[1] }
        add     eax,[edi]       { Result.L[3..1] = EDX:EAX + Result.L[1] }
        adc     edx,0
        stosd
        add     [edi],edx
        adc     dword ptr [edi+4],0  { Result.L[3] = Carry of EDX:EAX + Result.L[1] }

        lodsd
        mul     ecx             { EDX:EAX = A.L[1] * B.L[1] }
        add     eax,[edi]       { Result.L[3..2] = EDX:EAX + Result.L[2] }
        adc     edx,0
        stosd
        add     [edi],edx

        lodsd
        mul     ecx             { EDX:EAX = A.L[2] * B.L[1] }
        add     [edi],eax       { Result.L[3] = EAX + Result.L[3] }

        pop     edi             { No need to calculate A.L[3] * B.L[1], as this }
        pop     esi             {   does not contribute to the 128-bit result   }

@skz2:  lea     edi,[edi+4]     { EDI = address of Result.L[2] }
        mov     ecx,[ebx+8]     { ECX = B.L[2] }
        jecxz   @skz3

        push    esi
        push    edi

        lodsd
        mul     ecx             { EDX:EAX = A.L[0] * B.L[2] }
        add     eax,[edi]       { Result.L[3..2] = EDX:EAX + Result.L[2] }
        adc     edx,0
        stosd
        add     [edi],edx

        lodsd
        mul     ecx             { EDX:EAX = A.L[1] * B.L[2] }
        add     [edi],eax       { Result.L[3] = EAX + Result.L[3] }

        pop     edi             { No need to calculate A.L[2] * B.L[2] or }
        pop     esi             {   A.L[3] * B.L[2]                       }

@skz3:  lea     edi,[edi+4]     { EDI = address of Result.L[3] }
        mov     ecx,[ebx+12]    { ECX = B.L[3] }
        jecxz   @skz4

        lodsd
        mul     ecx             { EDX:EAX = A.L[0] * B.L[3] }
        add     [edi],eax       { Result.L[3] = EAX + Result.L[3] }

@skz4:
        pop     ebx
        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Divide two numbers, A and B, and return the quotient and remainder.
  A is the dividend and B is the divisor, ie, A := Quot * B + Rem.  If
  Signed is True, a signed division is performed.  Any overflow is
  ignored; however, if B is zero, a division by zero will occur.

  If Signed is True, A and B are treated as signed numbers.  The signs
  of Quot and Rem are set as follows:
      +A, +B => +Quot, +Rem
      +A, -B => -Quot, +Rem
      -A, +B => -Quot, -Rem
      -A, -B => +Quot, -Rem

  This routine implements a variation of the standard shift-left-and-
  subtract/increment algorithm.  Unfortunately, this algorithm requires
  128 iterations of the inner loop, in which 8 DWORD shifts, as well as
  other operations, are performed.  This makes for a slow division
  algorithm!

  For assembly programmers:  A, B, Quot and/or Rem may reference each other
  (although, for obvious reasons, Quot and Rem should be two different
  variables), as the values contained in A and B are copied into temporary
  storage at the beginning, with the results copied out at the end }
procedure iDivMod (const A, B : Int128; Signed : boolean; var Quot, Rem : Int128);

type
  Int256  = record
              Lo, Hi : Int128
            end;

var
  Temp    : Int256;         { 256-bit remainder and quotient storage }
  Divisor : Int128;
  iSigned : boolean;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, CL = value of Signed,
    [EBP + 8] = address of Quot, [EBP + 12] = address of Rem }

        push    esi
        push    edi
        push    ebx

        mov     [iSigned],cl    { Store the value of Signed }
        xor     ebx,ebx         { BL = sign of A, BH = sign of B }

        push    eax             { Test if B if zero; if so, divide by zero }
        push    edx
        mov     eax,B
        call    iSgn            { BH := iSgn(B) }
        mov     bh,al
        test    al,al
        pop     edx
        pop     eax
        jz      @divz           { if BH = 0 then DivideByZero }

        mov     ecx,4
        push    ecx

        mov     esi,A           { Temp.Lo := A }
        lea     edi,[Temp.Lo]
        rep
        movsd

        xor     eax,eax         { Temp.Hi := 0 }
        mov     ecx,[esp]       { Note that value of 4 has been saved above }
        rep
        stosd

        pop     ecx             { Divisor := B }
        mov     esi,B
        lea     edi,[Divisor]
        push    edi
        rep
        movsd
        pop     edi

        cmp     byte ptr [iSigned],False  { If Signed = True, may need to }
        je      @nsgn                     {   negate A and/or B           }

        test    bh,bh           { Check if B (Divisor) is negative }
        jnl     @skndv

        mov     eax,edi         { If so, negate it: Divisor := iNeg(Divisor) }
        mov     edx,edi         { Note EDI is set above }
        call    iNeg

@skndv: lea     eax,[Temp.Lo]   { Check if Temp.Lo (Dividend) is negative }
        push    eax
        call    iSgn
        pop     edx             { Note: EAX pushed, but EDX poped off }

        mov     bl,al           { BL := iSgn(Temp.Lo) }
        test    al,al
        jnl     @nsgn

        mov     eax,edx         { Temp.Lo := iNeg(Temp.Lo) }
        call    iNeg

@nsgn:                          { EDI = address of Divisor, from above }
        push    ebx
        mov     ecx,128         { ECX = Loop counter }
        lea     esi,[Temp.Lo]   { ESI = address of Temp.Lo }
        lea     ebx,[Temp.Hi]   { EBX = address of Temp.Hi }

@loop:  push    ecx

        shl     dword ptr [esi],1     { Shift Temp (256 bits) to the left }
        rcl     dword ptr [esi + 4],1 {   by one bit                      }
        rcl     dword ptr [esi + 8],1
        rcl     dword ptr [esi + 12],1
        rcl     dword ptr [esi + 16],1
        rcl     dword ptr [esi + 20],1
        rcl     dword ptr [esi + 24],1
        rcl     dword ptr [esi + 28],1

        mov     eax,ebx         { Check if Temp.Hi >= Divisor }
        mov     edx,edi
        xor     ecx,ecx
        call    iCmp            { EAX := iCmp(Temp.Hi, Divisor, False) }
        test    eax,eax
        jl      @ssub
                                { If so, perform subtract/increment operation }
        mov     eax,ebx
        mov     edx,edi
        mov     ecx,ebx
        call    iSub            { Temp.Hi := iSub(Temp.Hi, Divisor) }

        inc     dword ptr [esi] { Inc(Temp.L[0]): help form the quotient }

@ssub:  pop     ecx             { Repeat for a full 128 bits }
        loop    @loop

        mov     ecx,4
        push    ecx

        mov     edi,[Quot]      { Quot := Temp.Lo (ESI = Temp.Lo from above) }
        rep
        movsd

        pop     ecx             { Rem := Temp.Hi }
        mov     edi,[Rem]
        rep
        movsd

        pop     ebx             { BL = sign of A, BH = sign of B }

        cmp     byte ptr [iSigned], False  { If Signed = True, may need to }
        je      @exit                      {   negate Quot and/or Rem      }

        test    bl,bl           { If A < 0, need to negate the remainder }
        jnl     @snrm

        mov     eax,[Rem]       { Rem := iNeg(Rem) }
        mov     edx,eax
        call    iNeg

@snrm:  test    bh,bh           { If one and only one of (A < 0), (B < 0) is }
        setl    al              {   true, need to negate the quotient        }
        test    bl,bl
        setl    dl
        xor     al,dl
        jz      @exit

        mov     eax,[Quot]      { Quot := iNeg(Quot) }
        mov     edx,eax
        call    iNeg

        jmp     @exit

@divz:  xor     eax,eax         { Deliberately divide by zero to generate }
        div     eax             {   a divide by zero exception            }

@exit:
        pop     ebx
        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Divide two unsigned numbers, A and B, and return the quotient and
  remainder.  A is the 128-bit unsigned dividend and B is the 32-bit
  unsigned divisor, ie, A := Quot * B + Rem.  If B is zero, a division
  by zero will occur }
procedure iDivModU (const A : Int128; const B : LongWord;
                    var Quot : Int128; var Rem : LongWord);

var
  iB, iRem : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = value of B, ECX = address of Quot,
    [EBP + 8] = address of Rem }

        push    edi

        push    ecx             { Push address of Quot for iDivMod }
        lea     edi,[iRem]      { Push address of iRem for iDivMod }
        push    edi             { These two parameters are removed by iDivMod }

        push    eax

        lea     edi,[iB]        { iB := ZeroExtend(B) }
        mov     [edi],B
        xor     eax,eax
        mov     [edi + 4],eax
        mov     [edi + 8],eax
        mov     [edi + 12],eax

        mov     edx,edi         { EDX = address of iB }
        mov     ecx,eax         { ECX = False }
        pop     eax             { EAX = address of A }

        call    iDivMod         { iDivMod(A, iB, False, Quot, iRem) }

        mov     eax,dword ptr [iRem]
        mov     edi,[Rem]
        mov     [edi],eax       { Rem := iRem.L[0] }

        pop     edi
end;


{***********************************************************************}
{ Divide two numbers (A div B) and return the result.  If Signed is True,
  a signed division will be performed; otherwise, an unsigned division
  is done.  Any overflows are ignored; if B is zero, a divide by zero
  exception will occur }
function iDiv (const A, B : Int128; Signed : boolean) : Int128;

var
  Temp : Int128;

begin
  iDivMod(A, B, Signed, Result, Temp)
end;


{***********************************************************************}
{ Divide two numbers (A mod B) and return the remainder.  If Signed is
  True, a signed modulus will be performed; otherwise, an unsigned
  modulus is done.  Any overflows are ignored; if B is zero, a divide by
  zero exception will occur }
function iMod (const A, B : Int128; Signed : boolean) : Int128;

var
  Temp : Int128;

begin
  iDivMod(A, B, Signed, Temp, Result)
end;


{***********************************************************************}
{ Return the absolute value of A, treating A as a signed integer }
function iAbs (const A : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Result }

        cmp     dword ptr [A + 12], 0    { If sign bit is set, jump to iNeg }
        jns     @exit
        call    iNeg
@exit:
end;


{***********************************************************************}
{ Return the negative of A, treating A as a signed integer }
function iNeg (const A : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result
        xor     edx,edx

        lodsd                   { Result.L[0] := -A.L[0] }
        neg     eax
        stosd

        lodsd                   { Result.L[1] := 0 - A.L[1] - carry }
        mov     ecx,eax
        mov     eax,edx         { Note that EDX = 0 }
        sbb     eax,ecx
        stosd

        lodsd                   { Result.L[2] := 0 - A.L[2] - carry }
        mov     ecx,eax
        mov     eax,edx
        sbb     eax,ecx
        stosd

        lodsd                   { Result.L[3] := 0 - A.L[3] - carry }
        mov     ecx,eax
        mov     eax,edx
        sbb     eax,ecx
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the bitwise (one's) complement of A, ie, not A }
function iNot (const A : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := not A.L[0] }
        not     eax
        stosd

        lodsd                   { Result.L[1] := not A.L[1] }
        not     eax
        stosd

        lodsd                   { Result.L[2] := not A.L[2] }
        not     eax
        stosd

        lodsd                   { Result.L[3] := not A.L[3] }
        not     eax
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the bitwise conjunction (AND) of A and B }
function iAnd (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := A.L[0] and B.L[0] }
        and     eax,[B]
        stosd

        lodsd                   { Result.L[1] := A.L[1] and B.L[1] }
        and     eax,[B + 4]
        stosd

        lodsd                   { Result.L[2] := A.L[2] and B.L[2] }
        and     eax,[B + 8]
        stosd

        lodsd                   { Result.L[3] := A.L[3] and B.L[3] }
        and     eax,[B + 12]
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the one's complement of the conjunction of A and B, ie,
  not (A and B). }
function iNand (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := not (A.L[0] and B.L[0]) }
        and     eax,[B]
        not     eax
        stosd

        lodsd                   { Result.L[1] := not (A.L[1] and B.L[1]) }
        and     eax,[B + 4]
        not     eax
        stosd

        lodsd                   { Result.L[2] := not (A.L[2] and B.L[2]) }
        and     eax,[B + 8]
        not     eax
        stosd

        lodsd                   { Result.L[3] := not (A.L[3] and B.L[3]) }
        and     eax,[B + 12]
        not     eax
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the bitwise disjunction (OR) of A and B }
function iOr (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := A.L[0] or B.L[0] }
        or      eax,[B]
        stosd

        lodsd                   { Result.L[1] := A.L[1] or B.L[1] }
        or      eax,[B + 4]
        stosd

        lodsd                   { Result.L[2] := A.L[2] or B.L[2] }
        or      eax,[B + 8]
        stosd

        lodsd                   { Result.L[3] := A.L[3] or B.L[3] }
        or      eax,[B + 12]
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the one's complement of the bitwise disjunction of A and B, ie,
  not (A or B) }
function iNor (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := not (A.L[0] or B.L[0]) }
        or      eax,[B]
        not     eax
        stosd

        lodsd                   { Result.L[1] := not (A.L[1] or B.L[1]) }
        or      eax,[B + 4]
        not     eax
        stosd

        lodsd                   { Result.L[2] := not (A.L[2] or B.L[2]) }
        or      eax,[B + 8]
        not     eax
        stosd

        lodsd                   { Result.L[3] := not (A.L[3] or B.L[3]) }
        or      eax,[B + 12]
        not     eax
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the exclusive OR (XOR) of A and B }
function iXor (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := A.L[0] xor B.L[0] }
        xor     eax,[B]
        stosd

        lodsd                   { Result.L[1] := A.L[1] xor B.L[1] }
        xor     eax,[B + 4]
        stosd

        lodsd                   { Result.L[2] := A.L[2] xor B.L[2] }
        xor     eax,[B + 8]
        stosd

        lodsd                   { Result.L[3] := A.L[3] xor B.L[3] }
        xor     eax,[B + 12]
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the equivalence (exclusive NOR) of A and B, ie, not (A xor B) }
function iEqv (const A, B : Int128) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, ECX = address of Result }

        push    esi
        push    edi

        mov     esi,A           { Use ESI and EDI indexing instructions }
        mov     edi,Result

        lodsd                   { Result.L[0] := not (A.L[0] xor B.L[0]) }
        xor     eax,[B]
        not     eax
        stosd

        lodsd                   { Result.L[1] := not (A.L[1] xor B.L[1]) }
        xor     eax,[B + 4]
        not     eax
        stosd

        lodsd                   { Result.L[2] := not (A.L[2] xor B.L[2]) }
        xor     eax,[B + 8]
        not     eax
        stosd

        lodsd                   { Result.L[3] := not (A.L[3] xor B.L[3]) }
        xor     eax,[B + 12]
        not     eax
        stosd

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Return the sign of A, treating A as a signed integer.  In other words,
  return 0 if A is zero, 1 if A is positive, and -1 if A is negative }
function iSgn (const A : Int128) : integer;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A.  On exit, EAX = Result }

        mov     edx,A           { EDX = address of A }
        xor     eax,eax         { EAX = Result; set to zero }

        cmp     [edx + 12],eax  { If sign bit is set, number is < 0 }
        jns     @nneg
        dec     eax             { A is negative, Result := -1 }
        jmp     @exit

@nneg:  jnz     @posv           { A.L[3] is not negative and not zero }
        cmp     [edx],eax       { Check if A is non-zero; if so, it must be }
        jnz     @posv           {   a positive number                       }
        cmp     [edx + 4],eax
        jnz     @posv
        cmp     [edx + 8],eax
        jz      @exit
@posv:  inc     eax             { A is positive, Result := 1 }

@exit:
end;


{***********************************************************************}
{ Compare A and B, and return 0 if they are equal, -1 if A < B, or 1
  if A > B.  If Signed is True, a signed comparison is performed, else
  an unsigned comparison is performed }
function iCmp (const A, B : Int128; Signed : boolean) : integer;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of B, CL = Signed.
    On exit, EAX = Result }

        push    esi
        push    edi
        pushfd
        std                     { Comparisons must begin with the MSB and
                                  proceed to the LSB, ie, backwards }

        lea     esi,[A + 12]    { ESI = address of A.L[3] }
        lea     edi,[B + 12]    { EDI = address of B.L[3] }
        mov     edx,ecx         { DL = Signed }
        mov     ecx,4
        xor     eax,eax         { EAX = Result }

        or      dl,dl           { Check if Signed is False }
        jz      @ucmp

                                { Signed comparison:                           }
@scmp:  cmpsd                   {   for I := 3 downto 0 do                     }
        jl      @lt             {     if A.L[I] < B.L[I] then      Goto A_lt_B }
        jg      @gt             {     else if A.L[I] > B.L[I] then Goto A_gt_B }
        loop    @scmp           {     else ContinueLoop                        }
        jmp     @exit           {   Result := 0                                }

                                { Unsigned comparison:                         }
@ucmp:  cmpsd                   {   for I := 3 downto 0 do                     }
        jb      @lt             {     if A.L[I] < B.L[I] then      Goto A_lt_B }
        ja      @gt             {     else if A.L[I] > B.L[I] then Goto A_gt_B }
        loop    @ucmp           {     else ContinueLoop                        }
        jmp     @exit           {   Result := 0                                }

@lt:    dec     eax             { A < B: Result := -1 }
        jmp     @exit

@gt:    inc     eax             { A > B: Result := 1 }

@exit:  popfd
        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Extend the number of bits (specified in BitSize) from A into a full
  128-bit number and return this number.  If Signed is True, the number
  is sign-extended, otherwise it is zero-extended.  If BitSize <= 0,
  zero is returned.  If BitSize >= 128, A is returned unchanged }
function iExtend (const A : Int128; BitSize : integer; Signed : boolean) : Int128;

var
  iSigned : boolean;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = BitSize, CL = Signed, [EBP + 8] = address of
    Result }

        push    esi
        push    edi
        push    ebx

        mov     iSigned,Signed  { Preserve the value of Signed (CL) }
        mov     esi,A
        mov     edi,Result
        mov     ebx,4

        cmp     BitSize,0       { Check that BitSize is in legal range }
        jg      @nonz

        mov     ecx,ebx         { BitSize <= 0:  Assign 0 to Result }
        xor     eax,eax
        rep                     { Note that EDI points to Result }
        stosd
        jmp     @exit

@nonz:  mov     eax,BitSize     { EBX = 4 from above }
        shr     eax,5           { EAX = DWORD offset in variable A }

        cmp     eax,ebx         { Check if BitSize is > 128 }
        jle     @bok
        mov     eax,ebx         { BitSize > 128: only copy 4 DWORDs into Result }
@bok:
        mov     ebx,eax         { EBX = DWORD offset in variable A }

        mov     ecx,eax         { Copy ECX number of DWORDs from A to Result }
        jecxz   @cpz
        rep
        movsd
@cpz:   cmp     ebx,4           { Have we copied 4 DWORDs (ie, 128 bits)? }
        je      @exit           { Yes, we have finished (BitSize >= 128) }

        and     BitSize,$1F     { Only look at the lowest 5 bits of BitSize }
        mov     cl,dl           { Create a mask in EAX: }
        mov     eax,1           {   Mask := (1 shl (BitSize and $1F)) - 1 }
        shl     eax,cl
        dec     eax
        push    eax             { Save this mask for later use }

        lodsd                   { EAX = DWORD to be modified during copy }
        cmp     iSigned,False   { Do a signed extension? }
        je      @nsgn           { No: Signed = False }
        dec     BitSize         { Make BitSize a bit offset (0-30 or -1) }
        jns     @nprv           { If new BitSize < 0, old BitSize was 32 }

        cmp     dword ptr [esi - 8],0  { Check if previous DWORD's sign bit is set }
        jns     @nsgn
        jmp     @sgn

@nprv:  bt      eax,BitSize     { Set CF if BitSize'th bit is set }
        jnc     @nsgn           { Do an unsigned extension as sign bit is not set }

@sgn:   pop     edx             { Retrieve saved Mask into EDX }
        not     edx             { Mask := not Mask }

        or      eax,edx         { EAX = A.L[EBX] or Mask }
        cdq                     { EDX = -1 }
        jmp     @szx

@nsgn:  pop     edx             { Retrieve saved Mask into EDX }
        and     eax,edx         { EAX = A.L[EBX] and Mask }
        xor     edx,edx         { EDX = 0 }

@szx:   stosd                   { Result.L[EBX] := EAX }

        mov     eax,3           { Have we finished copying? }
        sub     eax,ebx         { EAX = number of DWORDs left to change }

        mov     ecx,eax         { Set the remaining DWORDs in Result to EDX, }
        mov     eax,edx         { which is the sign or zero extension }
        jecxz   @cpz2
        rep
        stosd
@cpz2:

@exit:  pop     ebx
        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Truncate the 128-bit number A to the number of bits specified in
  BitSize, setting the remaining bits to zero, and return the new
  number.  If BitSize <= 0, 0 is returned.  If BitSize >= 128, A is
  returned unchanged.

  iTruncate(A, BitSize) performs exactly the same function as
  iExtend(A, BitSize, False).  For this reason, this function calls
  iExtend with the additional parameter }
function iTruncate (const A : Int128; BitSize : integer) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = BitSize, ECX = address of Result }

        push    ecx             { Store address of Result for iExtend }
        xor     ecx,ecx         { CL ( = Signed) := False }
        call    iExtend         { Result := iExtend(A, BitSize, False) }
end;


{***********************************************************************}
{ Shift the number A to the left by Count bits.  Count is treated as
  a signed number if Signed is True; negative left shifts are treated
  as a call to iShr(A, iNeg(Count), True).  If Count >= 128, the result
  is zero, by definition (all of the bits have been "shifted off" from
  A).  If Count = 0, A is returned unchanged.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointer to A.  This is because Result is used to store
  intermediate results of the operation }
function iShl (const A, Count : Int128; Signed : boolean) : Int128;

var
  Temp : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Count, CL = Signed, [EBP+8] = address
    of Result}

        or      Signed,Signed   { Test if Signed is True and Count < 0 }
        jz      @nneg
        cmp     dword ptr [Count + 12], 0    { Test if sign bit is set }
        jns     @nneg

@cltz:  push    eax             { Count < 0; call iShr(A, iNeg(Count), True) }
        mov     eax,Count       { EAX = address of operand for iNeg }
        lea     edx,Temp        { EDX = address of result; use Temp to save }
        push    edx             {   Count from being changed                }
        call    iNeg            { Temp := iNeg(Count) }
        pop     edx             { EDX = address of new Count }
        pop     eax             { EAX = address of A for iShr }
        push    Result          { In iShr, [EBP+8] = address of Result }
        mov     cl,True         { CL = Signed := True }
        call    iShr            { Result := iShr(A, Temp, True) }
        jmp     @exns

@nneg:  push    esi
        push    edi
        push    ebx

        mov     esi,A           { ESI = address of A }
        mov     edi,Result      { EDI = address of Result }

        push    edi             { Set Result to zero }
        mov     ecx,4
        push    ecx
        xor     eax,eax
        rep
        stosd
        pop     ecx             { ECX = 4 }
        pop     edi

        cmp     eax,[Count + 12]    { Check if Count >= 128 }
        jne     @exit               { If so, Result is by definition zero }
        cmp     eax,[Count + 8]     { NB: EAX = 0 }
        jne     @exit
        cmp     eax,[Count + 4]
        jne     @exit
        mov     ebx,[Count]     { EBX = lower 32 bits of Count }
        mov     edx,ebx
        shr     edx,5           { EDX = Count shr 5 }
        cmp     edx,ecx         { Count >= 128 is the same as Count shr 5 >= 4 }
        jae     @exit

        sub     ecx,edx         { ECX = 4 - Count shr 5; loop count }
        push    edi
        lea     edi,[edi + edx * 4]
        rep                     { Copy from A[0..ECX] to Result[0+EDX..ECX+EDX] }
        movsd                   { ie, copy a whole number of DWORD shifts }
        pop     edi

        mov     ecx,ebx
        and     cl,$1F          { CL = lowest 5 bits of Count, ie, 0..31 }
        jz      @exit           { If CL = 0, no need for bit-wise shifting }

        mov     eax,[edi + 8]       { Copy Count AND $1F bits to the left }
        shld    [edi + 12], eax, cl {   in Result, using a bit-wise shift }
        mov     eax,[edi + 4]
        shld    [edi + 8], eax, cl
        mov     eax,[edi]
        shld    [edi + 4], eax, cl
        shl     eax, cl
        mov     [edi],eax

@exit:  pop     ebx
        pop     edi
        pop     esi
@exns:
end;


{***********************************************************************}
{ Shift the number A to the right by Count bits.  If Signed is True, an
  arithmetic right shift is done, where the right-most Count bits become
  the same as the sign bit in A (the right-most bit).  If Signed is
  False, a logical right shift is done, where the right-most Count bits
  become zero.  Count is treated as a signed number if Signed is True;
  negative right shifts are treated as a call to iShl(A, iNeg(Count), True).
  If Count = 0, A is returned unchanged.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointer to A.  This is because Result is used to store
  intermediate results of the operation }
function iShr (const A, Count : Int128; Signed : boolean) : Int128;

var
  Temp    : Int128;
  iSigned : boolean;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Count, CL = Signed, [EBP+8] = address
    of Result}

        or      Signed,Signed   { Test if Signed is True and Count < 0 }
        jz      @nneg
        cmp     dword ptr [Count + 12], 0    { Test if sign bit is set }
        jns     @nneg

@cltz:  push    eax             { Count < 0; call iShl(A, iNeg(Count), True) }
        mov     eax,Count       { EAX = address of operand for iNeg }
        lea     edx,Temp        { EDX = address of result; use Temp to save }
        push    edx             {   Count from being changed                }
        call    iNeg            { Temp := iNeg(Count) }
        pop     edx             { EDX = address of new Count }
        pop     eax             { EAX = address of A for iShl }
        push    Result          { In iShl, [EBP+8] = address of Result }
        mov     cl,True         { CL = Signed := True }
        call    iShl            { Result := iShl(A, Temp, True) }
        jmp     @exns

@nneg:  push    esi
        push    edi
        push    ebx

        mov     iSigned,Signed  { Preserve Signed (in CL) }
        mov     esi,A           { ESI = address of A }
        mov     edi,Result      { EDI = address of Result }

        xor     eax,eax         { EAX = 0 if not Signed, or sign bit if Signed }
        or      Signed,Signed
        jz      @nsgn
        cmp     [esi + 12],eax  { Test if sign bit on A is set }
        jns     @nsgn
        dec     eax
@nsgn:
        push    edi             { Set Result to either zero or -1, depending }
        mov     ecx,4           {   on whether Signed is True and A < 0      }
        push    ecx
        rep
        stosd
        pop     ecx             { ECX = 4 }
        pop     edi

        xor     eax,eax             { EAX = 0 }
        cmp     eax,[Count + 12]    { Check if Count >= 128 }
        jne     @exit               { If so, Result is by definition zero or -1 }
        cmp     eax,[Count + 8]
        jne     @exit
        cmp     eax,[Count + 4]
        jne     @exit
        mov     ebx,[Count]     { EBX = lower 32 bits of Count }
        mov     edx,ebx
        shr     edx,5           { EDX = Count shr 5 }
        cmp     edx,ecx         { Count >= 128 is the same as Count shr 5 >= 4 }
        jae     @exit

        sub     ecx,edx         { ECX = 4 - Count shr 5; loop count }
        push    edi
        lea     esi,[esi + edx * 4]
        rep                     { Copy from A[0+EDX..ECX+EDX] to Result[0..ECX] }
        movsd                   { ie, copy a whole number of DWORD shifts }
        pop     edi

        mov     ecx,ebx
        and     cl,$1F          { CL = lowest 5 bits of Count, ie, 0..31 }
        jz      @exit           { If CL = 0, no need for bit-wise shifting }

        mov     eax,[edi + 4]   { Copy Count AND $1F bits to the right }
        shrd    [edi], eax, cl  {   in Result, using a bit-wise shift  }
        mov     eax,[edi + 8]
        shrd    [edi + 4], eax, cl
        mov     eax,[edi + 12]
        shrd    [edi + 8], eax, cl
        cmp     iSigned, True   { Final shift is either SHR or SAR, depending }
        je      @ssar           { on if iSigned is False or True respectively }
        shr     eax,cl
        jmp     @sskp
@ssar:  sar     eax,cl
@sskp:  mov     [edi + 12],eax

@exit:  pop     ebx
        pop     edi
        pop     esi
@exns:
end;


{***********************************************************************}
{ Rotate the number A to the left by Count bits.  Count is treated as a
  signed number if Signed is True; negative left rotations are treated as
  a positive right rotation.  If Count = 0, A is returned unchanged.  Note
  that (Count mod 128) is used instead of just Count.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointer to A.  This is because Result is used to store
  intermediate results of the operation }
function iRol (const A, Count : Int128; Signed : boolean) : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Count, CL = Signed, [EBP+8] = address
    of Result}

        push    esi
        push    edi

        mov     esi,A           { ESI = address of A }
        mov     edi,Result      { EDI = address of Result }

        mov     eax,[Count]

        or      Signed,Signed   { Test if Signed is True and Count < 0 }
        jz      @nneg
        cmp     dword ptr [Count + 12], 0    { Test if sign bit is set }
        jns     @nneg
                                { If Signed and (Count < 0) then }
        add     eax,128         {   Count := 128 + Count         }

@nneg:  and     eax,$7F
        mov     edx,eax         { EDX = Count mod 128 }

        mov     ecx,4
        shr     eax,5           { EAX = Count shr 5; whole number of DWORDs }
        sub     ecx,eax         { ECX = 4 - Count shr 5; loop counter }

        push    edi
        lea     edi,[edi + eax * 4]
        rep                     { Copy from A[0..ECX] to Result[0+EAX..ECX+EAX] }
        movsd                   { ie, copy a whole number of DWORD shifts }
        mov     edi,[esp]
        mov     ecx,eax
        jecxz   @skcp           { Transfer the rest of the DWORDs from A to  }
        rep                     {   Result, ie, those that wrap around       }
        movsd                   {   This does not need to be done if ECX = 0 }
@skcp:  pop     edi

        mov     ecx,edx
        and     cl,$1F          { CL = lowest 5 bits of Count, ie, 0..31 }
        jz      @exit           { If CL = 0, no need for bit-wise shifting }

        mov     edx,[edi + 12]      { EDX = bits to save for lowest DWORD }
        mov     eax,[edi + 8]       { Copy Count AND $1F bits to the left  }
        shld    [edi + 12], eax, cl {   in Result, using a bit-wise rotate }
        mov     eax,[edi + 4]
        shld    [edi + 8], eax, cl
        mov     eax,[edi]
        shld    [edi + 4], eax, cl
        shld    eax, edx, cl        { For lowest DWORD, use saved bits in EDX }
        mov     [edi],eax

@exit:  pop     edi
        pop     esi
end;


{***********************************************************************}
{ Rotate the number A to the right by Count bits.  Count is treated as a
  signed number if Signed is True.  Note that a rotation to the right by
  Count bits is the same as a rotation to the left by (128 - Count mod 128);
  this is how this function is implemented.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointer to A.  This is because Result is used to store
  intermediate results of the operation }
function iRor (const A, Count : Int128; Signed : boolean) : Int128;

var
  Temp : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Count, CL = Signed, [EBP+8] = address
    of Result}

        push    esi
        push    edi

        push    eax

        push    ecx
        lea     edi,Temp        { Initialise Temp to zero }
        push    edi
        xor     eax,eax
        mov     ecx,4
        rep
        stosd
        pop     edi
        pop     ecx             { Note that EAX is still on the stack }

        mov     eax,[Count]     { if Signed and (Count < 0) then Temp := -Count }
        neg     eax             { else Temp := 128 - Count                      }
        or      Signed,Signed   {   This is implemented as Temp := -Count,      }
        jz      @nneg           {   followed by the comparison                  }
        cmp     dword ptr [Count + 12], 0
        js      @nskp
@nneg:  add     eax,128         { EAX = 128 + (-Count) }
@nskp:  and     eax,$7F         { Only look at 7 bits, from 0..127, for rotation }
        mov     [edi],eax

        pop     eax             { EAX = address of A }
                                { CL = value of Signed }
        push    Result          { In iRol, [EBP+8] = address of Result }
        mov     edx,edi         { EDX = address of Temp (our new Count) }
        call    iRol            { Result := iRol(A, 128-Count, Signed) }

        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Rotate bits 0..BitSize-1 of the number A to the left by Count bits.
  Count is treated as a signed number if Signed is True; negative left
  rotations are treated as a positive right rotation.  Note that only
  the lower BitSize bits are rotated; the upper bits (if any) are extended
  via a call to iExtend (using Signed as a parameter).  If Count = 0 or
  BitSize <= 1, A is returned appropriately extended.  Also note that
  (Count mod BitSize) is used instead of just Count.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointer to A.  This is because Result is used to store
  intermediate results of the operation }
function iRolBit (const A, Count : Int128; BitSize : integer;
                  Signed : boolean) : Int128;

var
  iBitSize, iCount, iMask, iTemp, iSubTemp : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Count, ECX = BitSize,
    [EBP+12] = Signed, [EBP+8] = address of Result }

        push    esi
        push    edi
        push    ebx

        mov     esi,A           { ESI = address of A }
        mov     edi,Result      { EDI = address of Result }
        mov     eax,BitSize     { EAX = BitSize }
        mov     ecx,dword ptr [Count]  { ECX = Count.L[0] }

        cmp     eax,1           { If BitSize <= 1, Result := A }
        jle     @copy

        cmp     eax,127         { If BitSize > 127, call iRol }
        jg      @irol

        xor     ebx,ebx         { BL = 0 if (not Signed) or (Count >= 0) }
        cmp     Signed,False
        je      @nsgn
        cmp     dword ptr [Count + 12],0
        sets    bl              { BL = 1 if Signed and (Count < 0) }
@nsgn:
        push    eax             { Calculate actual Count for special cases }
        or      bl,bl
        jz      @nadd           { if Signed and (Count < 0) then }
        add     ecx,eax         {   Count := BitSize + Count     }
@nadd:  dec     eax             { Count := Count mod BitSize     }
        and     ecx,eax         { Note that ECX = new value of Count, used }
        pop     eax             {   only during the special cases          }

        cmp     eax,8           { Special case if BitSize = 8 }
        je      @rol8
        cmp     eax,16          { Special case if BitSize = 16 }
        je      @rol16
        cmp     eax,32          { Special case if BitSize = 32 }
        je      @rol32
        cmp     eax,64          { Special case if BitSize = 64 }
        je      @rol64

        lea     ecx,iBitSize    { iBitSize := ZeroExtend(BitSize) }
        mov     [ecx],eax       { Note that we have already checked if }
        xor     eax,eax         {   BitSize is less than 0             }
        mov     [ecx + 4],eax
        mov     [ecx + 8],eax
        mov     [ecx + 12],eax

        mov     eax,Count       { EAX = address of Count }
        mov     edx,ecx         { EDX = address of iBitSize }
        push    eax
        push    edx
        xor     ecx,ecx         { CL = False (for Signed parameter) }
        call    iCmp            { EAX = iCmp(Count, iBitSize, False) }
        or      eax,eax         { Note that this is an UNSIGNED comparison }
        pop     edx
        pop     eax
        js      @skmod          { if Count < iBitSize then iCount := Count   }
                                { else iCount := Count mod iBitSize (Signed) }

        lea     ecx,iCount      { In function, [EBP+8] = address of iCount }
        push    ecx             { EAX and EDX retrieved from above }
        mov     cl,Signed       { CL = Signed }
        call    iMod            { iCount := iMod(Count, iBitSize, Signed) }
        jmp     @cont

                                { Perform a direct iRol since BitSize > 127 }
@irol:                          { Note EDX = address of Count from above }
        mov     eax,esi         { EAX = address of A }
        mov     cl,Signed       { CL = Signed }
        push    edi             { In iRol, [EBP+8] = address of Result }
        call    iRol            { Result := iRol(A, Count, Signed) }
        jmp     @exit

@rol8:  push    eax             { Special case: Rotate 8 bits of A }
        mov     eax,[esi]       { AL = lower 8 bits of A }
        rol     al,cl           { Perform the actual ROL operation }
        jmp     @extd

@rol16: push    eax             { Special case: Rotate 16 bits of A }
        mov     eax,[esi]       { AX = lower 16 bits of A }
        rol     ax,cl           { Perform the actual ROL operation }
        jmp     @extd

@rol32: push    eax             { Special case: Rotate 32 bits of A }
        mov     eax,[esi]       { EAX = lower 32 bits of A }
        rol     eax,cl          { Perform the actual ROL operation }
        jmp     @extd

@rol64: push    eax             { Special case: Rotate 64 bits of A }
        mov     ebx,[esi]       { EBX = A.L[0] }
        mov     edx,[esi + 4]   { EDX = A.L[1] }

        cmp     ecx,32          { SHLD cannot handle Count > 31 }
        jb      @r64x           { Compensate by exchanging A.L[0] and A.L[1] }
        xchg    ebx,edx         {   and making Count <= 31                   }
@r64x:                          { No need to do AND CL,31; done automatically }

        mov     eax,edx
        shld    eax,ebx,cl      { Shift A.L[1] with bits from A.L[0] }
        mov     [edi + 4],eax   {   and store into Result.L[1]       }
        mov     eax,ebx
        shld    eax,edx,cl      { Shift A.L[0] with bits from A.L[1] }
        jmp     @extd           {   and store into Result.L[0]       }

                                { BitSize <= 1 or (Count mod BitSize) = 0 }
@copy:  mov     edx,eax         { Note that EAX = BitSize on entry }
        mov     eax,esi         { EAX = address of A }
        mov     cl,Signed       { CL = Signed }
        push    edi             { In iExtend, [EBP+8] = address of Result }
        call    iExtend         { Result := iExtend(A, BitSize, Signed) }
        jmp     @exit

@skmod: push    esi             { iCount := Count, since Count < iBitSize }
        push    edi             {   (using an unsigned comparison)        }
        mov     esi,eax         { EAX = address of Count, from above }
        lea     edi,iCount
        mov     ecx,4
        rep
        movsd
        pop     edi
        pop     esi

@cont:  mov     eax,dword ptr [iBitSize]  { EAX = iBitSize (mod $100000000) }
        lea     edx,iCount      { EDX = address of iCount }
        mov     ecx,[edx]       { ECX = iCount (mod $100000000) }
                                { Note ESI = address of A, EDI = address of Result }

        jecxz   @copy           { if iCount = 0 then Result := A, extended }
        or      bl,bl           { BL = (Signed and (Count < 0)), from above }
        mov     ebx,eax         { EBX = iBitSize }
        jz      @nneg

        push    eax
        add     eax,ecx         { If Signed and (Count < 0) then }
        mov     ecx,eax         {   iCount := iBitSize + iCount  }
        mov     [edx],ecx       { Store the new iCount, making it positive }
        xor     eax,eax
        mov     [edx + 4],eax
        mov     [edx + 8],eax
        mov     [edx + 12],eax
        pop     eax             { EAX = iBitSize (mod $100000000) }
@nneg:
        push    ebx             { Save iBitSize for @extd2 }

                                { Calculate iMask := ((1 shl             }
                                {   (iBitSize - iCount)) - 1) shl iCount }

                                { Calculate iSubTemp := iBitSize - iCount }
        lea     edx,iSubTemp    { Do this directly instead of calling iSub  }
        sub     eax,ecx         {   since both numbers are guaranteed to be }
        mov     [edx],eax       {   positive and less than 32 bits in size  }
        xor     eax,eax         { Note EAX, EBX = iBitSize from above }
        mov     [edx + 4],eax   { Note ECX = iCount from above }
        mov     [edx + 8],eax
        mov     [edx + 12],eax

        lea     edx,iTemp       { iTemp := 1 }
        mov     [edx + 4],eax   { EAX = 0 from above }
        mov     [edx + 8],eax
        mov     [edx + 12],eax
        inc     eax
        mov     [edx],eax

        lea     eax,iMask       { Calculate iMask := 1 shl iSubTemp }
        push    eax             { Save for next function call }
        push    eax             { In iShl, [EBP+8] = address of iMask }
        mov     eax,edx         { EAX = address of iTemp }
        lea     edx,iSubTemp    { EDX = address of iSubTemp = (iBitSize - iCount) }
        xor     ecx,ecx         { CL = False (for Signed) }
        call    iShl            { iMask := iShl(iTemp, iSubTemp, False) }

                                { Calculate iTemp := (1 shl (iBitSize - iCount)) - 1 }
        mov     eax,[esp]       { Retrieve address of iMask from stack, above}
        lea     edx,iTemp       { iTemp is used instead of iMask as the result   }
        mov     ecx,edx         {   as iShl (below) requires different addresses }
                                {   for A and Result                             }
        call    iSub            { iTemp := iSub(iMask, iTemp) }

        lea     eax,iTemp       { Note address of iMask is already on stack }
        lea     edx,iCount
        xor     ecx,ecx         { CL = False (for Signed parameter) }
        call    iShl            { iMask := iShl(iTemp, iCount, False) }

                                { Perform first rotation }
        mov     eax,esi         { EAX = address of A }
        lea     edx,iCount      { EDX = address of iCount }
        mov     cl,Signed       { CL = Signed }
        push    edi             { In iRol, [EBP+8] = address of Result }
        call    iRol            { Result := iRol(A, iCount, Signed) }

        mov     eax,edi         { Calculate iTemp := Result AND iMask }
        lea     edx,iMask
        push    edx             { Save address of iMask for next function call }
        lea     ecx,iTemp
        call    iAnd            { iTemp := iAnd(Result,iMask) }

        pop     eax             { Retrieve address of iMask from above }
        mov     edx,eax
        call    iNot            { iMask := iNot(iMask) }

        lea     edx,iSubTemp    { Calculate iSubTemp := 128 - iBitSize }
        mov     eax,128         { Since iBitSize is a number less than 128, }
        sub     eax,ebx         {   this can be done directly.  Note that   }
        mov     [edx],eax       {   EBX = iBitSize from above               }
        xor     eax,eax
        mov     [edx + 4],eax
        mov     [edx + 8],eax
        mov     [edx + 12],eax
        push    edx             { Save address of iSubTemp for later }

                                { Perform second rotation }
        push    edx             { In iRol, [EBP+8] = address of iSubTemp }
        mov     eax,edi         { EAX = address of Result }
        xor     ecx,ecx         { CL = False (for Signed parameter) }
        call    iRol            { iSubTemp := iRol(Result, iSubTemp, False }

        mov     eax,[esp]       { Retrieve address of iSubTemp from above }
        lea     edx,iMask       { EDX = address of iMask }
        mov     ecx,eax
        call    iAnd            { iSubTemp := iAnd(iSubTemp, iMask) }
        pop     eax             { Retrieve address of iSubTemp from above }

        lea     edx,iTemp
        mov     ecx,edi
        call    iOr             { Result := iOr(iSubTemp, iTemp) }

        jmp     @extd2          { Extend the result appropriately }

@extd:  mov     [edi],eax       { Store result of ROL from special cases }
@extd2: pop     edx             { Retrieve saved BitSize into EDX }
        push    edi             { In iExtend, [EBP+8] = address of Result }
        mov     eax,edi         { EAX = address of Result }
        mov     cl,Signed       { CL = Signed }
        call    iExtend         { Result := iExtend(Result, BitSize, Signed) }

@exit:  pop     ebx
        pop     edi
        pop     esi
end;


{***********************************************************************}
{ Rotate bits 0..BitSize-1 of the number A to the right by Count bits.
  Count is treated as a signed number if Signed is True.  Note that a
  rotation to the right by Count bits is the same as a rotation to the
  left by (BitSize - Count mod BitSize); this is how this function is
  implemented.

  Notice for assembly programmers:  the Result pointer MUST be different
  from the pointer to A.  This is because Result is used to store
  intermediate results of the operation }
function iRorBit (const A, Count : Int128; BitSize : integer;
                  Signed : boolean) : Int128;

var
  Temp : Int128;

asm
  { This function uses the "register" calling convention.  Hence, on entry,
    EAX = address of A, EDX = address of Count, ECX = BitSize,
    [EBP+12] = Signed, [EBP+8] = address of Result }

        push    edi

        push    eax             { Save address of A for later }

        mov     eax,BitSize
        cmp     eax,1           { if BitSize <= 1 then Perform_iRolBit }
        jle     @rolx
        cmp     eax,127         { if BitSize > 127 then Perform_iRor }
        jg      @iror

        lea     edi,Temp        { BitSize is 2..127 }
        push    edi             { Temp := ZeroExtend(BitSize) }
        stosd
        xor     eax,eax
        stosd
        stosd
        stosd
                                { Calculate new Count: Temp := BitSize - Count }
        mov     eax,[esp]       { EAX = address of Temp, from stack }
        push    ecx             { Store BitSize for later }
        mov     ecx,eax         { ECX = address of Temp }
        call    iSub            { Temp := iSub(Temp, Count) }

        pop     ecx             { ECX = BitSize }
        pop     edx             { EDX = address of Temp (our new Count) }

                                { Perform call to iRolBit }
@rolx:  pop     eax             { EAX = address of A }
        push    dword ptr Signed  { In iRolBit, [EBP+12] = Signed }
        push    Result          { In iRolBit, [EBP+8] = address of Result }
        call    iRolBit         { Result := iRolBit(A, BitSize - Count, BitSize, Signed) }
        jmp     @exit
                                { Perform direct call to iRor }
@iror:  pop     eax             { EAX = address of A, from above }
        push    Result          { Note that EDX is already the address of Count }
        mov     cl,Signed
        call    iRor            { Result := iRor(A, Count, Signed) }

@exit:  pop     edi
end;

end.

