unit Tips;

{************************************************************************
*                                                                       *
*                  Tip of the Day Dialog Box Component                  *
*                   Copyright (C) 1999, John Zaitseff                   *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      23rd July, 1999.
  Version:   1.1

  This unit provides a very simple component that displays a Tip of the
  Day dialog box.  An application may include this component, TTipsDialog,
  on any form and call its Execute method.  For more information on using
  the TTipsDialog component, please see the description below.


  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only'
{$ENDIF}


interface


uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  StdCtrls, ExtCtrls, Forms, Dialogs;


type
  { Tip of the Day form (used internally only) }
  TTipWin     = class;


  { Tip of the Day dialog box component }
  TTipsDialog = class(TComponent)
      private
        FTipWin         : TTipWin;

        FCaption        : string;
        FTipItems       : TStrings;
        FTipIndex       : integer;
        FTipHeading     : string;
        FShowOnStartup  : boolean;

        procedure SetCaption (const Value : string);
        procedure SetTipItems (const Value : TStrings);
        procedure SetTipIndex (const Value : integer);
        procedure SetTipHeading (const Value : string);
        procedure SetShowOnStartup (const Value : boolean);

        procedure NextTipBtnClicked (Sender : TObject);

      public
        constructor Create (AOwner : TComponent); override;
        destructor Destroy; override;

        function Execute : boolean;

      published
        property Caption       : string   read FCaption       write SetCaption;

        property TipItems      : TStrings read FTipItems      write SetTipItems;
        property TipIndex      : integer  read FTipIndex      write SetTipIndex;
        property TipHeading    : string   read FTipHeading    write SetTipHeading;

        property ShowOnStartup : boolean  read FShowOnStartup write SetShowOnStartup;
    end;


{ TTipsDialog component:
  ======================

  The TTipsDialog component provides a dialog box that displays the Tip of
  the Day.  This dialog box provides the following properties:


  Caption : string
  ================

  Setting this property changes the dialog box window caption.  If Caption
  is left blank, the window caption will default to "Tip of the Day".


  TipHeading : string
  ===================

  Setting this property changes the heading within the tip panel.  If this
  is left blank, the dialog box tip heading will default to "Did you
  know..."


  TipItems : TStrings
  ===================

  TipItems is a string list that contains all of the tips to be displayed.
  For more information on using string lists, please see the Borland Delphi
  help for TStrings.  In particular, TipItems.Count returns the number of
  tips (which are numbered from 0 to the count - 1) and TipItems.Add adds to
  the end of the string list.

  Note that TipIndex contains the currently displayed tip from TipItems if
  such is the case.


  TipIndex : integer
  ==================

  This property contains the index of the currently selected tip text.  This
  number is an index into the TipItems string list, and hence must be
  between 0 and TipItems.Count - 1 inclusive.

  ShowOnStartup : boolean
  =======================

  This property contains the setting in the "Show the Tip of the Day the
  next time you start this application" check box.  Note that it is up to
  the application to store this value in an appropriate place (such as the
  system registry) and make sure it fulfils the user's wishes.  In other
  words, setting this value to True will NOT make the Tip of the Day dialog
  box appear automatically --- that is up to the application!


  Tag : integer
  =============

  As with all components, this property may be used for any purpose
  what-so-ever.
}


  { Tip of the Day form (used internally only) }
  TTipWin = class(TForm)
      HintPanel     : TPanel;
      TipImg        : TImage;
      TipHeadingLbl : TLabel;
      TipTextLbl    : TLabel;
      ShowTipsCB    : TCheckBox;
      NextBtn       : TButton;
      CloseBtn      : TButton;

      procedure NextBtnClick (Sender : TObject);
      procedure FormShow (Sender : TObject);

    private
      FOnOurNextTip : TNotifyEvent;
    end;


procedure Register;


implementation


{$R *.DFM}


{***********************************************************************}
{ Create the Tip of the Day dialog box component.  Note that this
  component is a wrapper around a TTipWin form, implemented in this
  unit }
constructor TTipsDialog.Create (AOwner : TComponent);

begin
  inherited Create(AOwner);

  { Set default values }
  FShowOnStartup := True;
  FTipItems := TStringList.Create
end;


{***********************************************************************}
{ Destroy the dialog box component (wrapper), including all allocated
  objects }
destructor TTipsDialog.Destroy;

begin
  FTipWin.Free;
  FTipItems.Free;
  inherited Destroy
end;


{***********************************************************************}
{ Display the Tip of the Day dialog box as a modal form }
function TTipsDialog.Execute : boolean;

begin
  Result := False;
  FTipWin := TTipWin.Create(Application);
  try
    SetCaption(FCaption);
    SetTipHeading(FTipHeading);
    SetTipIndex(FTipIndex);
    SetShowOnStartup(FShowOnStartup);
    FTipWin.FOnOurNextTip := NextTipBtnClicked;

    FTipWin.ShowModal;

    FShowOnStartup := FTipWin.ShowTipsCB.Checked;
    Result := True
  finally
    FTipWin.Free;
    FTipWin := nil
  end
end;


{***********************************************************************}
{ Set the dialog box window caption }
procedure TTipsDialog.SetCaption (const Value : string);

begin
  FCaption := Value;

  if Assigned(FTipWin) and (Value <> '') then
    FTipWin.Caption := Value
end;


{***********************************************************************}
{ Set the list of tips to a new value }
procedure TTipsDialog.SetTipItems (const Value : TStrings);

begin
  if FTipItems <> Value then
    FTipItems.Assign(Value);

  if Assigned(FTipWin) then
    SetTipIndex(FTipIndex)
end;


{***********************************************************************}
{ Set the tip text to an index within the tips list }
procedure TTipsDialog.SetTipIndex (const Value : integer);

begin
  FTipIndex := Value;

  if Assigned(FTipWin) then
    begin
      if FTipItems.Count = 0 then
        begin
          FTipIndex := 0;
          FTipWin.NextBtn.Enabled := False
        end
      else
        begin
          if (Value < 0) or (Value >= FTipItems.Count) then
            FTipIndex := 0;
          FTipWin.TipTextLbl.Caption := FTipItems[FTipIndex];
          FTipWin.NextBtn.Enabled := (FTipItems.Count > 1)
        end
    end
end;


{***********************************************************************}
{ Set the heading within the tip panel.  This is usually set to "Did
  you know..." }
procedure TTipsDialog.SetTipHeading (const Value : string);

begin
  FTipHeading := Value;

  if Assigned(FTipWin) and (Value <> '') then
    FTipWin.TipHeadingLbl.Caption := Value
end;


{***********************************************************************}
{ Set the "Show this Tip of the Day next time" check box to the requested
  value }
procedure TTipsDialog.SetShowOnStartup (const Value : boolean);

begin
  FShowOnStartup := Value;

  if Assigned(FTipWin) then
    FTipWin.ShowTipsCB.Checked := Value
end;


{***********************************************************************}
{ The "Next Tip" button has been clicked in the dialog box by the user.
  Select the next tip and call the OnNextTip event handler, if required }
procedure TTipsDialog.NextTipBtnClicked (Sender : TObject);

begin
  SetTipIndex(FTipIndex + 1)
end;


{***********************************************************************}
{ Centre the Tip of the Day dialog box above its parent form }
procedure TTipWin.FormShow (Sender : TObject);

var
  ParentForm : TForm;

begin
  if Owner is TApplication then
    ParentForm := TApplication(Owner).MainForm
  else if Owner is TForm then
    ParentForm := TForm(Owner)
  else
    ParentForm := nil;

  if Assigned(ParentForm) then
    begin
      Left := (ParentForm.Left + ParentForm.Width div 2) - Width div 2;
      Top := (ParentForm.Top + ParentForm.Height div 2) - Height div 2
    end
  else
    Position := poScreenCenter
end;


{***********************************************************************}
{ Process the "Next Tip" button click }
procedure TTipWin.NextBtnClick (Sender : TObject);

begin
  if Assigned(FOnOurNextTip) then
    FOnOurNextTip(Self)
end;


{***********************************************************************}
{ Register TTipsDialog with the Borland Delphi IDE }
procedure Register;

begin
  RegisterComponents('Base Calculator', [TTipsDialog])
end;

end.

