unit VerRes;

{************************************************************************
*                                                                       *
*                    Application Versioning Routines                    *
*                 Copyright (C) 1997-99, John Zaitseff                  *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@zap.org.au>
  Date:      1st September, 1999.
  Version:   2.3

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.

  This file provides an object, TVerRes, to access the VERSIONINFO
  resource in any executable file.  The following methods and properties
  are available:

  TVerRes.Create
  TVerRes.Create(Filename)
  ========================

  Creates the object, reading in the VERSIONINFO resource from the specified
  filename.  If Filename does not point to a valid file (EXE, DLL, SYS, etc),
  or the file does not have a VERSIONINFO resource, an exception is raised.
  If a blank string, or no parameter, is passed, the current module is used
  (Borland Delphi 4 only).  If only a module handle is available, use the
  Win32 function GetModuleFileName to convert it to a filename before passing
  it to Create.

  TVerRes.Free
  ============

  Destroys the object, freeing up all associated resources.

  TVerRes.FixedInfo  (read-only property)
  =======================================

  Returns the language-independent portion of the VERSIONINFO resource as
  type TFixedInfo.  The TFixedInfo structure has the following fields (all
  fields are 32-bits wide)

      FileVersionMS    - Internal file version, most significant 32 bits
      FileVersionLS    - Internal file version, least significant 32 bits
      ProductVersionMS - Internal product version, most significant 32 bits
      ProductVersionLS - Internal product version, least significant 32 bits
      FileFlagsMask    - A bit-mask of the valid bits in FileFlags
      FileFlags        - A bit-mask of file attributes, defined at compilation
      FileOS           - Operating system for which this file was designed for
      FileType         - General type of file (application, DLL, driver, etc)
      FileSubtype      - Sub-type of the file
      FileDateMS       - File creation date, most significant 32 bits
      FileDateLS       - File creation date, least significant 32 bits

  Warning: FileDateMS and FileDateLS often seem to be set to zero.

  For more information on the TFixedInfo structure, please see the Win32
  Software Development Kit documentation for VS_FIXEDFILEINFO.  This
  documentation is supplied with Borland Delphi.

  TVerRes.LangTable[]     (read-only property)
  TVerRes.LangTableCount  (read-only property)
  ============================================

  The LangTable[] property stores the VERSIONINFO language table as pairs of
  LanguageID and CharSetID (ie, LangTable[i].LanguageID and
  LangTable[i].CharSetID).  LangTableCount contains the number of entries in
  this table.  The first index is 1, the last is LangTableCount.

  TVerRes.CurLangTableIndex  (read/write property)
  ================================================

  The current language table index that is used when retrieving language-
  dependent strings.  Setting this property changes which language is used.

  TVerRes.DefLangTableIndex  (read-only property)
  ===============================================

  The default language table index as determined by TVerRes.  The algorithm
  used to determine this default language is as follows:

      1.  Read in the VarFileInfo\Translation part of the VERSIONINFO
          resource.  This is a table of language-character set pairs of
          languages present in the language-dependent part of the
          VERSIONINFO resource.  This table is stored as LangTable[].
      2.  If the user-specified language is present (via GetUserDefaultLangID),
          use that language.
      3.  If the user-specified language (minus the sub-language ID) matches
          one present (minus the sub-language ID), use that language.
      4.  If the system default language is present (via
          GetSystemDefaultLangID), use that language.
      5.  If the system default language (minus the sub-language ID) matches
          one present (minus the sub-language ID), use that language.
      6.  Failing all of the above, use the first language specified in the
          translation table.

  The sub-language ID differentiates between variants of the same language.
  For example, Windows treats British English and Australian English as
  variants of US English, assigning different sub-language IDs to each.

  TVerRes.VerResString(Res)
  =========================

  Returns the language-specific resource string in the VERSIONINFO resource.
  For example,
      VerResString('FileVersion') may return something like '1.2'.

  TVerRes.CompanyName
  TVerRes.FileDescription
  TVerRes.FileVersion
  TVerRes.ProductName
  TVerRes.ProductVersion
  TVerRes.InternalName
  TVerRes.OriginalFilename
  TVerRes.LegalCopyright
  TVerRes.LegalTrademarks
  TVerRes.PrivateBuild
  TVerRes.SpecialBuild
  TVerRes.Comments
  ========================

  These read-only properties contain the language-dependent strings
  associated with the like-named VERSIONINFO resources.  These are simply
  "short-cuts" to using the VerResString function (eg, TVerRes.CompanyName
  is the same as calling TVerRes.VerResString('CompanyName') ).


  Notes
  =====

  Note that the Win32 functions used to implement TVerRes state that
  they only work on Win32 files, NOT Win16 files.  Although not documented,
  however, using TVerRes with 16-bit EXEs and DLLs DOES seem to work.

  Version information strings can theoretically be of any length.  TVerRes,
  however, limits all strings to at most 512 characters.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only.'
{$ENDIF}


interface


uses
  SysUtils;


type
  { Language-independent version information.  See the Win32 description
    of VS_FIXEDFILEINFO.  Note that FileDateMS and FileDateLS seem to be
    always set to 0 }
  TFixedInfo = record
                 FileVersionMS,    FileVersionLS,
                 ProductVersionMS, ProductVersionLS,
                 FileFlagsMask,    FileFlags,
                 FileOS,
                 FileType,         FileSubtype,
                 FileDateMS,       FileDateLS        : cardinal;
               end;

  { Language table definition }
  TLangTable = record
                 LanguageID,
                 CharSetID   : word
               end;

  { Internal definitions }
  TLangTableArray = array [1..MaxInt div 4] of TLangTable;
  PLangTableArray = ^TLangTableArray;


  { Executable VERSIONINFO resource object definition }
  TVerRes = class
    private
      FVerResPtr          : Pointer;
      FFixedInfo          : TFixedInfo;
      FBlockPrefix        : string;

      FLangTableArray     : PLangTableArray;
      FLangTableAllocated : boolean;
      FLangTableCount     : integer;
      FCurLangTableIndex  : integer;
      FDefLangTableIndex  : integer;

      procedure ReadFixedInfo;
      procedure ReadLangTable;
      function GetLangTable (const AIndex : integer) : TLangTable;
      procedure SetCurLangTableIndex (const AIndex : integer);

      function GetStdVerRes (const AIndex : integer) : string;

    public
{$IFDEF VER90}
      constructor Create (Filename : string);
{$ENDIF}
{$IFDEF VER100}
      constructor Create (Filename : string);
{$ENDIF}
{$IFNDEF VER90}
{$IFNDEF VER100}
      constructor Create (Filename : string = '');
{$ENDIF}
{$ENDIF}

      destructor Destroy; override;

      function VerResString (const Res : string) : string;

      property FixedInfo : TFixedInfo read FFixedInfo;

      property LangTable [const AIndex : integer] : TLangTable read GetLangTable;
      property LangTableCount    : integer read FLangTableCount;
      property CurLangTableIndex : integer read FCurLangTableIndex
                                           write SetCurLangTableIndex;
      property DefLangTableIndex : integer read FDefLangTableIndex;

      property CompanyName       : string index 0  read GetStdVerRes;
      property FileDescription   : string index 1  read GetStdVerRes;
      property FileVersion       : string index 2  read GetStdVerRes;
      property ProductName       : string index 3  read GetStdVerRes;
      property ProductVersion    : string index 4  read GetStdVerRes;
      property InternalName      : string index 5  read GetStdVerRes;
      property OriginalFilename  : string index 6  read GetStdVerRes;
      property LegalCopyright    : string index 7  read GetStdVerRes;
      property LegalTrademarks   : string index 8  read GetStdVerRes;
      property PrivateBuild      : string index 9  read GetStdVerRes;
      property SpecialBuild      : string index 10 read GetStdVerRes;
      property Comments          : string index 11 read GetStdVerRes;
    end;

  { Exception object for TVerRes object }
  EVerRes = class(Exception)
              public
                ErrorCode : integer
            end;


implementation


uses
  Windows;


{$IFDEF VER90}
const
{$ELSE}
ResourceString
{$ENDIF}
  SModuleFilenameErr = 'Could not obtain the module filename';
  SVerReadErr        = 'Could not read VERSIONINFO resource from "%s"';
  SNoVerResErr       = 'No VERSIONINFO resource present in "%s"';
  SFixedInfoErr      = 'Could not read fixed file information from VERSIONINFO resource';
  SNoFixedInfoErr    = 'No fixed file information in VERSIONINFO resource';
  SLangTableRangeErr = 'Language table index %d is out of range [1..%d]';


const
  MaxResStrLen = 512;   { Maximum length of any VERSIONINFO resource string }


const
  StdVerResTable : array [0..11] of string =
    ('CompanyName', 'FileDescription', 'FileVersion', 'ProductName',
     'ProductVersion', 'InternalName', 'OriginalFilename', 'LegalCopyright',
     'LegalTrademarks', 'PrivateBuild', 'SpecialBuild', 'Comments');


{************************************************************************}
{ Create the TVerRes object by reading in the VERSIONINFO resource from
  the file pointed to by Filename.  If the filename is blank, the current
  module's filename is used.  If the file is not found or is in the
  incorrect format (ie, not a Win32 executable), an EInOut exception is
  raised with ErrorCode set to GetLastError.  If the executable is valid,
  but no VERSIONINFO resource exists, an EVerRes exception is raised. }
constructor TVerRes.Create (Filename : string);

var
  Len  : integer;
  Misc : DWORD;
  E    : EVerRes;

begin
  inherited Create;

  { If filename is blank, use the currently executing module's filename }
  if Filename = '' then
    begin
      SetLength(Filename, MAX_PATH);
      Len := GetModuleFileName(0, @Filename[1], MAX_PATH);
      if Len > 0 then
        SetLength(Filename, Len)
      else
        begin
          E := EVerRes.Create(SModuleFilenameErr);
          E.ErrorCode := GetLastError;
          raise E
        end
    end;

  { Get the length of the VERSIONINFO resource }
  Len := GetFileVersionInfoSize(PChar(Filename), Misc);
  if Len > 0 then
    begin
      { Allocate memory and read in the VERSIONINFO resource }
      GetMem(FVerResPtr, Len);
      if not GetFileVersionInfo(PChar(Filename), 0, Len, FVerResPtr) then
        begin
          E := EVerRes.CreateFmt(SVerReadErr, [Filename]);
          E.ErrorCode := GetLastError;
          raise E
        end
    end
  else if (Len = 0) and (GetLastError = 0) then
    { No VERSIONINFO resource is present in the file }
    raise EVerRes.CreateFmt(SNoVerResErr, [Filename])
  else
    begin
      { An error occurred: abort creation of the object }
      E := EVerRes.CreateFmt(SVerReadErr, [Filename]);
      E.ErrorCode := GetLastError;
      raise E
    end;

  ReadFixedInfo;
  ReadLangTable
end;


{************************************************************************}
{ Destroy the TVerRes object by freeing any associated memory }
destructor TVerRes.Destroy;

begin
  if Assigned(FVerResPtr) then
    FreeMem(FVerResPtr);

  if Assigned(FLangTableArray) and FLangTableAllocated then
    FreeMem(FLangTableArray);

  inherited Destroy
end;


{************************************************************************}
{ Read the language-independent portion (ie, fixed file information) of
  the VERSIONINFO resource into the FFixedInfo variable }
procedure TVerRes.ReadFixedInfo;

var
  FixedFilePtr : PVSFixedFileInfo;
  Len          : UINT;
  E            : EVerRes;

begin
  { Get the fixed file information from the VERSIONINFO resource }
  if not VerQueryValue(FVerResPtr, '\', Pointer(FixedFilePtr), Len) then
    begin
      { Could not read in the fixed file information for some reason }
      E := EVerRes.Create(SFixedInfoErr);
      E.ErrorCode := GetLastError;
      raise E
    end
  else if not Assigned(FixedFilePtr) then
    { No fixed file information present in the resource }
    raise EVerRes.Create(SNoFixedInfoErr)
  else
    begin
      { Found the fixed file information.  Copy it to FFixedInfo }
      with FFixedInfo do
        begin
          FileVersionMS := FixedFilePtr^.dwFileVersionMS;
          FileVersionLS := FixedFilePtr^.dwFileVersionLS;
          ProductVersionMS := FixedFilePtr^.dwProductVersionMS;
          ProductVersionLS := FixedFilePtr^.dwProductVersionLS;
          FileFlagsMask := FixedFilePtr^.dwFileFlagsMask;
          FileFlags := FixedFilePtr^.dwFileFlags;
          FileOS := FixedFilePtr^.dwFileOS;
          FileType := FixedFilePtr^.dwFileType;
          FileSubtype := FixedFilePtr^.dwFileSubtype;
          FileDateMS := FixedFilePtr^.dwFileDateMS;
          FileDateLS := FixedFilePtr^.dwFileDateLS
        end
    end
end;


{************************************************************************}
{ Internal procedure to TVerRes.  This reads in the entire language table,
  as stored in the VERSIONINFO resource, allocating enough memory to store
  it in FLangTableArray }
procedure TVerRes.ReadLangTable;

  { Internal function: scan the language table for the matching language
    ID appropriately masked.  Returns True if found (with AIndex set to
    the index in the table), False otherwise }
  function ScanTable (Lang, Mask : word; var AIndex : integer) : boolean;

  var
    I : integer;

  begin
    Result := False;
    for I := 1 to FLangTableCount do
      if (FLangTableArray^[I].LanguageID and Mask) = (Lang and Mask) then
        begin
          AIndex := I;
          Result := True;
          exit
        end
  end;

var
  Len : UINT;

begin { ReadLangTable }
  { If FLangTableArray is not nil, this function has already been called }
  if Assigned(FLangTableArray) then
    exit;

  FLangTableAllocated := False;
  if VerQueryValue(FVerResPtr, '\VarFileInfo\Translation', Pointer(FLangTableArray),
     Len) and Assigned(FLangTableArray) and (Len > 0) then
    begin
      { Store the number of records (not bytes) in the table }
      FLangTableCount := Len div SizeOf(TLangTable);

      { Determine the default language ID, using the algorithm presented in
        the comments at the top of this unit }
      if not ScanTable(GetUserDefaultLangID, $FFFF, FDefLangTableIndex) then
        if not ScanTable(GetUserDefaultLangID, $03FF, FDefLangTableIndex) then
          if not ScanTable(GetSystemDefaultLangID, $FFFF, FDefLangTableIndex) then
            if not ScanTable(GetSystemDefaultLangID, $03FF, FDefLangTableIndex) then
              FDefLangTableIndex := 1
    end
  else
    begin
      { No language table present!  Construct one, assuming US English }
      FLangTableAllocated := True;
      GetMem(FLangTableArray, SizeOf(TLangTable));
      FLangTableCount := 1;
      with FLangTableArray^[1] do
        begin
          LanguageID := $0409;  { US English }
          CharSetID := 1252     { Multilingual }
        end;

      { Set default indices into this table }
      FDefLangTableIndex := 1
    end;

  { Set the default block prefix string }
  FCurLangTableIndex := FDefLangTableIndex;
  with FLangTableArray^[FCurLangTableIndex] do
    FBlockPrefix := Format('\StringFileInfo\%4.4X%4.4X\', [LanguageID, CharSetID])
end;


{************************************************************************}
{ Return the language table values of LanguageID and CharSetID at the
  given index.  If index is out of range (less than 1 or larger than
  LangTableCount), an ERangeError exception is raised }
function TVerRes.GetLangTable (const AIndex : integer) : TLangTable;

begin
  if (AIndex < 1) or (AIndex > LangTableCount) then
    raise ERangeError.CreateFmt(SLangTableRangeErr, [AIndex, LangTableCount]);

  with FLangTableArray^[AIndex] do
    begin
      Result.LanguageID := LanguageID;
      Result.CharSetID := CharSetID
    end
end;


{************************************************************************}
{ Set the current language table index.  This will determine which language
  and character set is used to retrieve language-specific strings.  An
  ERangeError exception is raised if the index is out of range }
procedure TVerRes.SetCurLangTableIndex (const AIndex : integer);

begin
  if (AIndex < 1) or (AIndex > LangTableCount) then
    raise ERangeError.CreateFmt(SLangTableRangeErr, [AIndex, LangTableCount]);

  if AIndex = FCurLangTableIndex then
    exit;

  FCurLangTableIndex := AIndex;
  with FLangTableArray^[FCurLangTableIndex] do
    FBlockPrefix := Format('\StringFileInfo\%4.4X%4.4X\', [LanguageID, CharSetID])
end;


{************************************************************************}
{ Return the value stored in the language-dependent portion of the
  VERSIONINFO resource that has the keyword matching Res.  If no such
  keyword exists, a blank string is returned.  The language used to
  retrieve the string is that language in LangTable[CurLangTableIndex].LanguageID }
function TVerRes.VerResString (const Res : string) : string;

var
  SubBlock : string;
  Len      : UINT;
  P        : PChar;

begin
  SubBlock := FBlockPrefix + Res;

  Len := MaxResStrLen;
  if VerQueryValue(FVerResPtr, PChar(SubBlock), Pointer(P), Len) then
    Result := P
  else
    Result := ''
end;


{************************************************************************}
{ Return a standard version resource string, using the StdVerResTable
  table.  This is an internal function }
function TVerRes.GetStdVerRes (const AIndex : integer) : string;

begin
  Result := VerResString(StdVerResTable[AIndex])
end;

end.

